#include "dependency_default.h"
#include "dependency_default_audio.h"

static bool g_voice_key_pressed = false;
static mas_proxy_interface g_proxy_interface;

#include <stdlib.h>
#include <thread>
#include <atomic>
#include <chrono>

#include <dlog.h>
#include <audio_io.h>
#include <sound_manager.h>
#include <sound_manager_internal.h>

using namespace std;

static audio_in_h g_audio_in = NULL;
static audio_io_state_e g_audio_io_state = AUDIO_IO_STATE_IDLE;

static sound_stream_info_h g_volume_stream = NULL;
static virtual_sound_stream_h g_virtual_sound_stream = NULL;

static int g_device_id = -1;
static sound_stream_info_h g_stream_info = NULL;

static thread g_recorder_thread;
static atomic_bool g_stop_recorder_thread{false};

static long get_current_milliseconds_after_epoch()
{
	auto now = chrono::system_clock::now();
	auto now_ms = chrono::time_point_cast<chrono::milliseconds>(now);
	/* number of milliseconds since the epoch of system_clock */
	auto value = now_ms.time_since_epoch();

	return value.count();
}

static int _apply_device_for_stream_routing(void *user_data)
{
	sound_device_list_h deviceList = NULL;
	sound_device_h device = NULL;
	sound_device_type_e type;
	sound_device_io_direction_e ioDirection;

	if (0 != sound_manager_get_device_list(SOUND_DEVICE_IO_DIRECTION_IN_MASK, &deviceList)) {
		LOGE("[Recorder ERROR] Fail to get current device list");
		return -1;
	}

	bool isFound = false;
	while (0 == sound_manager_get_next_device(deviceList, &device)) {
		if (0 != sound_manager_get_device_type(device, &type)) {
			LOGE("[Recorder ERROR] Fail to get device type");
			continue;
		}
		if (0 != sound_manager_get_device_io_direction(device, &ioDirection)) {
			LOGE("[Recorder ERROR] Fail to get device io direction");
			continue;
		}

		if (SOUND_DEVICE_USB_AUDIO == type && SOUND_DEVICE_IO_DIRECTION_IN == ioDirection) {
			if (0 != sound_manager_add_device_for_stream_routing(g_stream_info, device)) {
				LOGE("[Recorder ERROR] Fail to add device");
				continue;
			}
			if (0 != sound_manager_apply_stream_routing(g_stream_info)) {
				LOGE("[Recorder ERROR] Fail to apply stream routing");
				continue;
			}
			isFound = true;
			break;
		}
	}

	sound_manager_free_device_list(deviceList);
	deviceList = NULL;

	if (true != isFound) {
		LOGI("[Recorder] No USB device");
	} else {
		LOGD("[Recorder] Apply device for stream routing");
	}
	return 0;
}

static void _device_connection_changed_cb(sound_device_h device, bool isConnected, void *user_data)
{
	sound_device_type_e type;
	if (isConnected) {
		if (0 != sound_manager_get_device_type(device, &type)) {
			LOGE("[Recorder ERROR] Fail to get device type");
			return;
		}
		if (type == SOUND_DEVICE_USB_AUDIO) {
			if (0 != sound_manager_remove_device_for_stream_routing(g_stream_info, device))
				LOGE("[Recorder ERROR] Fail to remove device");

			if (0 != sound_manager_add_device_for_stream_routing(g_stream_info, device)) {
				LOGE("[Recorder ERROR] Fail to add device");
				return;
			}
			if (0 != sound_manager_apply_stream_routing(g_stream_info)) {
				LOGE("[Recorder ERROR} Fail to apply stream routing");
				return;
			}
			LOGD("[Recorder] Apply device for stream routing");
		}
	}
	return;
}

static void _audio_in_state_changed_cb(audio_in_h handle, audio_io_state_e previous, audio_io_state_e current, bool by_policy, void *user_data)
{
	if (handle == g_audio_in) {
		g_audio_io_state = current;
	}
}

void dependency_default_audio_initialize(mas_proxy_interface interfaces)
{
	const int rate = 16000;
	const audio_channel_e channel = AUDIO_CHANNEL_MONO;
	const audio_sample_type_e type = AUDIO_SAMPLE_TYPE_S16_LE;

	g_proxy_interface = interfaces;

	int ret = audio_in_create(rate, channel, type, &g_audio_in);
	if (AUDIO_IO_ERROR_NONE != ret) {
		LOGE("[Recorder ERROR] Rate(%d) Channel(%d) Type(%d)", rate, channel, type);
		LOGE("[Recorder ERROR] Fail to create audio handle : %d", ret);
		return;
	}

	ret = sound_manager_add_device_connection_changed_cb(SOUND_DEVICE_IO_DIRECTION_IN_MASK, _device_connection_changed_cb, NULL, &g_device_id);
	if (0 != ret) {
		LOGE("[Recorder ERROR] Fail to add device connection changed callback");
		audio_in_destroy(g_audio_in);
		g_audio_in = NULL;
		return;
	}

	ret = sound_manager_create_stream_information_internal(
		SOUND_STREAM_TYPE_VOICE_RECOGNITION_SERVICE, NULL, NULL, &g_stream_info);
	if (0 != ret) {
		LOGE("[Recorder ERROR] Fail to create stream info");
		sound_manager_remove_device_connection_changed_cb(g_device_id);
		audio_in_destroy(g_audio_in);
		g_audio_in = NULL;
		return;
	}

	ret = _apply_device_for_stream_routing(NULL);
	if (0 != ret) {
		LOGE("[Recorder ERROR] Fail to apply device for stream routing: %d", ret);
		sound_manager_remove_device_connection_changed_cb(g_device_id);
		sound_manager_destroy_stream_information(g_stream_info);
		g_stream_info = NULL;
		audio_in_destroy(g_audio_in);
		g_audio_in = NULL;
		return;
	}

	ret = audio_in_set_state_changed_cb(g_audio_in, _audio_in_state_changed_cb, NULL);
	if (AUDIO_IO_ERROR_NONE != ret) {
		LOGE("[Recorder ERROR] Fail to set state changed callback : %d", ret);
		sound_manager_remove_device_connection_changed_cb(g_device_id);
		sound_manager_destroy_stream_information(g_stream_info);
		g_stream_info = NULL;
		audio_in_destroy(g_audio_in);
		g_audio_in = NULL;
		return;
	}

	ret = audio_in_set_sound_stream_info(g_audio_in, g_stream_info);
	if (AUDIO_IO_ERROR_NONE != ret) {
		LOGE("[Recorder ERROR] Fail to set stream info : %d", ret);
		sound_manager_remove_device_connection_changed_cb(g_device_id);
		sound_manager_destroy_stream_information(g_stream_info);
		g_stream_info = NULL;
		audio_in_unset_state_changed_cb(g_audio_in);
		audio_in_destroy(g_audio_in);
		g_audio_in = NULL;
		return;
	}

	ret = audio_in_prepare(g_audio_in);
	if (AUDIO_IO_ERROR_NONE != ret) {
		if (AUDIO_IO_ERROR_SOUND_POLICY == ret)
		{
			LOGE("[Recorder ERROR] Audio is busy.");
		} else {
			LOGE("[Recorder ERROR] Fail to start audio : %d", ret);
		}
		sound_manager_remove_device_connection_changed_cb(g_device_id);
		sound_manager_destroy_stream_information(g_stream_info);
		g_stream_info = NULL;
		audio_in_unset_state_changed_cb(g_audio_in);
		audio_in_destroy(g_audio_in);
		g_audio_in = NULL;
		return;
	}

	ret = audio_in_pause(g_audio_in);
	if (AUDIO_IO_ERROR_NONE != ret) {
		LOGE("[Recorder ERROR] Fail to pause audio in : %d", ret);
	}
}

void dependency_default_audio_deinitialize()
{
	dependency_default_audio_stop_recording();

	int ret = 0;
	if (g_virtual_sound_stream) {
		sound_manager_stop_virtual_stream(g_virtual_sound_stream);
		ret = sound_manager_destroy_virtual_stream(g_virtual_sound_stream);
		if (0 != ret) {
			LOGE("[Audio ERROR] Fail to destroy virtual stream, ret(%d)", ret);
		}
		g_virtual_sound_stream = NULL;
	}

	if (g_device_id) {
		ret = sound_manager_remove_device_connection_changed_cb(g_device_id);
		if (0 != ret)
			LOGE("[Audio ERROR] Fail to remove device connection changed callback, ret(%d)", ret);
	}

	if (g_volume_stream) {
		ret = sound_manager_destroy_stream_information(g_volume_stream);
		if (0 != ret) {
			LOGE("[Audio ERROR] Fail to destroy stream information, ret(%d)", ret);
		}
		g_volume_stream = NULL;
	}

	if (g_audio_in) {
		ret = audio_in_unprepare(g_audio_in);
		if (AUDIO_IO_ERROR_NONE != ret) {
			LOGD("[Recorder ERROR] Fail to stop audio : %d", ret);
		}
	}

	if (g_stream_info) {
		if (0 != sound_manager_destroy_stream_information(g_stream_info)) {
			LOGD("[Recorder ERROR] Fail to destroy stream info");
		}
		g_stream_info = NULL;
	}

	if (g_audio_in) {
		ret = audio_in_unset_state_changed_cb(g_audio_in);
		if (AUDIO_IO_ERROR_NONE != ret) {
			LOGD("[Recorder ERROR] Fail to unset state changed callback : %d", ret);
		}

		ret = audio_in_destroy(g_audio_in);
		if (AUDIO_IO_ERROR_NONE != ret) {
			LOGD("[Recorder ERROR] Fail to destroy audio : %d", ret);
		}
		g_audio_in = NULL;
	}
}

static void recorder_thread_func()
{
	const int FRAME_LENGTH = 160;
	const int BUFFER_LENGTH = FRAME_LENGTH * 2;

	static int buffer_count = 0;

	while (!(g_stop_recorder_thread.load())) {
		unsigned char buffer[BUFFER_LENGTH];
		memset(buffer, '\0', BUFFER_LENGTH);

		int read_bytes = audio_in_read(g_audio_in, buffer, BUFFER_LENGTH);
		if (0 > read_bytes) {
			LOGE("[Recorder WARNING] Fail to read audio : %d", read_bytes);
			break;
		}

		g_proxy_interface.feed_audio_data(WAKEUP_SPEECH_STREAMING_EVENT_CONTINUE,
			buffer, read_bytes);

		/* Audio read log */
		if (0 == buffer_count % 300) {
			LOGD("[Recorder][%d] Recording... : read_size(%d)", buffer_count, read_bytes);
		}

		buffer_count++;
	}
}

void dependency_default_audio_start_recording()
{
	if (g_audio_in) {
		int ret = AUDIO_IO_ERROR_NONE;
		if (AUDIO_IO_STATE_RUNNING != g_audio_io_state) {
			ret = audio_in_resume(g_audio_in);
		}
		if (AUDIO_IO_ERROR_NONE != ret) {
			LOGD("[Recorder ERROR] Fail to resume audio in : %d", ret);
		}
		g_stop_recorder_thread.store(false);
		LOGD("Starting recorder thread");
		g_recorder_thread = thread(recorder_thread_func);
	}
}

void dependency_default_audio_stop_recording()
{
	if (g_recorder_thread.joinable()) {
		LOGD("g_recorder_thread is joinable, trying join()");
		g_stop_recorder_thread.store(true);
		g_recorder_thread.join();
	}
	if (g_audio_in) {
		int ret = AUDIO_IO_ERROR_NONE;
		if (AUDIO_IO_STATE_RUNNING == g_audio_io_state) {
			ret = audio_in_pause(g_audio_in);
		}
		if (AUDIO_IO_ERROR_NONE != ret) {
			LOGD("[Recorder ERROR] Fail to pause audio in : %d", ret);
		}
	}
}

void dependency_default_audio_set_recording_session(unsigned int session)
{
	LOGD("Current recording session : [%s]",
		(RECORDING_SESSION_WAKE_WORD == session) ? "wake word detecting" :
		(RECORDING_SESSION_UTTERANCE == session) ? "utterance recording" :
		(RECORDING_SESSION_FOLLOW_UP == session) ? "follow-up speech recording" :
		"ERROR - Invalid recording session information"
	);
}

void dependency_default_audio_voice_key_pressed_set(bool pressed)
{
	g_voice_key_pressed = pressed;
}

void dependency_default_audio_set_background_volume(double ratio)
{
	int ret;

	if (ratio >= 0.5) {
		if (g_virtual_sound_stream) {
			ret = sound_manager_stop_virtual_stream(g_virtual_sound_stream);
			if (0 != ret) {
					LOGD("[Audio ERROR] Fail to stop virtual stream, ret(%d)", ret);
			}
			ret = sound_manager_destroy_virtual_stream(g_virtual_sound_stream);
			if (0 != ret) {
					LOGD("[Audio ERROR] Fail to destroy virtual stream, ret(%d)", ret);
					return;
			}
			g_virtual_sound_stream = NULL;
		}

		if (g_volume_stream) {
			ret = sound_manager_destroy_stream_information(g_volume_stream);
			if (0 != ret) {
					LOGD("[Audio ERROR] Fail to destroy stream information, ret(%d)", ret);
					return;
			}
			g_volume_stream = NULL;
		}
	} else {
		if (!g_volume_stream) {
			ret = sound_manager_create_stream_information_internal(SOUND_STREAM_TYPE_VOICE_RECOGNITION_SERVICE, NULL, NULL, &g_volume_stream);
			if (0 != ret) {
				LOGD("[Audio] Fail to create stream information, ret(%d)", ret);
				return;
			}
		}

		if (!g_virtual_sound_stream) {
			ret = sound_manager_create_virtual_stream(g_volume_stream, &g_virtual_sound_stream);
			if (0 != ret) {
				LOGD("[Audio ERROR] Fail to create virtual stream, ret(%d)", ret);
				return;
			}
		}
		if (g_virtual_sound_stream) {
			ret = sound_manager_start_virtual_stream(g_virtual_sound_stream);
			if (0 != ret) {
				LOGD("[Audio ERROR] Fail to start virtual stream, ret(%d)", ret);
				return;
			}
		}
	}
}

void dependency_default_audio_get_audio_format(int* rate, int* channel, int* audio_type)
{
	if (!audio_type || !rate || !channel) {
		return;
	}
	*rate = 16000;
	*channel = 0;
	*audio_type = 0;
}

void dependency_default_audio_get_audio_source_type(char** type)
{
	static char source_type[] = "default_audio";
	if (!type) {
		return;
	}
	*type = source_type;
}
