/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __HEAP_TRACER_H__
#define __HEAP_TRACER_H__

#include <list>
#include <mutex>
#include <unordered_map>
#include <string>
#include <set>

#ifndef __MODULE__
#define __MODULE__ (strrchr(__FILE__, '/') ? strrchr(__FILE__, '/') + 1 : __FILE__)
#endif

#define vm_calloc_simple(nmemb, size) vm_calloc(nmemb, size, "%s: %s(%d)", __MODULE__, __func__, __LINE__)
#define vm_malloc_simple(size) vm_malloc(size, "%s: %s(%d)", __MODULE__, __func__, __LINE__)
#define vm_free_simple(ptr) vm_free(ptr, "%s: %s(%d)", __MODULE__, __func__, __LINE__)
#define vm_strdup_simple(ptr) vm_strdup(ptr, "%s: %s(%d)", __MODULE__, __func__, __LINE__)
#define vm_mark_unmanaged_simple(ptr) vm_mark_unmanaged(ptr, "%s: %s(%d)", __MODULE__, __func__, __LINE__)

void *vm_calloc(size_t nmemb, size_t size, const char *fmt, ...);
void *vm_malloc(size_t size, const char *fmt, ...);
void vm_free(void *ptr, const char *fmt, ...);
char *vm_strdup(const char *s, const char *fmt, ...);
void *vm_mark_unmanaged(void *ptr, const char *fmt, ...);

class HeapTracer {
public:
    HeapTracer();
    virtual ~HeapTracer();

    bool Insert(void *ptr, std::string description, size_t size = 0);
    bool Delete(void *ptr, std::string description);
    std::string Find(void *ptr);

    void MarkUnmanaged(void *ptr, std::string description);

    bool Empty();
    void Trace();
protected:
    typedef struct {
        void *ptr{nullptr};
        std::string description;
        int size{0};
    } AllocationEntry;

    std::mutex mManagedTableMutex;
    std::unordered_map<void*, AllocationEntry> mManagedTable;

    std::mutex mUnmanagedRecordsMutex;
    std::list<AllocationEntry> mUnmanagedRecords;
    size_t mUnmanagedRecordsSize{0};

    enum class EntryOperation {
        INSERT,
        DELETE,
    };
    typedef struct {
        AllocationEntry entry;
        EntryOperation operation;
        struct timespec ts;
    } EntryHistory;
    std::list<EntryHistory> mEntryHistory;
};

#endif /* __HEAP_TRACER_H__ */