/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef _WAKEUP_AUDIO_MANAGER_H_
#define _WAKEUP_AUDIO_MANAGER_H_

#include "wakeup_interfaces.h"

#include <atomic>
#include <list>
#include <mutex>
#include <thread>
#include <vector>

#include <audio_io.h>
#include <sound_manager.h>
#include <sound_manager_internal.h>

namespace multiassistant
{
namespace wakeup
{

using namespace std;

class IAudioEventObserver
{
public:
	virtual ~IAudioEventObserver() = default;
	virtual bool on_recording_audio_data(long time, void* data, int len) = 0;
	virtual bool on_streaming_audio_data(
		wakeup_speech_streaming_event_e event, void* buffer, unsigned int len) = 0;
};

typedef enum
{
	RECORDING_SESSION_WAKE_WORD,
	RECORDING_SESSION_UTTERANCE,
	RECORDING_SESSION_FOLLOW_UP,
} recording_session;

class CAudioManager
{
public:
	CAudioManager();
	CAudioManager(IAudioEventObserver *observer);
	~CAudioManager();

	CAudioManager(const CAudioManager&) = delete;
	CAudioManager& operator=(const CAudioManager&) = delete;

	int initialize();
	int deinitialize();

	void sound_focus_changed();

	void subscribe(IAudioEventObserver *observer);
	void unsubscribe(IAudioEventObserver *observer);

	void start_recording(bool proactive);
	void stop_recording(bool proactive);

	void set_recording_session(recording_session session);

	void feed_audio_data(wakeup_speech_streaming_event_e event, void* buffer, int len);
	void finalize_audio_data();
	void clear_audio_data();

	void start_streaming_current_utterance_data(long start_time = 0);
	void stop_streaming_current_utterance_data();

	void start_streaming_previous_utterance_data();
	void stop_streaming_previous_utterance_data();

	void start_streaming_follow_up_data();
	void stop_streaming_follow_up_data();

	void set_background_volume(double ratio);
private:
	void add_audio_data(wakeup_speech_data& data, long time);
	void notify_audio_data_recording(long time, void* data, int len);

	void streaming_previous_audio_data_thread_func();
	void streaming_audio_data_thread_func(long start_time);

	int mSoundFocusWatchId{0};
	bool mRecordingRequired{false};
	bool mIsRecording{false};

	vector<IAudioEventObserver*> mObservers;

	thread mStreamingThread;
	atomic_bool mStopStreamingThread{false};

	thread mStreamingPreviousThread;
	atomic_bool mStopStreamingPreviousThread{false};

	static constexpr long mAudioRecordingDurationMilliseconds = 10 * 1000;
	typedef struct {
		long time;
		wakeup_speech_data data;
	} wakeup_speech_data_with_time;
	vector<wakeup_speech_data_with_time> mPreviousAudioData;
	list<wakeup_speech_data_with_time> mAudioData;

	mutex mMutex;
	bool mVoiceKeyPressed{false};
};

} // wakeup
} // multiassistant

#endif /* _WAKEUP_AUDIO_MANAGER_H_ */
