/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef _WAKEUP_MANAGER_H_
#define _WAKEUP_MANAGER_H_

#include "wakeup_manager_wrapper.h"
#include "wakeup_settings.h"
#include "wakeup_engine_manager.h"
#include "wakeup_audio_manager.h"
#include "wakeup_policy_default.h"

#include <memory>
#include <map>

namespace multiassistant
{
namespace wakeup
{

using namespace std;

enum class STREAMING_MODE {
	NONE,
	UTTERANCE,
	PREVIOUS_UTTERANCE,
	FOLLOW_UP,
};

class IWakeupEventObserver {
public:
	virtual ~IWakeupEventObserver() = default;
	virtual void on_wakeup(wakeup_event_info wakeup_info) = 0;
	virtual void on_streaming_audio_data(
		wakeup_speech_streaming_event_e event, void* buffer, unsigned int len) = 0;
	virtual void on_audio_streaming_data_section(ma_audio_streaming_data_section_e section) = 0;
};

class ISettingValueObserver {
public:
	virtual ~ISettingValueObserver() = default;
	virtual void on_value_changed() = 0;
};

/* If a wakeup event is raised by pressing a voice key,
	the wakeup_engine attribute of wakeup_event_info will have the following value */
#define WAKEUP_ENGINE_VOICE_KEY "voice_key"

class CWakeupManager
{
public:
	CWakeupManager(IWakeupEventObserver* wakeup_observer, ISettingValueObserver* setting_observer);
	virtual ~CWakeupManager();

	CWakeupManager(const CWakeupManager&) = delete;
	CWakeupManager& operator=(const CWakeupManager&) = delete;

	bool initialize();
	bool deinitialize();

	void subscribe_wakeup_observer(IWakeupEventObserver* observer);
	void unsubscribe_wakeup_observer(IWakeupEventObserver* observer);

	void subscribe_setting_observer(ISettingValueObserver* observer);
	void unsubscribe_setting_observer(ISettingValueObserver* observer);

	bool activate();
	bool deactivate();

	bool add_assistant_language(string appid, string language);
	bool add_assistant_wakeup_word(string appid, string wakeup_word, string language);
	bool set_assistant_wakeup_engine(string appid, string engine);
	bool set_assistant_language(string appid, string language);

	bool set_assistant_enabled(string appid, bool enabled);
	bool get_assistant_enabled(string appid);
	bool set_default_assistant(string appid);
	string get_default_assistant();

	bool update_voice_feedback_state(string appid, bool state);
	bool send_assistant_specific_command(string appid, string command);
	bool set_background_volume(string appid, double ratio);
	bool update_recognition_result(string appid, int result);
	bool process_event(ma_plugin_event_e event, void* data, int len);
	bool get_audio_format(int* rate, int* channel, int* audio_type);
	bool get_audio_source_type(char** type);
	bool set_language(string language);
	bool get_voice_key_pressed();
	bool set_wake_word_audio_require_flag(bool require);

	STREAMING_MODE get_streaming_mode();
	bool set_streaming_mode(STREAMING_MODE mode);

	bool start_streaming_utterance_data();
	bool stop_streaming_utterance_data();
	bool start_streaming_follow_up_data();
	bool stop_streaming_follow_up_data();
	bool start_streaming_previous_utterance_data();
	bool stop_streaming_previous_utterance_data();

	CWakeupPolicy* get_wakeup_policy();
	CWakeupEngineManager* get_engine_manager();
	CAudioManager* get_audio_manager();
	CWakeupSettings* get_wakeup_settings();

	vector<IWakeupEventObserver*> get_wakeup_observers();
	void set_last_wakeup_event_info(wakeup_event_info wakeup_info);

	vector<ISettingValueObserver*> get_setting_observers();

	bool change_manager_state(wakeup_manager_state_e state);
	wakeup_manager_state_e get_manager_state();

	void feed_audio_data(wakeup_speech_streaming_event_e event, void* buffer, int len);
private:
	class CEngineEventObserver : public IEngineEventObserver
	{
	public:
		bool on_wakeup_event(string engine_name, wakeup_event_info wakeup_info) override;
		bool on_speech_status(string engine_name, wakeup_service_speech_status_e status) override;
		bool on_error(string engine_name, int error_code, string error_message) override;
		bool on_audio_data_require_status(string engine_name, bool require) override;

		bool on_streaming_audio_data(
			wakeup_speech_streaming_event_e event, void* buffer, unsigned int len) override;
		bool on_audio_streaming_data_section(ma_audio_streaming_data_section_e section) override;

		void set_wakeup_manager(CWakeupManager *manager) { mWakeupManager = manager; }
	private:
		CWakeupManager *mWakeupManager{nullptr};
	};

	class CPolicyEventObserver : public IPolicyEventObserver
	{
	public:
		void on_wakeup(wakeup_event_info wakeup_info) override;

		void set_wakeup_manager(CWakeupManager *manager) { mWakeupManager = manager; }
	private:
		CWakeupManager *mWakeupManager{nullptr};
	};

	class CAudioEventObserver : public IAudioEventObserver
	{
	public:
		bool on_recording_audio_data(long time, void* data, int len) override;
		bool on_streaming_audio_data(
			wakeup_speech_streaming_event_e event, void* buffer, unsigned int len) override;

		void set_wakeup_manager(CWakeupManager *manager) { mWakeupManager = manager; }
		void set_wakeup_engine_manager(CWakeupEngineManager *manager) { mEngineManager = manager; }
	private:
		CWakeupManager *mWakeupManager{nullptr};
		CWakeupEngineManager *mEngineManager{nullptr};
	};

	class CSettingsEventObserver : public ISettingsEventObserver
	{
	public:
		bool on_voice_input_language_changed(const char* language) override;
		bool on_assistant_enabled_info_changed(const char* appid, bool enabled) override;
		bool on_default_assistant_appid_changed(const char* appid) override;

		void set_wakeup_manager(CWakeupManager *manager) { mWakeupManager = manager; }
	private:
		CWakeupManager *mWakeupManager{nullptr};
	};

	void initialize_wakeup_policy();

	typedef struct {
		string appid;
		vector<string> languageList;
	} AssistantLanguageInfo;
	vector<AssistantLanguageInfo> mAssistantLanguageInfo;
	map<string, bool> mAssistantSupportsCurrentLanguage;
	map<string, bool> mAssistantEnabled;

	vector<IWakeupEventObserver*> mWakeupObservers;
	vector<ISettingValueObserver*> mSettingObservers;

	unique_ptr<CWakeupPolicy> mWakeupPolicy;

	CAudioManager mAudioManager;
	CWakeupEngineManager mWakeupEngineManager;
	CWakeupSettings mWakeupSettings;

	CAudioEventObserver mAudioEventObserver;
	CEngineEventObserver mEngineEventObserver;
	CPolicyEventObserver mPolicyEventObserver;
	CSettingsEventObserver mSettingsEventObserver;

	thread mEngineDataThread;
	atomic_bool mStopEngineDataThread{false};

	bool mVoiceKeyPressed{false};
	string mCurrentLanguage;
	string mCurrentDefaultAssistant;

	STREAMING_MODE mStreamingMode{STREAMING_MODE::NONE};
	Ecore_Timer* mStreamingDurationTimer{nullptr};

	wakeup_manager_state_e mWakeupManagerState{WAKEUP_MANAGER_STATE_INACTIVE};

	wakeup_event_info mLastWakeupEventInfo;
};

} // wakeup
} // multiassistant

#endif /* _WAKEUP_MANAGER_H_ */