/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef _WAKEUP_MANAGER_WRAPPER_H_
#define _WAKEUP_MANAGER_WRAPPER_H_

#include <dlog.h>
#include <stdbool.h>
#include <tizen.h>
#include <tzplatform_config.h>

#include "wakeup_interfaces.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {
	WAKEUP_SPEECH_STATUS_NONE = -1, 					/**< None */
	WAKEUP_SPEECH_STATUS_BEGINNING_POINT_DETECTED = 1,	/**< Beginning point of speech is detected */
	WAKEUP_SPEECH_STATUS_END_POINT_DETECTED = 2 		/**< End point of speech is detected */
} wakeup_service_speech_status_e;

typedef enum {
	WAKEUP_MANAGER_STATE_INACTIVE = 0,
	WAKEUP_MANAGER_STATE_LISTENING = 1,
	WAKEUP_MANAGER_STATE_UTTERANCE = 2,
	WAKEUP_MANAGER_STATE_PROCESSING = 3,
	WAKEUP_MANAGER_STATE_VOICE_FEEDBACK = 4
} wakeup_manager_state_e;

typedef enum {
	MA_PREPROCESSING_ALLOW_NONE = 0,	/**< No preprocessing allowed */
	MA_PREPROCESSING_ALLOW_UTTERANCE,	/**< Preprocessing allowed for utterance audio only */
	MA_PREPROCESSING_ALLOW_FOLLOW_UP,	/**< Preprocessing allowed for follow-up audio only */
	MA_PREPROCESSING_ALLOW_ALL,			/**< Preprocessing allowed for all audio */
} ma_preprocessing_allow_mode_e;

typedef enum {
	MA_AUDIO_STREAMING_DATA_SECTION_UTTERANCE = 0,	/**< Utterance section started */
	MA_AUDIO_STREAMING_DATA_SECTION_WAKE_WORD,		/**< Wake word section started */
} ma_audio_streaming_data_section_e;

typedef void (*ma_audio_streaming_data_section_changed_cb)(ma_audio_streaming_data_section_e section, void* user_data);

typedef void (*wakeup_service_wakeup_event_cb)(wakeup_event_info wakeup_info, void* user_data);

typedef void (*wakeup_service_speech_streaming_cb)(wakeup_speech_streaming_event_e event, void* buffer, int len, void *user_data);

typedef void (*wakeup_service_speech_status_cb)(wakeup_service_speech_status_e status, void *user_data);

typedef void (*wakeup_service_setting_changed_cb)( void *user_data);

typedef void (*wakeup_service_error_cb)(int error, const char* err_msg, void* user_data);

typedef void (*wakeup_service_audio_data_require_status_cb)(bool require, void* user_data);

typedef void (*wakeup_service_streaming_section_changed_cb)(ma_audio_streaming_data_section_e section, void* user_data);

typedef enum {
	MA_PLUGIN_EVENT_VOICE_KEY_PRESSED = 0,
	MA_PLUGIN_EVENT_VOICE_KEY_RELEASED,
	MA_PLUGIN_EVENT_VOICE_KEY_RELEASED_AFTER_PUSH = MA_PLUGIN_EVENT_VOICE_KEY_RELEASED,
	MA_PLUGIN_EVENT_VOICE_KEY_RELEASED_AFTER_TAP,
} ma_plugin_event_e;

typedef enum {
	MA_SYSTEM_VOLUME_EVENT_CHANGE,
	MA_SYSTEM_VOLUME_EVENT_RECOVER
} ma_system_volume_event_e;

typedef struct {
	int plugin_version;
	bool ui_panel_enabled;
} ma_plugin_settings;

EXPORT_API int wakeup_manager_initialize(void);

EXPORT_API int wakeup_manager_deinitialize(void);

EXPORT_API int wakeup_manager_get_settings(ma_plugin_settings **settings, size_t *struct_size);

EXPORT_API int wakeup_manager_add_assistant_wakeup_word(const char* appid, const char* wakeup_word, const char* language);

EXPORT_API int wakeup_manager_add_assistant_language(const char* appid, const char* language);

EXPORT_API int wakeup_manager_set_assistant_wakeup_engine(const char* appid, const char *engine);

EXPORT_API int wakeup_manager_set_default_assistant(const char* appid);

EXPORT_API int wakeup_manager_get_default_assistant(const char** appid);

EXPORT_API int wakeup_manager_set_language(const char* language);

EXPORT_API int wakeup_manager_activate(void);

EXPORT_API int wakeup_manager_deactivate(void);

EXPORT_API int wakeup_manager_update_voice_feedback_state(const char *appid, int state);

EXPORT_API int wakeup_manager_send_assistant_specific_command(const char* appid, const char* command);

EXPORT_API int wakeup_manager_set_background_volume(const char *appid, double ratio);

EXPORT_API int wakeup_manager_update_recognition_result(const char *appid, int result);

EXPORT_API int wakeup_manager_process_event(int event, void* data, int len);

EXPORT_API int wakeup_manager_start_streaming_utterance_data(void);

EXPORT_API int wakeup_manager_stop_streaming_utterance_data(void);

EXPORT_API int wakeup_manager_start_streaming_previous_utterance_data(void);

EXPORT_API int wakeup_manager_stop_streaming_previous_utterance_data(void);

EXPORT_API int wakeup_manager_start_streaming_follow_up_data(void);

EXPORT_API int wakeup_manager_stop_streaming_follow_up_data(void);

EXPORT_API int wakeup_manager_get_audio_format(int *rate, int *channel, int *audio_type);

EXPORT_API int wakeup_manager_get_audio_source_type(char** type);

EXPORT_API int wakeup_manager_set_wake_word_audio_require_flag(bool require);

EXPORT_API int wakeup_manager_set_assistant_language(const char* appid, const char *language);

EXPORT_API int wakeup_manager_set_wakeup_event_callback(wakeup_service_wakeup_event_cb callback, void* user_data);

EXPORT_API int wakeup_manager_set_utterance_streaming_callback(wakeup_service_speech_streaming_cb callback, void* user_data);

EXPORT_API int wakeup_manager_set_previous_utterance_streaming_callback(wakeup_service_speech_streaming_cb callback, void* user_data);

EXPORT_API int wakeup_manager_set_follow_up_streaming_callback(wakeup_service_speech_streaming_cb callback, void* user_data);

EXPORT_API int wakeup_manager_set_speech_status_callback(wakeup_service_speech_status_cb callback, void* user_data);

EXPORT_API int wakeup_manager_set_setting_changed_callback(wakeup_service_setting_changed_cb callback, void* user_data);

EXPORT_API int wakeup_manager_set_error_callback(wakeup_service_error_cb callback, void* user_data);

EXPORT_API int wakeup_manager_set_streaming_section_changed_callback(wakeup_service_streaming_section_changed_cb callback, void* user_data);

/* Internal API declarations for dependency modules */

int wakeup_manager_feed_audio_data(wakeup_speech_streaming_event_e event, void* buffer, int len);

#ifdef __cplusplus
}
#endif

#endif /* _WAKEUP_MANAGER_WRAPPER_H_ */
