/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <tizen.h>
#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <vconf.h>

#include <stdexcept>

#include "multi_assistant_main.h"
#include "dependency_resolver.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif
#define LOG_TAG "dependency-resolver"

static void *g_handle = NULL;
static mas_dependency g_mas_dependency = { NULL, };

int dependency_resolver_initialize(mas_proxy_interface interface)
{
	MAS_LOGD("[Enter]");

	const int FILEPATH_LEN = 512;
	char filepath[FILEPATH_LEN] = {'\0', };
	char *vconf_str = vconf_get_str(MAS_DEPENDENCY_MODULE_PATH);
	if (vconf_str) {
		snprintf(filepath, FILEPATH_LEN - 1, "%s", vconf_str);
		free(vconf_str);
	} else {
		const char *default_path = MAS_DEPENDENCY_DEFAULT_PATH;
		snprintf(filepath, FILEPATH_LEN - 1, "%s/%s", default_path, MAS_DEPENDENCY_DEFAULT_FILENAME);
	}
	filepath[FILEPATH_LEN - 1] = '\0';

	char *error;
	g_handle = NULL;
	g_handle = dlopen(filepath, RTLD_LAZY);
	if (NULL != (error = dlerror())) {
		MAS_LOGE("[ERROR] Fail to dlopen(%s), error(%s)", filepath, error);
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	g_mas_dependency.initialize =
		(mas_dependency_initialize)dlsym(g_handle,
		MAS_DEPENDENCY_FUNC_INITIALIZE);
	g_mas_dependency.deinitialize =
		(mas_dependency_deinitialize)dlsym(g_handle,
		MAS_DEPENDENCY_FUNC_DEINITIALIZE);
	g_mas_dependency.set_error_callback =
		(mas_dependency_set_error_callback)dlsym(g_handle,
		MAS_DEPENDENCY_FUNC_SET_ERROR_CALLBACK);
	g_mas_dependency.start_recording =
		(mas_dependency_start_recording)dlsym(g_handle,
		MAS_DEPENDENCY_FUNC_START_RECORDING);
	g_mas_dependency.stop_recording =
		(mas_dependency_stop_recording)dlsym(g_handle,
		MAS_DEPENDENCY_FUNC_STOP_RECORDING);
	g_mas_dependency.set_recording_session =
		(mas_dependency_set_recording_session)dlsym(g_handle,
		MAS_DEPENDENCY_FUNC_SET_RECORDING_SESSION);
	g_mas_dependency.set_background_volume =
		(mas_dependency_set_background_volume)dlsym(g_handle,
		MAS_DEPENDENCY_FUNC_SET_BACKGROUND_VOLUME);
	g_mas_dependency.get_audio_format =
		(mas_dependency_get_audio_format)dlsym(g_handle,
		MAS_DEPENDENCY_FUNC_GET_AUDIO_FORMAT);
	g_mas_dependency.get_audio_source_type =
		(mas_dependency_get_audio_source_type)dlsym(g_handle,
		MAS_DEPENDENCY_FUNC_GET_AUDIO_SOURCE_TYPE);

	int ret = -1;
	int dependency_version = 0;
	if (NULL != g_handle) {
		mas_dependency_initialize func = g_mas_dependency.initialize;

		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MAS_DEPENDENCY_FUNC_INITIALIZE);
		} else {
			try {
				ret = func(interface, &dependency_version);
			} catch (const std::exception& e) {
				MAS_LOGE("[ERROR] %s of dependency module threw exception : %s",
					MAS_DEPENDENCY_FUNC_INITIALIZE, e.what());
			}
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to initialize, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	MAS_LOGD("g_handle : %p, dependency_version %d", g_handle, dependency_version);
	return ret;
}

int dependency_resolver_deinitialize(void)
{
	MAS_LOGD("g_handle : %p", g_handle);
	int ret = -1;
	if (NULL != g_handle) {
		mas_dependency_deinitialize func = g_mas_dependency.deinitialize;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MAS_DEPENDENCY_FUNC_DEINITIALIZE);
		} else {
			try {
				ret = func();
			} catch (const std::exception& e) {
				MAS_LOGE("[ERROR] %s of dependency module threw exception : %s",
					MAS_DEPENDENCY_FUNC_DEINITIALIZE, e.what());
			}
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to deinitialize, ret(%d)", ret);
			}
		}

		dlclose(g_handle);
		g_handle = NULL;
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}

	return ret;
}

int dependency_resolver_set_error_callback(mas_dependency_error_cb callback, void* user_data)
{
	int ret = -1;
	if (NULL != g_handle) {
		mas_dependency_set_error_callback func = g_mas_dependency.set_error_callback;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MAS_DEPENDENCY_FUNC_SET_ERROR_CALLBACK);
		} else {
			try {
				ret = func(callback, user_data);
			} catch (const std::exception& e) {
				MAS_LOGE("[ERROR] %s of dependency module threw exception : %s",
					MAS_DEPENDENCY_FUNC_DEINITIALIZE, e.what());
			}
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set error callback(%p, %p), ret(%d)", callback, user_data, ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int dependency_resolver_start_recording(void)
{
	int ret = -1;
	MAS_LOGD("g_handle : %p", g_handle);
	if (NULL != g_handle) {
		mas_dependency_start_recording func = g_mas_dependency.start_recording;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MAS_DEPENDENCY_FUNC_START_RECORDING);
		} else {
			try {
				ret = func();
			} catch (const std::exception& e) {
				MAS_LOGE("[ERROR] %s of dependency module threw exception : %s",
					MAS_DEPENDENCY_FUNC_DEINITIALIZE, e.what());
			}
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to start recording, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}

	return ret;
}

int dependency_resolver_stop_recording(void)
{
	int ret = -1;
	if (NULL != g_handle) {
		mas_dependency_stop_recording func = g_mas_dependency.stop_recording;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MAS_DEPENDENCY_FUNC_STOP_RECORDING);
		} else {
			try {
				ret = func();
			} catch (const std::exception& e) {
				MAS_LOGE("[ERROR] %s of dependency module threw exception : %s",
					MAS_DEPENDENCY_FUNC_DEINITIALIZE, e.what());
			}
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to stop recording, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}

	return ret;
}

int dependency_resolver_set_recording_session(unsigned int session)
{
	int ret = -1;
	MAS_LOGD("g_handle : %p", g_handle);
	if (NULL != g_handle) {
		mas_dependency_set_recording_session func = g_mas_dependency.set_recording_session;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MAS_DEPENDENCY_FUNC_SET_RECORDING_SESSION);
		} else {
			try {
				ret = func(session);
			} catch (const std::exception& e) {
				MAS_LOGE("[ERROR] %s of dependency module threw exception : %s",
					MAS_DEPENDENCY_FUNC_DEINITIALIZE, e.what());
			}
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set recording session, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}

	return ret;
}

int dependency_resolver_set_background_volume(double ratio)
{
	int ret = -1;
	if (NULL != g_handle) {
		mas_dependency_set_background_volume func = g_mas_dependency.set_background_volume;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MAS_DEPENDENCY_FUNC_SET_BACKGROUND_VOLUME);
		} else {
			try {
				ret = func(ratio);
			} catch (const std::exception& e) {
				MAS_LOGE("[ERROR] %s of dependency module threw exception : %s",
					MAS_DEPENDENCY_FUNC_DEINITIALIZE, e.what());
			}
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set background volume to %f, ret(%d)", ratio, ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}

	return ret;
}

int dependency_resolver_get_audio_format(int* rate, int* channel, int* audio_type)
{
	int ret = -1;
	if (NULL != g_handle) {
		mas_dependency_get_audio_format func = g_mas_dependency.get_audio_format;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MAS_DEPENDENCY_FUNC_GET_AUDIO_FORMAT);
		} else {
			try {
				ret = func(rate, channel, audio_type);
			} catch (const std::exception& e) {
				MAS_LOGE("[ERROR] %s of dependency module threw exception : %s",
					MAS_DEPENDENCY_FUNC_DEINITIALIZE, e.what());
			}
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to get audio format, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}

	return ret;
}

int dependency_resolver_get_audio_source_type(char** type)
{
	int ret = -1;
	if (NULL != g_handle) {
		mas_dependency_get_audio_source_type func = g_mas_dependency.get_audio_source_type;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MAS_DEPENDENCY_FUNC_GET_AUDIO_SOURCE_TYPE);
		} else {
			try {
				ret = func(type);
			} catch (const std::exception& e) {
				MAS_LOGE("[ERROR] %s of dependency module threw exception : %s",
					MAS_DEPENDENCY_FUNC_DEINITIALIZE, e.what());
			}
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to get audio source type, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}

	return ret;
}
