#include "heap_tracer.h"

#include <stdlib.h>
#include <stdarg.h>

#include <dlog.h>

#ifdef LOG_TAG
#undef LOG_TAG
#endif
#define LOG_TAG "ma_heap_tracer"

static HeapTracer _tracer;

const bool RECORD_UNMANAGED_CHUNKS = true;
const int MAX_UNMANAGED_RECORDS = 128;
const bool HEAP_TRACER_LOG_ENABLED = false
;

#define DESCRIPTION_LENGTH 255
//#define USE_HEAP_TRACER
//#define RECORD_ALL_HISTORY

void* vm_calloc(size_t nmemb, size_t size, const char *fmt, ...)
{
    void *ptr = calloc(nmemb, size);
#ifdef USE_HEAP_TRACER
    va_list argptr;
    va_start(argptr, fmt);

    char description[DESCRIPTION_LENGTH];
    vsnprintf(description, sizeof(description), fmt, argptr);
    _tracer.Insert(ptr, description, nmemb * size);

    va_end(argptr);
#endif
    return ptr;
}

void* vm_malloc(size_t size, const char *fmt, ...)
{
    void *ptr = malloc(size);
#ifdef USE_HEAP_TRACER
    va_list argptr;
    va_start(argptr, fmt);

    char description[DESCRIPTION_LENGTH];
    vsnprintf(description, sizeof(description), fmt, argptr);
    _tracer.Insert(ptr, description, size);

    va_end(argptr);
#endif
    return ptr;
}

void vm_free(void *ptr, const char *fmt, ...)
{
    if (ptr) {
#ifdef USE_HEAP_TRACER
        va_list argptr;
        va_start(argptr, fmt);

        char description[DESCRIPTION_LENGTH];
        vsnprintf(description, sizeof(description), fmt, argptr);
        if (false == _tracer.Delete(ptr, description)) {
            _tracer.Find(ptr);
        }
        va_end(argptr);
#endif
        free(ptr);
    } else {
        LOGD("Trying to free a null pointer");
    }
}

char* vm_strdup(const char *s, const char *fmt, ...)
{
    char *ptr = nullptr;
    if (s) {
        ptr = strdup(s);
#ifdef USE_HEAP_TRACER
        va_list argptr;
        va_start(argptr, fmt);

        char description[DESCRIPTION_LENGTH];
        vsnprintf(description, sizeof(description), fmt, argptr);
        _tracer.Insert(ptr, description, strlen(s) + 1);

        va_end(argptr);
#endif
    } else {
        LOGD("Trying to duplicate a null pointer");
    }
    return ptr;
}

void *vm_mark_unmanaged(void *ptr, const char *fmt, ...)
{
#ifdef USE_HEAP_TRACER
    va_list argptr;
    va_start(argptr, fmt);

    char description[DESCRIPTION_LENGTH];
    vsnprintf(description, sizeof(description), fmt, argptr);
    _tracer.MarkUnmanaged(ptr, description);

    va_end(argptr);
#endif
    return ptr;
}

HeapTracer::HeapTracer()
{
    LOGD("HeapTracer()");
}

HeapTracer::~HeapTracer()
{
    LOGD("~HeapTracer()");

    Trace();
}

bool HeapTracer::Insert(void *ptr, std::string description, size_t size)
{
    if (HEAP_TRACER_LOG_ENABLED) {
        LOGD("Heap allocated %p [size %zu] : %s", ptr, size, description.c_str());
    }

    AllocationEntry entry;
    entry.ptr = ptr;
    entry.description = description;
    entry.size = size;

    std::lock_guard<std::mutex> lock(mManagedTableMutex);
    mManagedTable[ptr] = entry;

#ifdef RECORD_ALL_HISTORY
    EntryHistory history;
    history.entry = entry;
    history.operation = EntryOperation::INSERT;
    clock_gettime(CLOCK_REALTIME, &(history.ts));
    mEntryHistory.push_back(history);
#endif
    return true;
}

bool HeapTracer::Delete(void *ptr, std::string description)
{
    if (HEAP_TRACER_LOG_ENABLED) {
        LOGD("Heap deallocated %p : %s", ptr, description.c_str());
    }

    std::lock_guard<std::mutex> lock(mManagedTableMutex);
    if (mManagedTable.find(ptr) != mManagedTable.end()) {
#ifdef RECORD_ALL_HISTORY
        EntryHistory history;
        history.entry = mManagedTable[ptr];
        history.entry.description = description;
        history.operation = EntryOperation::DELETE;
        clock_gettime(CLOCK_REALTIME, &(history.ts));
        mEntryHistory.push_back(history);
#endif
        mManagedTable.erase(ptr);
    } else {
        LOGE("Heap Trace Error : Trying to delete an entry that does not exist : %p, %s",
            ptr, description.c_str());
        return false;
    }
    return true;
}

std::string HeapTracer::Find(void *ptr)
{
    std::lock_guard<std::mutex> lock(mManagedTableMutex);
#ifdef RECORD_ALL_HISTORY
    for (auto entry : mEntryHistory) {
        if (entry.entry.ptr == ptr) {
            LOGD("Heap history record for %p : %s, %lld.%.9ld, %s",
                ptr,
                (entry.operation == EntryOperation::INSERT ? "INSERT" : "DELETE"),
                (long long)(entry.ts.tv_sec), entry.ts.tv_nsec,
                entry.entry.description.c_str());
        }
    }
#endif
    if (mManagedTable.find(ptr) != mManagedTable.end()) {
        return mManagedTable[ptr].description;
    }
    return std::string("Not found");
}

void HeapTracer::MarkUnmanaged(void *ptr, std::string description)
{
    LOGD("Marking as unmanaged : %p %s", ptr, description.c_str());

    std::lock_guard<std::mutex> lock(mManagedTableMutex);
    auto iter = mManagedTable.find(ptr);
    if (iter != mManagedTable.end()) {
        if (RECORD_UNMANAGED_CHUNKS) {
            std::lock_guard<std::mutex> lock(mUnmanagedRecordsMutex);
            if (mUnmanagedRecordsSize >= MAX_UNMANAGED_RECORDS) {
                mUnmanagedRecords.pop_front();
                mUnmanagedRecordsSize--;
            }
            mUnmanagedRecords.push_back(iter->second);
            mUnmanagedRecordsSize++;
        }
        mManagedTable.erase(ptr);
    } else {
        LOGE("Heap Trace Error : Trying to mark an entry that does not exist as unmanaged : %p", ptr);
    }
}

bool HeapTracer::Empty()
{
    std::lock_guard<std::mutex> lock(mManagedTableMutex);
    return mManagedTable.empty();
}

void HeapTracer::Trace()
{
    LOGD("Heap Trace Start");

    std::lock_guard<std::mutex> lock(mManagedTableMutex);
    for (auto item : mManagedTable) {
        LOGD("No dealloc information exists for %p [size %d] : %s", item.first, item.second.size, item.second.description.c_str());
    }
    LOGD("Heap Trace End");
}
