#include "wakeup_audio_manager.h"
#include "wakeup_manager_main.h"
#include "wakeup_interfaces.h"
#include "dependency_resolver.h"
#include "heap_tracer.h"

#include <algorithm>

#include <Ecore.h>

namespace multiassistant
{
namespace wakeup
{

/* Need to check whether this value needs to be configurable */
static int g_speech_pcm_wait_count = 800;

static long get_current_milliseconds_after_epoch()
{
	auto now = chrono::system_clock::now();
	auto now_ms = chrono::time_point_cast<chrono::milliseconds>(now);
	/* number of milliseconds since the epoch of system_clock */
	auto value = now_ms.time_since_epoch();

	return value.count();
}

CAudioManager::CAudioManager()
{
}

CAudioManager::~CAudioManager()
{
}

CAudioManager::CAudioManager(IAudioEventObserver *observer) : CAudioManager()
{
	subscribe(observer);
}

void recording_focus_state_watch_cb(int id, sound_stream_focus_mask_e focus_mask,
	sound_stream_focus_state_e focus_state, sound_stream_focus_change_reason_e reason,
	const char *extra_info, void *user_data)
{
	MWR_LOGD("[Recorder] focus_mask : %d, focus_state : %d, reason : %d, [%s]",
		focus_mask, focus_state, reason, extra_info);

	if (nullptr == user_data) return;

	/* The API description states that calling sound_manager_get_current_recording_focus()
		function inside sound_stream_focus_state_watch_cb() function is prohibited */
	ecore_main_loop_thread_safe_call_async(
		[](void* data) {
			CAudioManager *manager = static_cast<CAudioManager*>(data);
			if (manager) {
				manager->sound_focus_changed();
			}
		}, user_data);
}

int CAudioManager::initialize(void)
{
	sound_manager_add_focus_state_watch_cb(SOUND_STREAM_FOCUS_FOR_RECORDING,
		recording_focus_state_watch_cb, this, &mSoundFocusWatchId);
	return 0;
}

int CAudioManager::deinitialize(void)
{
	clear_audio_data();
	if (mStreamingThread.joinable()) {
		MWR_LOGD("mStreamingThread is joinable, trying join()");
		mStopStreamingThread.store(true);
		try {
			mStreamingThread.join();
		} catch (std::exception &e) {
			MWR_LOGE("Exception thrown : %s", e.what());
		}
	}
	mStopStreamingThread.store(false);

	sound_manager_remove_focus_state_watch_cb(mSoundFocusWatchId);

	return 0;
}

void CAudioManager::sound_focus_changed()
{
	sound_stream_focus_change_reason_e acquired_by;
	int sound_behavior;
	char* extra_info = NULL;
	int focus = sound_manager_get_current_recording_focus(&acquired_by, &sound_behavior, &extra_info);
	MWR_LOGD("[Recorder] sound focus has changed : %d %d %d %s", focus,
		(SOUND_MANAGER_ERROR_NO_DATA != focus ? acquired_by : -1),
		(SOUND_MANAGER_ERROR_NO_DATA != focus ? sound_behavior : -1),
		(SOUND_MANAGER_ERROR_NO_DATA != focus ? extra_info : ""));
	if (SOUND_MANAGER_ERROR_NO_DATA == focus) {
		if (mRecordingRequired && !mIsRecording) {
			MWR_LOGD("[Recorder] Currently no other process has acquired sound focus, start recording");
			start_recording(false);
		}
	} else {
		if (mIsRecording) {
			MWR_LOGW("[Recorder] Sound focus acquired by other process, stop recording");
			stop_recording(false);
		}
	}
	if (extra_info) {
		free(extra_info);
		extra_info = NULL;
	}
}

void CAudioManager::subscribe(IAudioEventObserver *observer)
{
	mObservers.push_back(observer);
}

void CAudioManager::unsubscribe(IAudioEventObserver *observer)
{
	auto iter = find(mObservers.begin(), mObservers.end(), observer);
	if (iter != mObservers.end()) {
		mObservers.erase(iter);
	}
}

void CAudioManager::stop_recording(bool proactive)
{
	dependency_resolver_stop_recording();
	if (proactive) {
		mRecordingRequired = false;
	}
	mIsRecording = false;
}

void CAudioManager::start_recording(bool proactive)
{
	if (mIsRecording) {
		stop_recording(false);
	}
	if (proactive) {
		mRecordingRequired = true;
	}

	sound_stream_focus_change_reason_e acquired_by;
	int sound_behavior;
	char* extra_info = NULL;
	if (SOUND_MANAGER_ERROR_NO_DATA == sound_manager_get_current_recording_focus(&acquired_by, &sound_behavior, &extra_info)) {
		MWR_LOGD("[Recorder] Currently no other process has acquired sound focus, start recording");
		dependency_resolver_start_recording();
		mIsRecording = true;
	} else {
		MWR_LOGW("[Recorder] Currently sound focus is acquired by other process, skip recording");
	}
	if (extra_info) {
		free(extra_info);
		extra_info = NULL;
	}
}

void CAudioManager::set_recording_session(recording_session session)
{
	dependency_resolver_set_recording_session((unsigned int)session);
}

/* Need to consider adapting conventional producer-consumer model */
void CAudioManager::streaming_previous_audio_data_thread_func()
{
	MWR_LOGD("[ENTER]");

	unique_lock<mutex> lock(mMutex, defer_lock);

	int ret = -1;
	int cnt = 0;

	/* get feedback data */
	size_t audio_data_size = 0;
	lock.lock();
	audio_data_size = mPreviousAudioData.size();
	for (int index = 0; index < audio_data_size; index++) {
		wakeup_speech_data& speech_data = mPreviousAudioData.at(index).data;
		for (const auto& observer : mObservers) {
			if (observer) {
				if (!observer->on_streaming_audio_data(
					speech_data.event, speech_data.buffer, speech_data.len)) {
					LOGE("[Recorder WARNING] One of the observer returned false");
				}
			}

			if (WAKEUP_SPEECH_STREAMING_EVENT_FINISH == speech_data.event) {
				MWR_LOGI("[INFO] Finish to send previous speech data");
				return;
			}
		}

		if (mStopStreamingThread.load()) {
			MWR_LOGI("[INFO] Stop Streaming Requested, returning");
			return;
		}
	}
	lock.unlock();

	MWR_LOGD("[EXIT]");
}

void CAudioManager::streaming_audio_data_thread_func(long start_time)
{
	MWR_LOGD("[ENTER]");

	unique_lock<mutex> lock(mMutex, defer_lock);

	lock.lock();
	auto lead = mAudioData.begin();
	auto iter = lead;
	while (lead != mAudioData.end() && lead->time < start_time) {
		iter = lead;
		advance(lead, 1);
	}
	MWR_LOGD("data_count : %zu", mAudioData.size());
	lock.unlock();

	while (!(mStopStreamingThread.load())) {
		int ret = -1;
		int cnt = 0;

		/* get feedback data */
		lock.lock();
		auto end = mAudioData.end();
		lock.unlock();
		if (lead == end) {
			/* empty queue */
			MWR_LOGD("[DEBUG] No feedback data. Waiting mode : %d", ret);

			/* waiting */
			while (!(mStopStreamingThread.load())) {
				this_thread::sleep_for(chrono::milliseconds(10));
				lock.lock();
				end = mAudioData.end();
				auto begin = mAudioData.begin();
				lock.unlock();
				if (iter == end) {
					iter = begin;
				}
				lead = iter;
				if (lead != end) {
					advance(lead, 1);
				}
				if (lead != end) {
					MWR_LOGI("[INFO] Resume thread");
					break;
				}
				if (g_speech_pcm_wait_count < cnt) {
					MWR_LOGE("[ERROR] Wrong request, there's no pcm data");
					for (const auto& observer : mObservers) {
						if (observer) {
							if (!observer->on_streaming_audio_data(
								WAKEUP_SPEECH_STREAMING_EVENT_FAIL, NULL, 0)) {
								LOGE("[Recorder WARNING] One of the observer returned false");
							}
						}
					}
					MWR_LOGD("[EXIT]");
					return;
				}
				cnt++;
			}
			MWR_LOGI("[INFO] Finish to wait for new feedback data come");

			/* resume feedback thread */
			continue;
		}

		iter = lead;

		/* FIXME : Extracted audio data here should be used as previous audio data*/

		wakeup_speech_data& speech_data = iter->data;
		for (const auto& observer : mObservers) {
			if (observer) {
				if (!observer->on_streaming_audio_data(
					speech_data.event, speech_data.buffer, speech_data.len)) {
					LOGE("[Recorder WARNING] One of the observer returned false");
				}
			}
		}

		if (WAKEUP_SPEECH_STREAMING_EVENT_FINISH == speech_data.event) {
			MWR_LOGI("[INFO] Finish to get and send speech data");
			break;
		}

		advance(lead, 1);
	}
	MWR_LOGD("[EXIT]");
}

void CAudioManager::add_audio_data(wakeup_speech_data& data, long time)
{
	long delta = mAudioRecordingDurationMilliseconds;

	notify_audio_data_recording(time, data.buffer, data.len);

	wakeup_speech_data_with_time data_with_time;
	data_with_time.data = data;
	data_with_time.time = time;

	lock_guard<mutex> lock(mMutex);

	/* Pop items only when the streaming is not activated */
	while(false == mAudioData.empty() && mAudioData.front().time < time - delta) {
		const auto &front = mAudioData.front();
		if (front.data.buffer) {
			vm_free_simple(front.data.buffer);
		}
		mAudioData.pop_front();
	}
	mAudioData.push_back(data_with_time);
}

void CAudioManager::feed_audio_data(wakeup_speech_streaming_event_e event, void* buffer, int len)
{
	if (NULL == buffer || 0 == len) return;

	wakeup_speech_data speech_data;
	speech_data.buffer = vm_malloc_simple(len);
	if (speech_data.buffer) {
		long time = get_current_milliseconds_after_epoch();

		speech_data.event = event;
		speech_data.len = len;
		memcpy(speech_data.buffer, buffer, len);
		add_audio_data(speech_data, time);
	}
}

void CAudioManager::finalize_audio_data()
{
	unsigned char final_buffer[2] = {'\0', };
	wakeup_speech_data speech_data;
	speech_data.event = WAKEUP_SPEECH_STREAMING_EVENT_FINISH;
	speech_data.len = sizeof(final_buffer);
	speech_data.buffer = vm_malloc_simple(speech_data.len);
	if (speech_data.buffer) {
		long time = get_current_milliseconds_after_epoch();

		memcpy(speech_data.buffer, final_buffer, speech_data.len);
		add_audio_data(speech_data, time);
	}
}

void CAudioManager::clear_audio_data()
{
	lock_guard<mutex> lock(mMutex);
	while(!mAudioData.empty()) {
		const auto &front = mAudioData.front();
		if (front.data.buffer) {
			vm_free_simple(front.data.buffer);
		}
		mAudioData.pop_front();
	}
	mAudioData.clear();
}

void CAudioManager::notify_audio_data_recording(long time, void* data, int len)
{
	for (const auto& observer : mObservers) {
		if (observer) {
			if (!observer->on_recording_audio_data(time, data, len)) {
				LOGE("[Recorder WARNING] One of the observer returned false");
				return;
			}
		}
	}
}

void CAudioManager::start_streaming_current_utterance_data(long start_time)
{
	if (mStreamingThread.joinable()) {
		MWR_LOGE("ERROR : mStreamingThread is joinable, will not start a new thread");
		return;
	}
	lock_guard<mutex> lock(mMutex);
	mStreamingThread = thread(&CAudioManager::streaming_audio_data_thread_func, this, start_time);
}

void CAudioManager::stop_streaming_current_utterance_data()
{
	if (mStreamingThread.joinable()) {
		MWR_LOGD("mStreamingThread is joinable, trying join()");
		mStopStreamingThread.store(true);
		try {
			mStreamingThread.join();
		} catch (std::exception &e) {
			MWR_LOGE("Exception thrown : %s", e.what());
		}
	}
	mStopStreamingThread.store(false);

	/* FIXME : Need to move all the speech data to previous speech data for later use */
}

void CAudioManager::start_streaming_previous_utterance_data()
{
	if (mStreamingPreviousThread.joinable()) {
		MWR_LOGE("ERROR : mStreamingPreviousThread is joinable, will not start a new thread");
		return;
	}
	lock_guard<mutex> lock(mMutex);
	mStreamingPreviousThread = thread(&CAudioManager::streaming_previous_audio_data_thread_func, this);
}

void CAudioManager::stop_streaming_previous_utterance_data()
{
	if (mStreamingPreviousThread.joinable()) {
		MWR_LOGD("mStreamingPreviousThread is joinable, trying join()");
		mStopStreamingPreviousThread.store(true);
		try {
			mStreamingPreviousThread.join();
		} catch (std::exception &e) {
			MWR_LOGE("Exception thrown : %s", e.what());
		}
	}
	mStopStreamingThread.store(false);
}

void CAudioManager::start_streaming_follow_up_data()
{
	if (mStreamingThread.joinable()) {
		MWR_LOGE("ERROR : mStreamingThread is joinable, will not start a new thread");
		return;
	}

	mStreamingThread = thread(&CAudioManager::streaming_audio_data_thread_func, this, 0);
}

void CAudioManager::stop_streaming_follow_up_data()
{
	if (mStreamingThread.joinable()) {
		MWR_LOGD("mStreamingThread is joinable, trying join()");
		mStopStreamingThread.store(true);
		mStreamingThread.join();
	}
	mStopStreamingThread.store(false);
}

void CAudioManager::set_background_volume(double ratio)
{
	dependency_resolver_set_background_volume(ratio);
}

} // wakeup
} // multiassistant
