#include "wakeup_engine_manager.h"
#include "wakeup_manager_main.h"

#include <dlfcn.h>
#include <algorithm>
#include <pkgmgr-info.h>
#include <deque>

namespace multiassistant
{
namespace wakeup
{

/* Need to check whether this value needs to be configurable */
static int g_speech_pcm_wait_count = 400;

/* Utility function for checking if an element exists in a container */
template<class C, class T>
static auto contains(const C& v, const T& x) -> decltype(end(v), true)
{
	return end(v) != find(begin(v), end(v), x);
}

CWakeupEngineManager::CWakeupEngineManager()
{
}

CWakeupEngineManager::~CWakeupEngineManager()
{
	MWR_LOGI("Wakeup Engine Manager is now being destroyed");
}

CWakeupEngineManager::CWakeupEngineManager(IEngineEventObserver *observer) : CWakeupEngineManager()
{
	subscribe(observer);
}

void CWakeupEngineManager::initialize()
{
	DIR* dp = opendir(MA_WAKEUP_ENGINE_PATH);
	if (nullptr == dp) {
		MWR_LOGD("Failed opening directory : %s", (const char*)MA_WAKEUP_ENGINE_PATH);
	} else {
		struct dirent *dirp = nullptr;
		char dirpath[_POSIX_PATH_MAX];
		do {
			dirp = readdir(dp);

			if (nullptr != dirp) {
				const string current_directory{"."};
				const string parent_directory{".."};
				if (0 == current_directory.compare(dirp->d_name) ||
					0 == parent_directory.compare(dirp->d_name))
					continue;

				if (DT_DIR != dirp->d_type) /* If not a directory */
					continue;

				int dirpath_len = strlen(MA_WAKEUP_ENGINE_PATH) + strlen(dirp->d_name) + 1;
				if (dirpath_len >= _POSIX_PATH_MAX) {
					MWR_LOGD("File path is too long : %s", dirp->d_name);
					closedir(dp);
					return;
				}

				memset(dirpath, '\0', _POSIX_PATH_MAX);
				snprintf(dirpath, _POSIX_PATH_MAX, "%s/%s",
					(const char*)(MA_WAKEUP_ENGINE_PATH), dirp->d_name);

				add_engine_directory(string{dirp->d_name}, dirpath);
			}
		} while (nullptr != dirp);

		closedir(dp);
	}
}

void CWakeupEngineManager::deinitialize()
{
	MWR_LOGI("[START]");
	if (mStreamingThread.joinable()) {
		MWR_LOGD("mStreamingThread is joinable, trying join()");
		mStopStreamingThread.store(true);
		mStreamingThread.join();
	}
	mStopStreamingThread.store(false);

	for (auto& info : mEngineInfo) {
		try {
			if (info.interface.set_wakeup_event_callback) {
				info.interface.set_wakeup_event_callback(nullptr, nullptr);
			}
			if (info.interface.set_speech_status_callback) {
				info.interface.set_speech_status_callback(nullptr, nullptr);
			}
			if (info.interface.set_error_callback) {
				info.interface.set_error_callback(nullptr, nullptr);
			}
			if (info.interface.set_audio_data_require_status_callback) {
				info.interface.set_audio_data_require_status_callback(nullptr, nullptr);
			}
			MWR_LOGI("Trying to deinitialize engine : %s", info.engine_name.c_str());
			if (info.interface.deinitialize) {
				int ret = info.interface.deinitialize();
				MWR_LOGI("Deinitialization [%s] returned %d", info.engine_name.c_str(), ret);
			}
		} catch (const std::exception& e) {
			MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
				info.engine_name.c_str(), e.what());
		}
		if (info.engine_handle) {
			int ret = dlclose(info.engine_handle);
			MWR_LOGI("Closing [%s] returned %d, [%s]", info.engine_name.c_str(), ret,
				((0 == ret) ? "" : dlerror()));
			info.engine_handle = nullptr;
		}
	}
	mSelectedEngine = nullptr;
	mEngineInfo.clear();
	MWR_LOGI("[END]");
}

void CWakeupEngineManager::subscribe(IEngineEventObserver *observer)
{
	mObservers.push_back(observer);
	MWR_LOGD("Added Observer : %p %zu", observer, mObservers.size());
}

void CWakeupEngineManager::unsubscribe(IEngineEventObserver *observer)
{
	auto iter = find(mObservers.begin(), mObservers.end(), observer);
	if (iter != mObservers.end()) {
		mObservers.erase(iter);
	}
}

bool CWakeupEngineManager::get_audio_data_required()
{
	return mAudioDataRequired;
}

void CWakeupEngineManager::set_selected_wakeup_info(wakeup_event_info wakeup_info)
{
	mSelectedEngine = nullptr;
	for (const auto& info : mEngineInfo) {
		string appid = string{wakeup_info.wakeup_appid};
		bool found = contains(info.assistant_list, appid);

		if (found) {
			mSelectedEngine = &info;
			MWR_LOGD("Selected : %s", info.engine_name.c_str());
		}
	}
}

bool CWakeupEngineManager::set_language(string language)
{
	for (const auto& info : mEngineInfo) {
		if (info.interface.set_language) {
			try {
				info.interface.set_language(language.c_str());
			} catch (const std::exception& e) {
				MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
					info.engine_name.c_str(), e.what());
			}
		}
	}
	return true;
}

bool CWakeupEngineManager::set_assistant_language(string appid, string language)
{
	for (auto& info : mEngineInfo) {
		const auto& iter = find_if(info.assistant_list.begin(), info.assistant_list.end(),
			[appid](const string& assistant) {
				return (0 == assistant.compare(appid));
			});

		/* If the appid is in the assistant list */
		if (info.assistant_list.end() != iter) {
			try {
				int ret = info.interface.set_language(language.c_str());
				MWR_LOGD("set_language returned %d : %s %s %s",
					ret, appid.c_str(), info.engine_name.c_str(), language.c_str());
			} catch (const std::exception& e) {
				MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
					info.engine_name.c_str(), e.what());
			}
		}
	}
	return true;
}

void CWakeupEngineManager::set_assistant_activated(string appid, bool activated)
{
	MWR_LOGD("[ENTER] : %s %d", appid.c_str(), activated);
	for (auto& info : mEngineInfo) {
		const auto& iter = find_if(info.assistant_list.begin(), info.assistant_list.end(),
			[appid](const string& assistant) {
				return (0 == assistant.compare(appid));
			});

		/* If the appid is in the assistant list */
		if (info.assistant_list.end() != iter) {
			bool previously_activated = info.activated;
			if (activated) {
				info.activated_assistants.insert(appid);
			} else {
				info.activated_assistants.erase(appid);
			}
			info.activated = (info.activated_assistants.size() > 0);
			if (previously_activated != info.activated) {
				if (info.activated) {
					try {
						info.interface.activate();
					} catch (const std::exception& e) {
						MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
							info.engine_name.c_str(), e.what());
					}
				} else {
					try {
						info.interface.deactivate();
					} catch (const std::exception& e) {
						MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
							info.engine_name.c_str(), e.what());
					}
				}
				/* Activated status changed, need to update audio_data_require_status too */
				on_audio_data_require_status(info.engine_name, info.audio_data_require_status);
			}
		}
	}
}

void CWakeupEngineManager::set_wake_word_audio_require_flag(bool require)
{
	MWR_LOGD("[ENTER]");
	mWakeWordAudioRequired = require;
	for (const auto& info : mEngineInfo) {
		if (info.interface.set_wake_word_audio_require_flag) {
			try {
				info.interface.set_wake_word_audio_require_flag(require);
			} catch (const std::exception& e) {
				MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
					info.engine_name.c_str(), e.what());
			}
		}
	}
}

void CWakeupEngineManager::streaming_speech_data_thread_func()
{
	MWR_LOGD("[ENTER]");

	if (nullptr == mSelectedEngine)
		return;

	const wakeup_engine_interface *interface = &(mSelectedEngine->interface);

	if (NULL == interface ||
		NULL == interface->get_utterance_data ||
		NULL == interface->get_utterance_data_count)
		return;

	MWR_LOGD("data_count : %d", interface->get_utterance_data_count());

	wakeup_speech_data speech_data;
	int index = 0;
	bool finish_event_sent = false;

	if (mWakeWordAudioRequired &&
		NULL != interface->get_wake_word_data &&
		NULL != interface->get_wake_word_data_count) {
		for (const auto& observer : mObservers) {
			if (observer) {
				if (!observer->on_audio_streaming_data_section(MA_AUDIO_STREAMING_DATA_SECTION_WAKE_WORD)) {
					LOGE("[Recorder WARNING] One of the observer returned false");
				}
			}
		}
		int count = interface->get_wake_word_data_count();
		while (!(mStopStreamingThread.load()) && index < count) {
			int ret = interface->get_wake_word_data(index, &speech_data);
			if (0 == ret) {
				for (const auto& observer : mObservers) {
					if (observer) {
						if (!observer->on_streaming_audio_data(
							speech_data.event, speech_data.buffer, speech_data.len)) {
							LOGE("[Recorder WARNING] One of the observer returned false");
						}
					}
				}
			} else {
				break;
			}
			index++;
		}
		index = 0;
		for (const auto& observer : mObservers) {
			if (observer) {
				if (!observer->on_audio_streaming_data_section(MA_AUDIO_STREAMING_DATA_SECTION_UTTERANCE)) {
					LOGE("[Recorder WARNING] One of the observer returned false");
				}
			}
		}
	}

	while (!(mStopStreamingThread.load())) {
		int ret = -1;
		int cnt = 0;

		/* get feedback data */
		if (interface && interface->get_utterance_data) {
			ret = interface->get_utterance_data(index, &speech_data);
			if (0 != ret) {
				/* empty queue */
				MWR_LOGD("[DEBUG] No feedback data. Waiting mode : %d", ret);

				/* waiting */
				while (!(mStopStreamingThread.load())) {
					this_thread::sleep_for(chrono::milliseconds(10));
					if (index < interface->get_utterance_data_count()) {
						MWR_LOGI("[INFO] Resume thread");
						break;
					}
					if (g_speech_pcm_wait_count < cnt) {
						MWR_LOGE("[ERROR] Wrong request, there's no pcm data");
						for (const auto& observer : mObservers) {
							if (observer) {
								if (!observer->on_streaming_audio_data(
									WAKEUP_SPEECH_STREAMING_EVENT_FAIL, NULL, 0)) {
									LOGE("[Recorder WARNING] One of the observer returned false");
								}
							}
						}
						return;
					}
					cnt++;
				}
				MWR_LOGI("[INFO] Finish to wait for new feedback data come");

				/* resume feedback thread */
				continue;
			}

			for (const auto& observer : mObservers) {
				if (observer) {
					if (!observer->on_streaming_audio_data(
						speech_data.event, speech_data.buffer, speech_data.len)) {
						LOGE("[Recorder WARNING] One of the observer returned false");
					}
				}
			}

			if (WAKEUP_SPEECH_STREAMING_EVENT_FINISH == speech_data.event) {
				MWR_LOGI("[INFO] Finish to get and send speech data");
				finish_event_sent = true;
				break;
			}

			index++;
		}
	}

	if (true != finish_event_sent) {
		unsigned char final_buffer[2] = {'\0', };
		for (const auto& observer : mObservers) {
			if (observer) {
				if (!observer->on_streaming_audio_data(
					WAKEUP_SPEECH_STREAMING_EVENT_FINISH, final_buffer, sizeof(final_buffer))) {
					LOGE("[Recorder WARNING] One of the observer returned false");
				}
			}
		}
	}
}

void CWakeupEngineManager::start_streaming_current_utterance_data()
{
	if (mStreamingThread.joinable()) {
		MWR_LOGE("ERROR : mStreamingThread is joinable, will not start a new thread");
		return;
	}
	mStreamingThread = thread(&CWakeupEngineManager::streaming_speech_data_thread_func, this);
}

void CWakeupEngineManager::stop_streaming_current_utterance_data()
{
	if (mStreamingThread.joinable()) {
		MWR_LOGD("mStreamingThread is joinable, trying join()");
		mStopStreamingThread.store(true);
		mStreamingThread.join();
	}
	mStopStreamingThread.store(false);
}

void CWakeupEngineManager::update_manager_state(wakeup_manager_state_e state)
{
	for (const auto& info : mEngineInfo) {
		if (info.interface.update_manager_state) {
			try {
				info.interface.update_manager_state(state);
			} catch (const std::exception& e) {
				MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
					info.engine_name.c_str(), e.what());
			}
		}
	}
}

void CWakeupEngineManager::update_recognition_result(string appid, int result)
{
	if (mSelectedEngine) {
		if (mSelectedEngine->interface.update_recognition_result) {
			mSelectedEngine->interface.update_recognition_result(appid.c_str(), result);
		}
	}
}

void CWakeupEngineManager::engine_add_target_assistant(string engine_name, string appid)
{
	const auto& iter = find_if(mEngineInfo.begin(), mEngineInfo.end(),
		[engine_name](const EngineInfo& info) {
			return (0 == info.engine_name.compare(engine_name));
		});

	if (mEngineInfo.end() == iter) {
		/* Not found, add new library */
		pkgmgrinfo_appinfo_h handle;
		int ret = pkgmgrinfo_appinfo_get_appinfo(engine_name.c_str(), &handle);
		if (PMINFO_R_OK == ret) {
			char *root_path = nullptr;
			ret = pkgmgrinfo_appinfo_get_root_path(handle, &root_path);
			if (PMINFO_R_OK == ret && nullptr != root_path) {
				string path = root_path;
				path += "/";
				path += MA_WAKEUP_DEDICATED_ENGINE_PATH;
				add_engine(engine_name, path);
			}
			pkgmgrinfo_appinfo_destroy_appinfo(handle);
		}
		/* Find again to add appid to the newly created engine's assistant list */
		const auto &new_iter = find_if(mEngineInfo.begin(), mEngineInfo.end(),
			[engine_name](const EngineInfo& info) {
				return (0 == info.engine_name.compare(engine_name));
			});
		if (mEngineInfo.end() != new_iter) {
			new_iter->assistant_list.push_back(appid);
		}
	} else {
		/* If the engine already exists, simply add the appid to the assistant list */
		iter->assistant_list.push_back(appid);
	}
}

void CWakeupEngineManager::engine_add_wakeup_word(string appid, string wakeup_word, string language)
{
	for (const auto& info : mEngineInfo) {
		bool found = contains(info.assistant_list, appid);
		if (found) {
			if (info.interface.add_wakeup_word) {
				try {
					info.interface.add_wakeup_word(appid.c_str(), wakeup_word.c_str(), language.c_str());
				} catch (const std::exception& e) {
					MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
						info.engine_name.c_str(), e.what());
				}
			}
		}
	}
}

void CWakeupEngineManager::engine_set_assistant_specific_command(string appid, string command)
{
	for (const auto& info : mEngineInfo) {
		bool found = contains(info.assistant_list, appid);
		if (found) {
			if (info.interface.set_assistant_specific_command) {
				try {
					info.interface.set_assistant_specific_command(appid.c_str(), command.c_str());
				} catch (const std::exception& e) {
					MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
						info.engine_name.c_str(), e.what());
				}
			}
		}
	}
}

void CWakeupEngineManager::engine_feed_audio_data(long time, void* data, int len)
{
	for (const auto& info : mEngineInfo) {
		if (info.activated &&
			info.audio_data_require_status &&
			info.interface.feed_audio_data) {
			try {
				int ret = info.interface.feed_audio_data(time, data, len);
				if (0 != ret) {
					LOGE("[ERROR] Fail to feed speech data, ret(%d) : %s", ret, info.engine_name.c_str());
				}
			} catch (const std::exception& e) {
				MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
					info.engine_name.c_str(), e.what());
			}
		}
	}
}

bool CWakeupEngineManager::on_wakeup_event(string engine_name, wakeup_event_info info)
{
	MWR_LOGD("[ENTER]");

	for (const auto& observer : mObservers) {
		if (observer) {
			if (!observer->on_wakeup_event(engine_name, info)) {
				LOGE("[Recorder WARNING] One of the observer returned false");
			}
		}
	}

	return true;
}

bool CWakeupEngineManager::on_speech_status(string engine_name, wakeup_service_speech_status_e status)
{
	MWR_LOGD("[ENTER]");

	for (const auto& observer : mObservers) {
		if (observer) {
			if (!observer->on_speech_status(engine_name, status)) {
				LOGE("[Recorder WARNING] One of the observer returned false");
			}
		}
	}

	return true;
}

bool CWakeupEngineManager::on_error(string engine_name, int error_code, string error_message)
{
	MWR_LOGD("[ENTER]");

	for (const auto& observer : mObservers) {
		if (observer) {
			if (!observer->on_error(engine_name, error_code, error_message)) {
				LOGE("[Recorder WARNING] One of the observer returned false");
			}
		}
	}

	return true;
}

bool CWakeupEngineManager::on_audio_data_require_status(string engine_name, bool require)
{
	MWR_LOGD("[ENTER] %s, %d", engine_name.c_str(), require);

	bool found = false;
	// LOCK REQUIRED
	int count = 0;
	for (auto& info : mEngineInfo) {
		if (info.engine_name.compare(engine_name) == 0) {
			found = true;
			info.audio_data_require_status = require;
		}
		if (info.activated && info.audio_data_require_status) {
			count++;
		}
	}
	MWR_LOGD("count : %d", count);
	if (count > 0) {
		mAudioDataRequired = true;
	} else {
		mAudioDataRequired = false;
	}

	if (found) {
		for (const auto& observer : mObservers) {
			if (observer) {
				if (!observer->on_audio_data_require_status(engine_name, require)) {
					LOGE("[Recorder WARNING] One of the observer returned false");
				}
			}
		}
	}
	// UNLOCK REQUIRED
	return true;
}

void CWakeupEngineManager::add_engine(string name, string path)
{
	MWR_LOGD("Name (%s), Filepath(%s)", name.c_str(), path.c_str());

	char* error = NULL;
	EngineInfo info;
	info.engine_handle = dlopen(path.c_str(), RTLD_LAZY);
	if (nullptr != (error = dlerror()) || nullptr == info.engine_handle) {
		MWR_LOGD("[ERROR] Fail to dlopen(%s), error(%s)", path.c_str(), error);
		if (info.engine_handle) dlclose(info.engine_handle);
		return;
	}

	/* Interfaces without version information */
	info.interface.initialize =
		(wakeup_engine_initialize)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_INITIALIZE);
	info.interface.deinitialize =
		(wakeup_engine_deinitialize)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_DEINITIALIZE);
	info.interface.activate =
		(wakeup_engine_activate)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_ACTIVATE);
	info.interface.deactivate =
		(wakeup_engine_deactivate)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_DEACTIVATE);
	info.interface.add_wakeup_word =
		(wakeup_engine_add_wakeup_word)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_ADD_WAKEUP_WORD);
	info.interface.add_language =
		(wakeup_engine_add_language)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_ADD_LANGUAGE);
	info.interface.set_language =
		(wakeup_engine_set_language)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_SET_LANGUAGE);
	info.interface.update_manager_state =
		(wakeup_engine_update_manager_state)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_UPDATE_MANAGER_STATE);
	info.interface.update_recognition_result =
		(wakeup_engine_update_recognition_result)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_UPDATE_RECOGNITION_RESULT);
	info.interface.set_audio_format =
		(wakeup_engine_set_audio_format)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_SET_AUDIO_FORMAT);
	info.interface.get_audio_format =
		(wakeup_engine_get_audio_format)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_GET_AUDIO_FORMAT);
	info.interface.feed_audio_data =
		(wakeup_engine_feed_audio_data)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_FEED_AUDIO_DATA);
	info.interface.get_utterance_data_count =
		(wakeup_engine_get_utterance_data_count)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_GET_UTTERANCE_DATA_COUNT);
	info.interface.get_utterance_data =
		(wakeup_engine_get_utterance_data)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_GET_UTTERANCE_DATA);
	info.interface.get_wake_word_data_count =
		(wakeup_engine_get_wake_word_data_count)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_GET_WAKE_WORD_DATA_COUNT);
	info.interface.get_wake_word_data =
		(wakeup_engine_get_wake_word_data)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_GET_WAKE_WORD_DATA);
	info.interface.set_assistant_specific_command =
		(wakeup_engine_set_assistant_specific_command)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_SET_ASSISTANT_SPECIFIC_COMMAND);
	info.interface.set_wake_word_audio_require_flag =
		(wakeup_engine_set_wake_word_audio_require_flag)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_SET_WAKE_WORD_AUDIO_REQUIRE_FLAG);
	info.interface.set_wakeup_event_callback =
		(wakeup_engine_set_wakeup_event_callback)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_SET_WAKEUP_EVENT_CALLBACK);
	info.interface.set_speech_status_callback =
		(wakeup_engine_set_speech_status_callback)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_SET_SPEECH_STATUS_CALLBACK);
	info.interface.set_error_callback =
		(wakeup_engine_set_error_callback)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_SET_ERROR_CALLBACK);
	info.interface.set_audio_data_require_status_callback =
		(wakeup_engine_set_audio_data_require_status_callback)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_SET_AUDIO_DATA_REQUIRE_STATUS_CALLBACK);

	/* Interfaces after version 1 */
	info.interface.get_version =
		(wakeup_engine_get_version)dlsym(info.engine_handle,
		MA_WAKEUP_ENGINE_FUNC_GET_VERSION);

	info.version = 0;
	info.engine_path = path;
	info.engine_name = name;

	info.activated = false;
	info.audio_data_require_status = false;

	/* All the necessary information has already been set properly */
	mEngineInfo.push_back(info);

	MWR_LOGD("Initializing wakeup engine : %s %p",
		info.engine_path.c_str(),
		info.interface.initialize);

	/* Workaround for registering C-style callbacks */
	typedef struct {
		CWakeupEngineManager *manager;
		string engine_name;
	} CallbackUserData;

	static deque<CallbackUserData> callback_user_data;

	CallbackUserData user_data;
	user_data.manager = this;
	user_data.engine_name = info.engine_name;
	callback_user_data.push_back(user_data);

	try {
		if (info.interface.set_wakeup_event_callback) {
			info.interface.set_wakeup_event_callback(
				[](wakeup_event_info info, void* user_data) {
					CallbackUserData *data = static_cast<CallbackUserData*>(user_data);
					if (nullptr == data) return;
					if (nullptr == data->manager) return;
					info.wakeup_engine = data->engine_name.c_str();
					data->manager->on_wakeup_event(data->engine_name, info);
				}, &(callback_user_data.back()));
		}

		if (info.interface.set_audio_data_require_status_callback) {
			info.interface.set_audio_data_require_status_callback(
				[](bool require, void* user_data) {
					CallbackUserData *data = static_cast<CallbackUserData*>(user_data);
					if (nullptr == data) return;
					if (nullptr == data->manager) return;
					data->manager->on_audio_data_require_status(data->engine_name, require);
				}, &(callback_user_data.back()));
		}

		if (info.interface.initialize) {
			info.interface.initialize();
		}
		if (info.interface.get_version) {
			int version;
			if (0 == info.interface.get_version(&version)) {
				info.version = version;
			}
		}
	} catch (const std::exception& e) {
		MWR_LOGE("[ERROR] wakeup engine %s threw exception : %s",
			info.engine_name.c_str(), e.what());
	}
}

void CWakeupEngineManager::add_engine_directory(string name, string path)
{
	if (0 == path.size()) return;

	DIR* dp = opendir(path.c_str());
	if (NULL == dp) {
		MWR_LOGD("Failed opening directory : %s", path.c_str());
	} else {
		struct dirent *dirp = NULL;
		string filepath;
		do {
			dirp = readdir(dp);

			if (NULL != dirp) {
				if (!strcmp(".", dirp->d_name) || !strcmp("..", dirp->d_name))
					continue;

				if (DT_REG != dirp->d_type) /* If not a regular file */
					continue;

				filepath = path;
				filepath += "/";
				filepath += dirp->d_name;

				if (filepath.length() >= _POSIX_PATH_MAX) {
					MWR_LOGD("File path is too long : %s", filepath.c_str());
					closedir(dp);
					return;
				}
				add_engine(name, filepath);
			}
		} while (NULL != dirp);

		closedir(dp);
	}
}

} // wakeup
} // multiassistant
