/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wakeup_manager.h"
#include "wakeup_manager_main.h"
#include "wakeup_policy_default.h"
#include "dependency_resolver.h"

#include <algorithm>
namespace multiassistant
{
namespace wakeup
{

static bool check_language_valid(string language)
{
	return true;
}

static bool initialize_wakeup_event_info(wakeup_event_info* wakeup_info)
{
	bool ret = false;
	if (wakeup_info) {
		ret = true;

		wakeup_info->wakeup_appid = nullptr;
		wakeup_info->wakeup_word = nullptr;
		wakeup_info->wakeup_language = nullptr;
		wakeup_info->wakeup_voice_id = nullptr;
		wakeup_info->wakeup_engine = nullptr;
		wakeup_info->wakeup_confidence_score = 0.0f;

		wakeup_info->wakeup_start_time = 0;
		wakeup_info->wakeup_end_time = 0L;
		wakeup_info->wakeup_time_valid = false;

		wakeup_info->extra_data = nullptr;
		wakeup_info->extra_data_length = 0;
		wakeup_info->extra_data_description = nullptr;
	}
	return ret;
}

CWakeupManager::CWakeupManager(IWakeupEventObserver* wakeup_observer, ISettingValueObserver* setting_observer)
{
	initialize_wakeup_event_info(&mLastWakeupEventInfo);

	if (wakeup_observer) {
		subscribe_wakeup_observer(wakeup_observer);
	}
	if (setting_observer) {
		subscribe_setting_observer(setting_observer);
	}
}

CWakeupManager::~CWakeupManager()
{
	MWR_LOGI("Wakeup Manager is now being destroyed");
}

void CWakeupManager::initialize_wakeup_policy()
{
	mWakeupPolicy.reset(new CWakeupPolicyDefault{&mPolicyEventObserver});

	/* Default Policy specific initialization */
	CWakeupPolicyDefault *policy =
		dynamic_cast<CWakeupPolicyDefault*>(mWakeupPolicy.get());
	if (policy) {
		int priority = 0;

		policy->set_delay(mWakeupSettings.get_wakeup_policy_delay());
		MWR_LOGD("Setting Delay : %f", mWakeupSettings.get_wakeup_policy_delay());
		for (const auto& assistant : mWakeupSettings.get_wakeup_policy_priority()) {
			policy->set_assistant_priority(assistant, ++priority);
			MWR_LOGD("Setting Priority : %d %s", priority, assistant.c_str());
		}
	}
}

bool CWakeupManager::initialize()
{
	MWR_LOGD("[ENTER]");

	mPolicyEventObserver.set_wakeup_manager(this);
	mEngineEventObserver.set_wakeup_manager(this);
	mAudioEventObserver.set_wakeup_manager(this);
	mSettingsEventObserver.set_wakeup_manager(this);

	mWakeupSettings.subscribe(&mSettingsEventObserver);
	mWakeupSettings.initialize();

	mAudioEventObserver.set_wakeup_engine_manager(&mWakeupEngineManager);
	mAudioManager.subscribe(&mAudioEventObserver);
	mAudioManager.initialize();

	initialize_wakeup_policy();

	mWakeupEngineManager.subscribe(&mEngineEventObserver);
	mWakeupEngineManager.initialize();

	mas_proxy_interface interface;
	interface.process_event = wakeup_manager_process_event;
	interface.feed_audio_data = wakeup_manager_feed_audio_data;

	dependency_resolver_initialize(interface);

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::deinitialize()
{
	MWR_LOGI("[ENTER]");

	dependency_resolver_deinitialize();

	mWakeupEngineManager.unsubscribe(&mEngineEventObserver);
	mWakeupEngineManager.deinitialize();

	mAudioManager.unsubscribe(&mAudioEventObserver);
	mAudioManager.deinitialize();

	mWakeupSettings.deinitialize();
	mAssistantLanguageInfo.clear();

	MWR_LOGI("[END]");
	return true;
}

void CWakeupManager::subscribe_wakeup_observer(IWakeupEventObserver *observer)
{
	mWakeupObservers.push_back(observer);
}

void CWakeupManager::unsubscribe_wakeup_observer(IWakeupEventObserver *observer)
{
	auto iter = find(mWakeupObservers.begin(), mWakeupObservers.end(), observer);
	if (iter != mWakeupObservers.end()) {
		mWakeupObservers.erase(iter);
	}
}

void CWakeupManager::subscribe_setting_observer(ISettingValueObserver* observer)
{
	mSettingObservers.push_back(observer);
}

void CWakeupManager::unsubscribe_setting_observer(ISettingValueObserver* observer)
{
	auto iter = find(mSettingObservers.begin(), mSettingObservers.end(), observer);
	if (iter != mSettingObservers.end()) {
		mSettingObservers.erase(iter);
	}
}

bool CWakeupManager::activate(void)
{
	MWR_LOGD("[ENTER]");

	if (WAKEUP_MANAGER_STATE_INACTIVE != mWakeupManagerState)
		return false;

	/* Activate assistants that supports current voice input language */
	set_language(mWakeupSettings.get_current_language());

	change_manager_state(WAKEUP_MANAGER_STATE_LISTENING);
	if (mWakeupEngineManager.get_audio_data_required()) {
		mAudioManager.set_recording_session(RECORDING_SESSION_WAKE_WORD);
		mAudioManager.start_recording(true);
	}

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::deactivate(void)
{
	MWR_LOGD("[ENTER]");

	if (WAKEUP_MANAGER_STATE_INACTIVE == mWakeupManagerState)
		return false;

	mAudioManager.stop_recording(true);
	change_manager_state(WAKEUP_MANAGER_STATE_INACTIVE);

	stop_streaming_utterance_data();
	stop_streaming_previous_utterance_data();
	stop_streaming_follow_up_data();

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::add_assistant_language(string appid, string language)
{
	MWR_LOGD("[ENTER]");
	bool found = false;
	for (auto& info : mAssistantLanguageInfo) {
		if(0 == info.appid.compare(appid)) {
			info.languageList.push_back(language);
			found = true;
		}
	}
	if(false == found) {
		AssistantLanguageInfo info;
		info.appid = appid;
		info.languageList.push_back(language);
		mAssistantLanguageInfo.push_back(info);
	}
	if (0 == mCurrentLanguage.compare(language)) {
		mAssistantSupportsCurrentLanguage[appid] = true;
		if (true == mAssistantEnabled[appid]) {
			mWakeupEngineManager.set_assistant_activated(appid, true);
		}
	}
	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::add_assistant_wakeup_word(string appid, string wakeup_word, string language)
{
	MWR_LOGD("[ENTER]");

	mWakeupEngineManager.engine_add_wakeup_word(appid, wakeup_word, language);

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::set_assistant_wakeup_engine(string appid, string engine)
{
	MWR_LOGD("[ENTER]");

	mWakeupEngineManager.engine_add_target_assistant(engine, appid);

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::set_assistant_language(string appid, string language)
{
	MWR_LOGD("[ENTER] : %s, %s", appid.c_str(), language.c_str());

	mWakeupEngineManager.set_assistant_language(appid, language);

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::set_assistant_enabled(string appid, bool enabled)
{
	MWR_LOGD("[ENTER]");

	mAssistantEnabled[appid] = enabled;
	bool activated = enabled;
	if (false == mAssistantSupportsCurrentLanguage[appid]) {
		activated = false;
	}
	if (0 == appid.compare(mCurrentDefaultAssistant)) {
		activated = true;
	}
	mWakeupEngineManager.set_assistant_activated(appid, activated);

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::set_default_assistant(string appid)
{
	MWR_LOGD("[ENTER]");

	/* Check if previous default assistant has to be deactivated */
	bool activated = true;
	if (false == mAssistantSupportsCurrentLanguage[mCurrentDefaultAssistant]) {
		activated = false;
	}
	if (false == mAssistantEnabled[mCurrentDefaultAssistant]) {
		activated = false;
	}
	mWakeupEngineManager.set_assistant_activated(mCurrentDefaultAssistant, activated);

	/* New default assistant has to be activated no matter what */
	mWakeupEngineManager.set_assistant_activated(appid, true);
	mCurrentDefaultAssistant = appid;

	MWR_LOGD("[END]");
	return true;
}

string CWakeupManager::get_default_assistant()
{
	return mCurrentDefaultAssistant;
}

bool CWakeupManager::get_assistant_enabled(string appid)
{
	return mAssistantEnabled[appid];
}

bool CWakeupManager::set_language(string language)
{
	bool ret = false;
	MWR_LOGD("[ENTER] : %s", language.c_str());

	if (check_language_valid(language)) {
		mCurrentLanguage = language;

		bool found = false;
		for (auto& info : mAssistantLanguageInfo) {
			found = false;
			for(auto it = info.languageList.begin(); it != info.languageList.end(); it++) {
				if(language == *it) {
					found = true;
					break;
				}
			}
			if(false == found) {
				mAssistantSupportsCurrentLanguage[info.appid] = false;
			} else {
				mAssistantSupportsCurrentLanguage[info.appid] = true;
			}

			bool activated = found;
			if (false == mAssistantEnabled[info.appid]) {
				activated = false;
			}
			if (0 == info.appid.compare(mCurrentDefaultAssistant)) {
				activated = true;
			}
			mWakeupEngineManager.set_assistant_activated(info.appid, activated);
		}

		mWakeupEngineManager.set_language(language);
		ret = true;
	} else {
		MWR_LOGE("[ERROR] Not supported language (%s)", language.c_str());
	}

	MWR_LOGD("[END]");
	return ret;
}

bool CWakeupManager::get_voice_key_pressed()
{
	return mVoiceKeyPressed;
}

STREAMING_MODE CWakeupManager::get_streaming_mode()
{
	return mStreamingMode;
}

bool CWakeupManager::set_streaming_mode(STREAMING_MODE mode)
{
	mStreamingMode = mode;
	return true;
}

bool CWakeupManager::change_manager_state(wakeup_manager_state_e state)
{
	mWakeupManagerState = state;
	mWakeupEngineManager.update_manager_state(state);
	return true;
}

wakeup_manager_state_e CWakeupManager::get_manager_state()
{
	return mWakeupManagerState;
}

bool CWakeupManager::update_voice_feedback_state(string appid, bool state)
{
	MWR_LOGD("[ENTER]");

	if (state) {
		if (WAKEUP_MANAGER_STATE_LISTENING == mWakeupManagerState ||
			WAKEUP_MANAGER_STATE_PROCESSING == mWakeupManagerState) {
			change_manager_state(WAKEUP_MANAGER_STATE_VOICE_FEEDBACK);
		}
	} else {
		if (WAKEUP_MANAGER_STATE_VOICE_FEEDBACK == mWakeupManagerState) {
			change_manager_state(WAKEUP_MANAGER_STATE_LISTENING);
		}
	}

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::send_assistant_specific_command(string appid, string command)
{
	MWR_LOGD("[ENTER]");

	/*
	static const string voice_key_pressed{"voice_key_pressed"};
	static const string voice_key_released{"voice_key_released"};

	if (0 == command.compare(voice_key_pressed)) {
		process_event(MA_PLUGIN_EVENT_VOICE_KEY_PRESSED, NULL, 0);
	} else if (0 == command.compare(voice_key_released)) {
		process_event(MA_PLUGIN_EVENT_VOICE_KEY_RELEASED, NULL, 0);
	}
	*/

	mWakeupEngineManager.engine_set_assistant_specific_command(appid, command);

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::set_background_volume(string appid, double ratio)
{
	MWR_LOGD("[DEBUG] set background volume (%f)", ratio);

	int ret = 0;
	mAudioManager.set_background_volume(ratio);
	return ret;
}

bool CWakeupManager::update_recognition_result(string appid, int result)
{
	MWR_LOGD("[ENTER]");
	if (WAKEUP_MANAGER_STATE_UTTERANCE == mWakeupManagerState) {
		mAudioManager.set_recording_session(RECORDING_SESSION_WAKE_WORD);
	}
	mWakeupEngineManager.update_recognition_result(appid, result);
	if (WAKEUP_MANAGER_STATE_PROCESSING == mWakeupManagerState ||
		WAKEUP_MANAGER_STATE_UTTERANCE == mWakeupManagerState) {
		change_manager_state(WAKEUP_MANAGER_STATE_LISTENING);
		stop_streaming_utterance_data();
		stop_streaming_previous_utterance_data();
		stop_streaming_follow_up_data();
	}
	MWR_LOGD("[END]");
	return true;
}

static long get_current_milliseconds_after_epoch()
{
	auto now = chrono::system_clock::now();
	auto now_ms = chrono::time_point_cast<chrono::milliseconds>(now);
	/* number of milliseconds since the epoch of system_clock */
	auto value = now_ms.time_since_epoch();

	return value.count();
}

bool CWakeupManager::process_event(ma_plugin_event_e event, void* data, int len)
{
	MWR_LOGD("[ENTER] : %d", event);
	if (WAKEUP_MANAGER_STATE_INACTIVE == mWakeupManagerState)
		return false;

	// LOCK REQUIRED
	if (MA_PLUGIN_EVENT_VOICE_KEY_PRESSED == event) {
		if (mVoiceKeyPressed != true) {
			mAudioManager.stop_recording(true);

			stop_streaming_utterance_data();
			stop_streaming_previous_utterance_data();
			stop_streaming_follow_up_data();

			mAudioManager.clear_audio_data();
			change_manager_state(WAKEUP_MANAGER_STATE_UTTERANCE);

			mVoiceKeyPressed = true;

			/* Start recorder thread using appropriate recording device */
			mAudioManager.set_recording_session(RECORDING_SESSION_UTTERANCE);
			mAudioManager.start_recording(true);

			/* Wakeup default assistant */
			/* TODO: apply conversation timeout for selecting assistant here */
			wakeup_event_info wakeup_info;
			initialize_wakeup_event_info(&wakeup_info);
			/* Make sure to use background data */
			wakeup_info.wakeup_time_valid = true;
			wakeup_info.wakeup_end_time = get_current_milliseconds_after_epoch();
			wakeup_info.wakeup_engine = WAKEUP_ENGINE_VOICE_KEY;

			string appid = mWakeupSettings.get_default_assistant_appid();
			wakeup_info.wakeup_appid = appid.c_str();
			MWR_LOGD("wakeup_appid : %s", wakeup_info.wakeup_appid);

			set_last_wakeup_event_info(wakeup_info);
			mWakeupEngineManager.set_selected_wakeup_info(wakeup_info);
			for (const auto& observer : mWakeupObservers) {
				observer->on_wakeup(wakeup_info);
			}
		}
	} else if (MA_PLUGIN_EVENT_VOICE_KEY_RELEASED_AFTER_PUSH == event) {
		if (mVoiceKeyPressed != false) {
			mAudioManager.finalize_audio_data();
			mVoiceKeyPressed = false;

			if (STREAMING_MODE::UTTERANCE == mStreamingMode) {
				change_manager_state(WAKEUP_MANAGER_STATE_PROCESSING);
			} else {
				change_manager_state(WAKEUP_MANAGER_STATE_LISTENING);
			}

			if (mWakeupEngineManager.get_audio_data_required()) {
				/* Restart recorder thread using appropriate recording device */
				mAudioManager.stop_recording(true);
				mAudioManager.set_recording_session(RECORDING_SESSION_WAKE_WORD);
				mAudioManager.start_recording(true);
			} else {
				mAudioManager.stop_recording(true);
			}
		}
	} else if (MA_PLUGIN_EVENT_VOICE_KEY_RELEASED_AFTER_TAP == event) {
		if (mVoiceKeyPressed != false) {
			mVoiceKeyPressed = false;
		}
	}
	// UNLOCK REQUIRED

	MWR_LOGD("[END]");
	return true;
}

vector<IWakeupEventObserver*> CWakeupManager::get_wakeup_observers()
{
	return mWakeupObservers;
}

void CWakeupManager::set_last_wakeup_event_info(wakeup_event_info wakeup_info)
{
	mLastWakeupEventInfo = wakeup_info;
}

vector<ISettingValueObserver*> CWakeupManager::get_setting_observers()
{
	return mSettingObservers;
}

static Eina_Bool streaming_duration_expired(void *data)
{
	MWR_LOGD("[ENTER]");
	CWakeupManager *wakeup_manager = static_cast<CWakeupManager*>(data);
	if (nullptr == wakeup_manager) return ECORE_CALLBACK_CANCEL;

	CAudioManager *audio_manager = wakeup_manager->get_audio_manager();
	CWakeupEngineManager *engine_manager = wakeup_manager->get_engine_manager();

	if (nullptr == audio_manager) return ECORE_CALLBACK_CANCEL;
	if (nullptr == engine_manager) return ECORE_CALLBACK_CANCEL;

	switch(wakeup_manager->get_streaming_mode()) {
		case STREAMING_MODE::UTTERANCE:
			audio_manager->stop_streaming_current_utterance_data();
			engine_manager->stop_streaming_current_utterance_data();
		break;
		case STREAMING_MODE::PREVIOUS_UTTERANCE:
			audio_manager->stop_streaming_previous_utterance_data();
		break;
		case STREAMING_MODE::FOLLOW_UP:
			audio_manager->stop_streaming_follow_up_data();
		break;
	}

	unsigned char final_buffer[2] = {'\0', };
	vector<IWakeupEventObserver*> observers = wakeup_manager->get_wakeup_observers();
	for (const auto& observer : observers) {
		observer->on_streaming_audio_data(
			WAKEUP_SPEECH_STREAMING_EVENT_FINISH, final_buffer, sizeof(final_buffer));
	}
	wakeup_manager->set_streaming_mode(STREAMING_MODE::NONE);

	if (WAKEUP_MANAGER_STATE_UTTERANCE == wakeup_manager->get_manager_state()) {
		wakeup_manager->change_manager_state(WAKEUP_MANAGER_STATE_PROCESSING);
	}

	return ECORE_CALLBACK_CANCEL;
}

bool CWakeupManager::start_streaming_utterance_data()
{
	MWR_LOGD("[ENTER]");

	mAudioManager.stop_streaming_current_utterance_data();
	mWakeupEngineManager.stop_streaming_current_utterance_data();

	mStreamingMode = STREAMING_MODE::UTTERANCE;

	bool streaming_by_manager = true;
	if (false == mLastWakeupEventInfo.wakeup_time_valid) {
		mWakeupEngineManager.start_streaming_current_utterance_data();
		streaming_by_manager = false;
	} else {
		mAudioManager.start_streaming_current_utterance_data(mLastWakeupEventInfo.wakeup_end_time);
	}

	ecore_thread_main_loop_begin();
	if (mStreamingDurationTimer) {
		ecore_timer_del(mStreamingDurationTimer);
		mStreamingDurationTimer = nullptr;
	}
	if (streaming_by_manager) {
		mStreamingDurationTimer = ecore_timer_add(
			mWakeupSettings.get_streaming_duration_max(),
			streaming_duration_expired, this);
	}
	ecore_thread_main_loop_end();

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::stop_streaming_utterance_data()
{
	MWR_LOGD("[ENTER]");
	if (STREAMING_MODE::UTTERANCE != mStreamingMode) return false;

	if (mStreamingDurationTimer) {
		ecore_thread_main_loop_begin();
		ecore_timer_del(mStreamingDurationTimer);
		mStreamingDurationTimer = nullptr;
		ecore_thread_main_loop_end();
	}
	if (WAKEUP_MANAGER_STATE_UTTERANCE == mWakeupManagerState) {
		change_manager_state(WAKEUP_MANAGER_STATE_PROCESSING);
	}
	mAudioManager.stop_streaming_current_utterance_data();
	mWakeupEngineManager.stop_streaming_current_utterance_data();

	mStreamingMode = STREAMING_MODE::NONE;

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::start_streaming_follow_up_data()
{
	MWR_LOGD("[ENTER]");

	mAudioManager.stop_streaming_follow_up_data();
	mWakeupEngineManager.stop_streaming_current_utterance_data();

	mStreamingMode = STREAMING_MODE::FOLLOW_UP;

	/* For the follow up streaming, audio data should be recorded from now on */
	mAudioManager.stop_recording(true);
	mAudioManager.clear_audio_data();
	change_manager_state(WAKEUP_MANAGER_STATE_UTTERANCE);
	mAudioManager.set_recording_session(RECORDING_SESSION_FOLLOW_UP);
	mAudioManager.start_recording(true);

	mAudioManager.start_streaming_follow_up_data();

	ecore_thread_main_loop_begin();
	if (mStreamingDurationTimer) {
		ecore_timer_del(mStreamingDurationTimer);
		mStreamingDurationTimer = nullptr;
	}
	mStreamingDurationTimer = ecore_timer_add(
		mWakeupSettings.get_streaming_duration_max(),
		streaming_duration_expired, this);
	ecore_thread_main_loop_end();

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::stop_streaming_follow_up_data()
{
	MWR_LOGD("[ENTER]");
	if (STREAMING_MODE::FOLLOW_UP != mStreamingMode) return false;

	if (mStreamingDurationTimer) {
		ecore_thread_main_loop_begin();
		ecore_timer_del(mStreamingDurationTimer);
		mStreamingDurationTimer = nullptr;
		ecore_thread_main_loop_end();
	}
	if (WAKEUP_MANAGER_STATE_UTTERANCE == mWakeupManagerState) {
		change_manager_state(WAKEUP_MANAGER_STATE_PROCESSING);
	}
	mAudioManager.stop_streaming_follow_up_data();
	mWakeupEngineManager.stop_streaming_current_utterance_data();

	mStreamingMode = STREAMING_MODE::NONE;

	mAudioManager.stop_recording(true);
	mAudioManager.clear_audio_data();
	mAudioManager.set_recording_session(RECORDING_SESSION_WAKE_WORD);
	mAudioManager.start_recording(true);

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::start_streaming_previous_utterance_data()
{
	MWR_LOGD("[ENTER]");

	mAudioManager.stop_streaming_previous_utterance_data();
	mWakeupEngineManager.stop_streaming_current_utterance_data();

	mStreamingMode = STREAMING_MODE::PREVIOUS_UTTERANCE;
	mAudioManager.start_streaming_previous_utterance_data();

	ecore_thread_main_loop_begin();
	if (mStreamingDurationTimer) {
		ecore_timer_del(mStreamingDurationTimer);
		mStreamingDurationTimer = nullptr;
	}
	mStreamingDurationTimer = ecore_timer_add(
		mWakeupSettings.get_streaming_duration_max(),
		streaming_duration_expired, this);
	ecore_thread_main_loop_end();

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::stop_streaming_previous_utterance_data()
{
	MWR_LOGD("[ENTER]");
	if (STREAMING_MODE::PREVIOUS_UTTERANCE != mStreamingMode) return false;

	if (mStreamingDurationTimer) {
		ecore_thread_main_loop_begin();
		ecore_timer_del(mStreamingDurationTimer);
		mStreamingDurationTimer = nullptr;
		ecore_thread_main_loop_end();
	}
	if (WAKEUP_MANAGER_STATE_UTTERANCE == mWakeupManagerState) {
		change_manager_state(WAKEUP_MANAGER_STATE_PROCESSING);
	}
	mAudioManager.stop_streaming_previous_utterance_data();
	mWakeupEngineManager.stop_streaming_current_utterance_data();

	mStreamingMode = STREAMING_MODE::NONE;

	MWR_LOGD("[END]");
	return true;
}

bool CWakeupManager::get_audio_format(int* rate, int* channel, int* audio_type)
{
	MWR_LOGD("[ENTER]");

	if (!audio_type || !rate || !channel) {
		MWR_LOGE("[ERROR] Invalid parameter");
		return false;
	}

	dependency_resolver_get_audio_format(rate, channel, audio_type);

	MWR_LOGD("[END] rate(%d), channel(%d), audio_type(%d)", *rate, *channel, *audio_type);
	return true;
}

bool CWakeupManager::get_audio_source_type(char** type)
{
	MWR_LOGD("[ENTER]");

	if (!type) {
		MWR_LOGE("[ERROR] Invalid parameter");
		return false;
	}

	dependency_resolver_get_audio_source_type(type);

	MWR_LOGD("[END] type(%s)", *type);
	return true;
}

bool CWakeupManager::set_wake_word_audio_require_flag(bool require)
{
	MWR_LOGD("[ENTER]");

	mWakeupEngineManager.set_wake_word_audio_require_flag(require);

	MWR_LOGD("[END]");
	return true;
}

CWakeupPolicy* CWakeupManager::get_wakeup_policy()
{
	return mWakeupPolicy.get();
}

CWakeupEngineManager* CWakeupManager::get_engine_manager()
{
	return &mWakeupEngineManager;
}

CAudioManager* CWakeupManager::get_audio_manager()
{
	return &mAudioManager;
}

CWakeupSettings* CWakeupManager::get_wakeup_settings()
{
	return &mWakeupSettings;
}

void CWakeupManager::feed_audio_data(wakeup_speech_streaming_event_e event, void* buffer, int len)
{
	mAudioManager.feed_audio_data(event, buffer, len);
}

bool CWakeupManager::CEngineEventObserver::on_wakeup_event(string engine_name, wakeup_event_info wakeup_info)
{
	MWR_LOGD("[ENTER]");
	if (nullptr == mWakeupManager) return false;
	if (nullptr == wakeup_info.wakeup_appid) return false;

	if (0 != mWakeupManager->get_default_assistant().compare(wakeup_info.wakeup_appid)) {
		if (false == mWakeupManager->get_assistant_enabled(string{wakeup_info.wakeup_appid})) {
			MWR_LOGE("Wakeup event with deactivated appid : %s", wakeup_info.wakeup_appid);
			return false;
		}
	}

	CWakeupPolicy* policy = mWakeupManager->get_wakeup_policy();
	if (policy) {
		policy->wakeup_candidate(wakeup_info);
	}
	return true;
}

bool CWakeupManager::CEngineEventObserver::on_speech_status(string engine_name, wakeup_service_speech_status_e status)
{
	MWR_LOGD("[ENTER]");
	if (nullptr == mWakeupManager) return false;

	return true;
}

bool CWakeupManager::CEngineEventObserver::on_error(string engine_name, int error_code, string error_message)
{
	MWR_LOGD("[ENTER]");
	if (nullptr == mWakeupManager) return false;

	return true;
}

bool CWakeupManager::CEngineEventObserver::on_audio_data_require_status(string engine_name, bool require)
{
	MWR_LOGD("[ENTER]");
	if (nullptr == mWakeupManager) return false;
	if (WAKEUP_MANAGER_STATE_INACTIVE == mWakeupManager->get_manager_state()) return false;

	CAudioManager *audio_manager = mWakeupManager->get_audio_manager();
	CWakeupEngineManager *engine_manager = mWakeupManager->get_engine_manager();

	if (audio_manager && engine_manager) {
		if (engine_manager->get_audio_data_required()) {
			if (mWakeupManager->get_voice_key_pressed() != true) {
				audio_manager->set_recording_session(RECORDING_SESSION_WAKE_WORD);
				audio_manager->start_recording(true);
			}
		} else {
			if (mWakeupManager->get_voice_key_pressed() != true) {
				audio_manager->stop_recording(true);
			}
		}
	}
	return true;
}

bool CWakeupManager::CEngineEventObserver::on_streaming_audio_data(
	wakeup_speech_streaming_event_e event, void* buffer, unsigned int len)
{
	if (nullptr == mWakeupManager) return false;

	vector<IWakeupEventObserver*> observers = mWakeupManager->get_wakeup_observers();
	for (const auto& observer : observers) {
		observer->on_streaming_audio_data(event, buffer, len);
	}
	if (WAKEUP_SPEECH_STREAMING_EVENT_FINISH == event) {
		mWakeupManager->set_streaming_mode(STREAMING_MODE::NONE);
	}

	return true;
}

bool CWakeupManager::CEngineEventObserver::on_audio_streaming_data_section(
	ma_audio_streaming_data_section_e section)
{
	if (nullptr == mWakeupManager) return false;

	vector<IWakeupEventObserver*> observers = mWakeupManager->get_wakeup_observers();
	for (const auto& observer : observers) {
		observer->on_audio_streaming_data_section(section);
	}

	return true;
}

void CWakeupManager::CPolicyEventObserver::on_wakeup(wakeup_event_info wakeup_info)
{
	if (nullptr == mWakeupManager) return;

	CAudioManager *audio_manager = mWakeupManager->get_audio_manager();
	CWakeupEngineManager *engine_manager = mWakeupManager->get_engine_manager();
	if (nullptr == audio_manager || nullptr == engine_manager) return;

	mWakeupManager->stop_streaming_utterance_data();
	mWakeupManager->stop_streaming_previous_utterance_data();
	mWakeupManager->stop_streaming_follow_up_data();
	mWakeupManager->change_manager_state(WAKEUP_MANAGER_STATE_UTTERANCE);

	mWakeupManager->set_last_wakeup_event_info(wakeup_info);
	engine_manager->set_selected_wakeup_info(wakeup_info);
	vector<IWakeupEventObserver*> observers = mWakeupManager->get_wakeup_observers();
	for (const auto& observer : observers) {
		observer->on_wakeup(wakeup_info);
	}

	audio_manager->set_recording_session(RECORDING_SESSION_UTTERANCE);
}

bool CWakeupManager::CAudioEventObserver::on_recording_audio_data(long time, void* data, int len)
{
	if (nullptr == mWakeupManager) return false;
	if (nullptr == mEngineManager) return false;

	if (false == mEngineManager->get_audio_data_required()) return false;

	if (mWakeupManager->get_voice_key_pressed() != true) {
		/* When follow-up streaming in progress, no need to feed audio data to wakeup engines */
		if (STREAMING_MODE::FOLLOW_UP != mWakeupManager->get_streaming_mode()) {
			mEngineManager->engine_feed_audio_data(time, data, len);
		}
	}

	return true;
}

bool CWakeupManager::CAudioEventObserver::on_streaming_audio_data(
	wakeup_speech_streaming_event_e event, void* buffer, unsigned int len)
{
	if (nullptr == mWakeupManager) return false;

	vector<IWakeupEventObserver*> observers = mWakeupManager->get_wakeup_observers();
	for (const auto& observer : observers) {
		observer->on_streaming_audio_data(event, buffer, len);
	}
	if (WAKEUP_SPEECH_STREAMING_EVENT_FINISH == event) {
		mWakeupManager->set_streaming_mode(STREAMING_MODE::NONE);
	}

	return true;
}

bool CWakeupManager::CSettingsEventObserver::on_voice_input_language_changed(
	const char* language)
{
	if (nullptr == mWakeupManager || nullptr == language) return false;
	mWakeupManager->set_language(std::string(language));
	vector<ISettingValueObserver*> observers = mWakeupManager->get_setting_observers();
	for (const auto& observer : observers) {
		observer->on_value_changed();
	}
	return true;
}

bool CWakeupManager::CSettingsEventObserver::on_assistant_enabled_info_changed(
	const char* appid, bool enabled)
{
	if (nullptr == mWakeupManager || nullptr == appid) return false;
	mWakeupManager->set_assistant_enabled(std::string(appid), enabled);
	vector<ISettingValueObserver*> observers = mWakeupManager->get_setting_observers();
	for (const auto& observer : observers) {
		observer->on_value_changed();
	}
	return true;
}

bool CWakeupManager::CSettingsEventObserver::on_default_assistant_appid_changed(
	const char* appid)
{
	if (nullptr == mWakeupManager || nullptr == appid) return false;
	mWakeupManager->set_default_assistant(std::string(appid));
	vector<ISettingValueObserver*> observers = mWakeupManager->get_setting_observers();
	for (const auto& observer : observers) {
		observer->on_value_changed();
	}
	return true;
}

} // wakeup
} // multiassistant
