/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "wakeup_manager_wrapper.h"
#include "wakeup_manager_main.h"
#include "wakeup_manager.h"

#include <memory>

using namespace multiassistant::wakeup;

static wakeup_service_wakeup_event_cb g_wakeup_event_cb;
static void* g_wakeup_event_user_data;

static wakeup_service_speech_streaming_cb g_utterance_streaming_cb;
static void* g_utterance_streaming_user_data;

static wakeup_service_speech_streaming_cb g_previous_utterance_streaming_cb;
static void* g_previous_utterance_streaming_user_data;

static wakeup_service_speech_streaming_cb g_follow_up_streaming_cb;
static void* g_follow_up_streaming_user_data;

static wakeup_service_speech_status_cb g_speech_status_cb;
static void* g_speech_status_user_data;

static wakeup_service_setting_changed_cb g_setting_changed_cb;
static void* g_setting_changed_user_data;

static wakeup_service_error_cb g_error_cb;
static void* g_error_user_data;

static wakeup_service_streaming_section_changed_cb g_streaming_section_changed_cb;
static void* g_streaming_section_changed_user_data;

class CWakeupEventObserver : public IWakeupEventObserver
{
	void on_wakeup(wakeup_event_info wakeup_info) override;
	void on_streaming_audio_data(
		wakeup_speech_streaming_event_e event, void* buffer, unsigned int len) override;
	void on_audio_streaming_data_section(ma_audio_streaming_data_section_e section) override;
};

class CSettingValueObserver : public ISettingValueObserver
{
	void on_value_changed() override;
};

static std::unique_ptr<CWakeupEventObserver> g_wakeup_event_observer;
static std::unique_ptr<CSettingValueObserver> g_setting_value_observer;
static std::unique_ptr<CWakeupManager> g_wakeup_manager;

int wakeup_manager_initialize(void)
{
	MWR_LOGD("[ENTER]");

	try {
		g_wakeup_event_observer.reset(new CWakeupEventObserver);
		g_setting_value_observer.reset(new CSettingValueObserver);
		if (NULL == g_wakeup_event_observer || NULL == g_setting_value_observer) {
			return -1;
		}
		g_wakeup_manager.reset(new CWakeupManager(g_wakeup_event_observer.get(), g_setting_value_observer.get()));
		if (NULL == g_wakeup_manager) {
			return -1;
		}
	} catch (const std::exception& e) {
		MWR_LOGE("Exception caught : %s", e.what());
		return -1;
	}

	g_wakeup_event_cb = NULL;
	g_wakeup_event_user_data = NULL;

	g_utterance_streaming_cb = NULL;
	g_utterance_streaming_user_data = NULL;

	g_follow_up_streaming_cb = NULL;
	g_follow_up_streaming_user_data = NULL;

	g_speech_status_cb = NULL;
	g_speech_status_user_data = NULL;

	g_setting_changed_cb = NULL;
	g_setting_changed_user_data = NULL;

	g_error_cb = NULL;
	g_error_user_data = NULL;

	g_wakeup_manager->initialize();

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_deinitialize(void)
{
	MWR_LOGI("[ENTER]");

	if (nullptr == g_wakeup_manager) return -1;

	try {
		g_wakeup_manager->deinitialize();
		g_wakeup_manager.reset();

		g_wakeup_event_observer.reset();
		g_setting_value_observer.reset();
	} catch (const std::exception& e) {
		MWR_LOGE("Exception caught : %s", e.what());
		return -1;
	}

	MWR_LOGI("[END]");
	return 0;
}

int wakeup_manager_get_settings(ma_plugin_settings **settings, size_t *struct_size)
{
	if (NULL == settings || NULL == struct_size) {
		MWR_LOGD("[ERROR] Parameter is invalid, settings(%p), struct_size(%p)", settings, struct_size);
		return -1;
	}

	if (nullptr == g_wakeup_manager) return -1;
	CWakeupSettings* wakeup_settings = g_wakeup_manager->get_wakeup_settings();
	if (wakeup_settings) {
		const int PLUGIN_VERSION = 1;
		static ma_plugin_settings current_settings;
		current_settings.plugin_version = PLUGIN_VERSION;
		current_settings.ui_panel_enabled = wakeup_settings->get_ui_panel_enabled();
		*struct_size = sizeof(current_settings);
		*settings = &current_settings;
	}

	return 0;
}

int wakeup_manager_add_assistant_wakeup_word(const char* appid, const char* wakeup_word, const char* language)
{
	MWR_LOGD("[ENTER]");

	if (NULL == appid || NULL == wakeup_word) {
		MWR_LOGD("[ERROR] Parameter is invalid, appid(%s), wakeup_word(%s), language(%s)", appid, wakeup_word, language);
		return -1;
	}

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->add_assistant_wakeup_word(
		string{appid}, string{wakeup_word}, (language ? string{language} : string{}));

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_add_assistant_language(const char* appid, const char* language)
{
	MWR_LOGD("[ENTER]");

	if (NULL == appid || NULL == language) {
		MWR_LOGD("[ERROR] Parameter is invalid, appid(%s), language(%s)", appid, language);
		return -1;
	}

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->add_assistant_language(string{appid}, string{language});

	MWR_LOGD("[DEBUG] language(%s)", language);

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_assistant_wakeup_engine(const char* appid, const char* engine)
{
	MWR_LOGD("[ENTER]");

	if (NULL == appid || NULL == engine) {
		MWR_LOGD("[ERROR] Parameter is invalid, appid(%s), wakeup engine(%s)", appid, engine);
		return -1;
	}

	if (nullptr == g_wakeup_manager) return -1;
	MWR_LOGD("[DEBUG] appid(%s), wakeup engine(%s)", appid, engine);
	g_wakeup_manager->set_assistant_wakeup_engine(string{appid}, string{engine});

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_default_assistant(const char* appid)
{
	MWR_LOGD("[ENTER]");

	if (NULL == appid) {
		MWR_LOGD("[ERROR] Parameter is invalid, appid(%s)", appid);
		return -1;
	}

	if (nullptr == g_wakeup_manager) return -1;
	MWR_LOGD("[DEBUG] default_assistant appid(%s)", appid);
	g_wakeup_manager->set_default_assistant(string{appid});

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_get_default_assistant(const char** appid)
{
	static string default_assistant;

	if (nullptr == g_wakeup_manager) return -1;
	default_assistant = g_wakeup_manager->get_default_assistant();

	if (default_assistant.empty()) {
		*appid = nullptr;
	} else {
		*appid = default_assistant.c_str();
	}
	return 0;
}

int wakeup_manager_set_language(const char* language)
{
	MWR_LOGD("[ENTER]");

	if (NULL == language) {
		MWR_LOGD("[ERROR] Parameter is invalid, language(%s)", language);
		return -1;
	}

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->set_language(string{language});

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_change_state(wakeup_manager_state_e state)
{
	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->change_manager_state(state);
	return 0;
}

int wakeup_manager_activate(void)
{
	MWR_LOGD("[ENTER]");

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->activate();

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_deactivate(void)
{
	MWR_LOGD("[ENTER]");

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->deactivate();

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_update_voice_feedback_state(const char* appid, int state)
{
	MWR_LOGD("[ENTER]");

	string appid_string;
	if (NULL == appid) {
		MWR_LOGD("[ERROR] Parameter is invalid, appid(%s)", appid);
	} else {
		appid_string = appid;
	}

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->update_voice_feedback_state(appid_string, state);

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_send_assistant_specific_command(const char* appid, const char* command)
{
	MWR_LOGD("[ENTER]");

	string appid_string;
	string command_string;
	if (NULL == appid || NULL == command) {
		MWR_LOGD("[ERROR] Parameter is invalid, appid(%s), command(%s)", appid, command);
	}
	if (appid) appid_string = appid;
	if (command) command_string = command;

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->send_assistant_specific_command(appid_string, command_string);

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_background_volume(const char* appid, double ratio)
{
	MWR_LOGD("[ENTER]");

	string appid_string;
	if (NULL == appid) {
		MWR_LOGD("[ERROR] Parameter is invalid, appid(%s)", appid);
	} else {
		appid_string = appid;
	}

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->set_background_volume(appid_string, ratio);

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_update_recognition_result(const char* appid, int result)
{
	MWR_LOGD("[ENTER]");

	string appid_string;
	if (NULL == appid) {
		MWR_LOGD("[ERROR] Parameter is invalid, appid(%s)", appid);
	} else {
		appid_string = appid;
	}

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->update_recognition_result(appid_string, result);

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_process_event(int event, void* data, int len)
{
	MWR_LOGD("[ENTER]");

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->process_event(static_cast<ma_plugin_event_e>(event), data, len);

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_start_streaming_utterance_data(void)
{
	MWR_LOGD("[ENTER]");

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->start_streaming_utterance_data();

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_stop_streaming_utterance_data(void)
{
	MWR_LOGD("[ENTER]");

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->stop_streaming_utterance_data();

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_start_streaming_follow_up_data(void)
{
	MWR_LOGD("[ENTER]");

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->start_streaming_follow_up_data();

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_stop_streaming_follow_up_data(void)
{
	MWR_LOGD("[ENTER]");

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->stop_streaming_follow_up_data();

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_start_streaming_previous_utterance_data(void)
{
	MWR_LOGD("[ENTER]");

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->start_streaming_previous_utterance_data();

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_stop_streaming_previous_utterance_data(void)
{
	MWR_LOGD("[ENTER]");

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->stop_streaming_previous_utterance_data();

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_get_audio_format(int* rate, int* channel, int* audio_type)
{
	MWR_LOGD("[ENTER]");

	if (!audio_type || !rate || !channel) {
		MWR_LOGE("[ERROR] Invalid parameter");
		return -1;
	}

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->get_audio_format(rate, channel, audio_type);

	MWR_LOGD("[END] rate(%d), channel(%d), audio_type(%d)", *rate, *channel, *audio_type);
	return 0;
}

int wakeup_manager_get_audio_source_type(char** type)
{
	MWR_LOGD("[ENTER]");

	if (!type) {
		MWR_LOGE("[ERROR] Invalid parameter");
		return -1;
	}

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->get_audio_source_type(type);

	MWR_LOGD("[END] type(%s)", *type);
	return 0;
}

int wakeup_manager_set_wake_word_audio_require_flag(bool require)
{
	MWR_LOGD("[ENTER] : %d", require);

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->set_wake_word_audio_require_flag(require);

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_assistant_language(const char* appid, const char* language)
{
	MWR_LOGD("[ENTER]");

	string appid_string;
	string language_string;
	if (NULL == appid || NULL == language) {
		MWR_LOGE("[ERROR] Parameter is invalid, appid(%s), language(%s)", appid, language);
		return -1;
	}
	appid_string = appid;
	language_string = language;

	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->set_assistant_language(appid_string, language_string);

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_wakeup_event_callback(wakeup_service_wakeup_event_cb callback, void* user_data)
{
	MWR_LOGD("[ENTER]");

	if (NULL == callback) {
		MWR_LOGE("[ERROR] Input parameter is NULL");
		return -1;
	}

	g_wakeup_event_cb = callback;
	g_wakeup_event_user_data = user_data;

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_utterance_streaming_callback(wakeup_service_speech_streaming_cb callback, void* user_data)
{
	MWR_LOGD("[ENTER]");

	if (NULL == callback) {
		MWR_LOGE("[ERROR] Input parameter is NULL");
		return -1;
	}

	g_utterance_streaming_cb = callback;
	g_utterance_streaming_user_data = user_data;

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_previous_utterance_streaming_callback(wakeup_service_speech_streaming_cb callback, void* user_data)
{
	MWR_LOGD("[ENTER]");

	if (NULL == callback) {
		MWR_LOGE("[ERROR] Input parameter is NULL");
		return -1;
	}

	g_previous_utterance_streaming_cb = callback;
	g_previous_utterance_streaming_user_data = user_data;

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_follow_up_streaming_callback(wakeup_service_speech_streaming_cb callback, void* user_data)
{
	MWR_LOGD("[ENTER]");

	if (NULL == callback) {
		MWR_LOGE("[ERROR] Input parameter is NULL");
		return -1;
	}

	g_follow_up_streaming_cb = callback;
	g_follow_up_streaming_user_data = user_data;

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_speech_status_callback(wakeup_service_speech_status_cb callback, void* user_data)
{
	MWR_LOGD("[ENTER]");

	if (NULL == callback) {
		MWR_LOGE("[ERROR] Input parameter is NULL");
		return -1;
	}

	g_speech_status_cb = callback;
	g_speech_status_user_data = user_data;

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_setting_changed_callback(wakeup_service_setting_changed_cb callback, void* user_data)
{
	MWR_LOGD("[ENTER]");

	if (NULL == callback) {
		MWR_LOGE("[ERROR] Input parameter is NULL");
		return -1;
	}

	g_setting_changed_cb = callback;
	g_setting_changed_user_data = user_data;

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_error_callback(wakeup_service_error_cb callback, void* user_data)
{
	MWR_LOGD("[ENTER]");

	if (NULL == callback) {
		MWR_LOGE("[ERROR] Input parameter is NULL");
		return -1;
	}

	g_error_cb = callback;
	g_error_user_data = user_data;

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_set_streaming_section_changed_callback(wakeup_service_streaming_section_changed_cb callback, void* user_data)
{
	MWR_LOGD("[ENTER]");

	if (NULL == callback) {
		MWR_LOGE("[ERROR] Input parameter is NULL");
		return -1;
	}

	g_streaming_section_changed_cb = callback;
	g_streaming_section_changed_user_data = user_data;

	MWR_LOGD("[END]");
	return 0;
}

int wakeup_manager_feed_audio_data(wakeup_speech_streaming_event_e event, void* buffer, int len)
{
	if (nullptr == g_wakeup_manager) return -1;
	g_wakeup_manager->feed_audio_data(event, buffer, len);
	return 0;
}

void CWakeupEventObserver::on_wakeup(wakeup_event_info wakeup_info)
{
	if (NULL != g_wakeup_event_cb) {
		g_wakeup_event_cb(wakeup_info, g_wakeup_event_user_data);
	}
}

void CWakeupEventObserver::on_streaming_audio_data(
	wakeup_speech_streaming_event_e event, void* buffer, unsigned int len)
{
	if (WAKEUP_SPEECH_STREAMING_EVENT_START == event) {
		MWR_LOGD("streaming_cb START");
	}
	if (WAKEUP_SPEECH_STREAMING_EVENT_FINISH == event) {
		MWR_LOGD("streaming_cb FINISH");
	}
	if (NULL != g_utterance_streaming_cb) {
		g_utterance_streaming_cb(event, buffer, len, g_utterance_streaming_user_data);
	} else {
		MWR_LOGI("[INFO] No service streaming callback");
	}
}

void CWakeupEventObserver::on_audio_streaming_data_section(
	ma_audio_streaming_data_section_e section)
{
	if (g_streaming_section_changed_cb) {
		g_streaming_section_changed_cb(section, g_streaming_section_changed_user_data);
	}
}

void CSettingValueObserver::on_value_changed()
{
	if (g_setting_changed_cb) {
		g_setting_changed_cb(g_setting_changed_user_data);
	}
}