#include "wakeup_settings.h"
#include "wakeup_manager_main.h"

#include <sstream>
#include <algorithm>

namespace multiassistant
{
namespace wakeup
{

/* Utility function for checking if an element exists in a container */
template<class C, class T>
static auto contains(const C& v, const T& x) -> decltype(end(v), true)
{
	return end(v) != find(begin(v), end(v), x);
}

CWakeupSettings::CWakeupSettings()
{
}

CWakeupSettings::~CWakeupSettings()
{
}

static void wakeup_setting_input_language_changed_cb(keynode_t* node, void* data)
{
	MWR_LOGD("[ENTER]");
	if (nullptr == node) return;

	CWakeupSettings* settings = static_cast<CWakeupSettings*>(data);
	if (nullptr == settings) return;

	if (VCONF_TYPE_STRING == node->type) {
		const char* value = static_cast<const char*>(node->value.s);
		vector<ISettingsEventObserver*> observers = settings->get_observers();
		for (const auto& observer : observers) {
			if (observer) {
				if (!observer->on_voice_input_language_changed(value)) {
					LOGW("[Settings WARNING] One of the observer returned false");
				}
			}
		}
	} else {
		LOGE("[Settings ERROR] the value type is not string : %d", node->type);
	}
}

static void wakeup_setting_enabled_assistants_changed_cb(keynode_t* node, void* data)
{
	MWR_LOGD("[ENTER]");
	if (nullptr == node) return;

	CWakeupSettings* settings = static_cast<CWakeupSettings*>(data);
	if (nullptr == settings) return;

	if (VCONF_TYPE_STRING == node->type) {
		vector<string> newlyAddedAssistants;
		vector<string> newlyRemovedAssistants;
		const char* value = static_cast<const char*>(node->value.s);
		if (value) {
			vector<string> previouslyEnabledAssistants = settings->get_enabled_assistants();
			vector<string> currentlyEnabledAssistants;
			string token;
			istringstream iss(value);
			currentlyEnabledAssistants.clear();
			while (getline(iss, token, ';')) {
				currentlyEnabledAssistants.push_back(token);
				MWR_LOGD("enabled_assistants : %s", token.c_str());
			}

			for (const auto& assistant : currentlyEnabledAssistants) {
				if (!contains(previouslyEnabledAssistants, assistant)) {
					newlyAddedAssistants.push_back(assistant);
				}
			}
			for (const auto& assistant : previouslyEnabledAssistants) {
				if (!contains(currentlyEnabledAssistants, assistant)) {
					newlyRemovedAssistants.push_back(assistant);
				}
			}
		}

		vector<ISettingsEventObserver*> observers = settings->get_observers();
		for (const auto& observer : observers) {
			if (observer) {
				for (const auto& assistant : newlyAddedAssistants) {
					if (!observer->on_assistant_enabled_info_changed(assistant.c_str(), true)) {
						LOGW("[Settings WARNING] One of the observer returned false");
					}
				}
				for (const auto& assistant : newlyRemovedAssistants) {
					if (!observer->on_assistant_enabled_info_changed(assistant.c_str(), false)) {
						LOGW("[Settings WARNING] One of the observer returned false");
					}
				}
			}
		}
	} else {
		LOGE("[Settings ERROR] the value type is not string : %d", node->type);
	}
}

static void wakeup_setting_default_assistant_appid_changed_cb(keynode_t* node, void* data)
{
	MWR_LOGD("[ENTER]");
	if (nullptr == node) return;

	CWakeupSettings* settings = static_cast<CWakeupSettings*>(data);
	if (nullptr == settings) return;

	if (VCONF_TYPE_STRING == node->type) {
		const char* value = static_cast<const char*>(node->value.s);
		vector<ISettingsEventObserver*> observers = settings->get_observers();
		for (const auto& observer : observers) {
			if (observer) {
				if (!observer->on_default_assistant_appid_changed(value)) {
					LOGW("[Settings WARNING] One of the observer returned false");
				}
			}
		}
	} else {
		LOGE("[Settings ERROR] the value type is not string : %d", node->type);
	}
}

void CWakeupSettings::initialize()
{
	int vconf_ret;
	char *vconf_str;
	int vconf_bool;
	double vconf_double;

	vconf_str = vconf_get_str(WAKEUP_SETTINGS_KEY_DEFAULT_ASSISTANT_APPID);
	if (vconf_str) {
		mDefaultAssistantAppid = vconf_str;
		MWR_LOGD("default_assistant_appid : %s", mDefaultAssistantAppid.c_str());
		for (const auto& observer : mObservers) {
			if (observer) {
				if (!observer->on_default_assistant_appid_changed(vconf_str)) {
					LOGW("[Settings WARNING] One of the observer returned false");
				}
			}
		}
		free(vconf_str);
		vconf_str = nullptr;
	}
	vconf_ret = vconf_get_bool(WAKEUP_SETTINGS_KEY_UI_PANEL_ENABLED, &vconf_bool);
	if (0 == vconf_ret) {
		mUiPanelEnabled = vconf_bool;
		MWR_LOGD("ui_panel_enabled : %s", (mUiPanelEnabled ? "true" : "false"));
	}
	vconf_ret = vconf_get_dbl(WAKEUP_SETTINGS_KEY_CONVERSATION_TIMEOUT, &vconf_double);
	if (0 == vconf_ret) {
		mConversationTimeout = vconf_double;
		MWR_LOGD("conversation_timeout : %f", mConversationTimeout);
	}
	vconf_ret = vconf_get_bool(WAKEUP_SETTINGS_KEY_MULTIPLE_MODE, &vconf_bool);
	if (0 == vconf_ret) {
		mMultipleMode = vconf_bool;
		MWR_LOGD("multiple_mode : %s", (mMultipleMode ? "true" : "false"));
	}
	if (true == mMultipleMode) {
		vconf_str = vconf_get_str(WAKEUP_SETTINGS_KEY_ENABLED_ASSISTANTS);
		if (vconf_str) {
			string token;
			istringstream iss(vconf_str);
			mEnabledAssistants.clear();
			while (getline(iss, token, ';')) {
				mEnabledAssistants.push_back(token);
				MWR_LOGD("enabled_assistants : %s", token.c_str());
				for (const auto& observer : mObservers) {
					if (observer) {
						if (!observer->on_assistant_enabled_info_changed(token.c_str(), true)) {
							LOGW("[Settings WARNING] One of the observer returned false");
						}
					}
				}
			}
			free(vconf_str);
			vconf_str = nullptr;
		}
	}
	vconf_ret = vconf_get_dbl(WAKEUP_SETTINGS_KEY_WAKEUP_POLICY_DELAY, &vconf_double);
	if (0 == vconf_ret) {
		mWakeupPolicyDelay = vconf_double;
		MWR_LOGD("conversation_timeout : %f", mWakeupPolicyDelay);
	}
	vconf_str = vconf_get_str(WAKEUP_SETTINGS_KEY_WAKEUP_POLICY_PRIORITY);
	if (vconf_str) {
		string token;
		istringstream iss(vconf_str);
		mWakeupPolicyPriority.clear();
		while (getline(iss, token, ';')) {
			mWakeupPolicyPriority.push_back(token);
			MWR_LOGD("wakeup_policy_priority : %s", token.c_str());
		}
		free(vconf_str);
		vconf_str = nullptr;
	}
	vconf_ret = vconf_get_dbl(WAKEUP_SETTINGS_KEY_STREAMING_DURATION_MAX, &vconf_double);
	if (0 == vconf_ret) {
		mStreamingDurationMax = vconf_double;
		MWR_LOGD("streaming_duration_max : %f", mStreamingDurationMax);
	}
	vconf_str = vconf_get_str(WAKEUP_SETTINGS_KEY_VOICE_INPUT_LANGUAGE);
	if (vconf_str) {
		mVoiceInputLanguage = vconf_str;
		MWR_LOGD("voice input language : %s", mVoiceInputLanguage.c_str());
		for (const auto& observer : mObservers) {
			if (observer) {
				if (!observer->on_voice_input_language_changed(vconf_str)) {
					LOGW("[Settings WARNING] One of the observer returned false");
				}
			}
		}
		free(vconf_str);
		vconf_str = nullptr;
	}

	vconf_notify_key_changed(WAKEUP_SETTINGS_KEY_VOICE_INPUT_LANGUAGE,
		wakeup_setting_input_language_changed_cb, this);
	vconf_notify_key_changed(WAKEUP_SETTINGS_KEY_DEFAULT_ASSISTANT_APPID,
		wakeup_setting_default_assistant_appid_changed_cb, this);
	if (true == mMultipleMode) {
		vconf_notify_key_changed(WAKEUP_SETTINGS_KEY_ENABLED_ASSISTANTS,
			wakeup_setting_enabled_assistants_changed_cb, this);
	}
}

void CWakeupSettings::deinitialize()
{
	vconf_ignore_key_changed(WAKEUP_SETTINGS_KEY_VOICE_INPUT_LANGUAGE, NULL);
	vconf_ignore_key_changed(WAKEUP_SETTINGS_KEY_DEFAULT_ASSISTANT_APPID, NULL);
	if (true == mMultipleMode) {
		vconf_ignore_key_changed(WAKEUP_SETTINGS_KEY_ENABLED_ASSISTANTS, NULL);
	}
}

void CWakeupSettings::subscribe(ISettingsEventObserver *observer)
{
	mObservers.push_back(observer);
}

void CWakeupSettings::unsubscribe(ISettingsEventObserver *observer)
{
	auto iter = find(mObservers.begin(), mObservers.end(), observer);
	if (iter != mObservers.end()) {
		mObservers.erase(iter);
	}
}

vector<ISettingsEventObserver*> CWakeupSettings::get_observers()
{
	return mObservers;
}

string CWakeupSettings::get_default_assistant_appid()
{
	char *vconf_str;
	vconf_str = vconf_get_str(WAKEUP_SETTINGS_KEY_DEFAULT_ASSISTANT_APPID);
	MWR_LOGD("default_assistant_appid : %s", vconf_str);
	if (vconf_str) {
		mDefaultAssistantAppid = vconf_str;
		MWR_LOGD("default_assistant_appid : %s", mDefaultAssistantAppid.c_str());
		free(vconf_str);
		vconf_str = nullptr;
	}

	return mDefaultAssistantAppid;
}

bool CWakeupSettings::get_ui_panel_enabled()
{
	return mUiPanelEnabled;
}

float CWakeupSettings::get_conversation_timeout()
{
	return mConversationTimeout;
}

bool CWakeupSettings::get_multiple_mode()
{
	return mMultipleMode;
}

vector<string> CWakeupSettings::get_enabled_assistants()
{
	return mEnabledAssistants;
}

float CWakeupSettings::get_wakeup_policy_delay()
{
	return mWakeupPolicyDelay;
}

vector<string> CWakeupSettings::get_wakeup_policy_priority()
{
	return mWakeupPolicyPriority;
}

float CWakeupSettings::get_streaming_duration_max()
{
	return mStreamingDurationMax;
}

std::string CWakeupSettings::get_current_language(void)
{
	std::string result{"en_US"};
	char* language = vconf_get_str(WAKEUP_SETTINGS_KEY_VOICE_INPUT_LANGUAGE);
	if (language) {
		result = language;
		free(language);
	}
	return result;
}

} // wakeup
} // multiassistant
