/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <libxml/parser.h>
#include <sys/types.h>
#include <dirent.h>

#include "multi_assistant_config.h"
#include "multi_assistant_main.h"

int mas_config_parse_assistant_info(mas_config_assistant_info_cb callback, const char *path, void* user_data)
{
	xmlDocPtr doc = NULL;
	xmlNodePtr cur = NULL;
	xmlChar *key;

	int loop;
	int retry_count = 0;

	while (NULL == doc) {
		doc = xmlParseFile(path);
		if (NULL != doc) {
			break;
		}

		if (MA_RETRY_COUNT == retry_count++) {
			MAS_LOGE("[ERROR] Fail to parse file error : %s", path);
			xmlCleanupParser();
			return -1;
		}
		usleep(10000);
	}

	cur = xmlDocGetRootElement(doc);
	if (cur == NULL) {
		MAS_LOGE("[ERROR] Empty document");
		xmlFreeDoc(doc);
		return -1;
	}

	if (xmlStrcmp(cur->name, (const xmlChar *) MA_TAG_ASSISTANT_BASE)) {
		MAS_LOGE("[ERROR] The wrong type, root node is NOT %s", MA_TAG_ASSISTANT_BASE);
		xmlFreeDoc(doc);
		return -1;
	}

	cur = cur->xmlChildrenNode;
	if (cur == NULL) {
		MAS_LOGE("[ERROR] Empty document");
		xmlFreeDoc(doc);
		return -1;
	}

	/* alloc assistant info */
	ma_assistant_info_s* temp;
	temp = (ma_assistant_info_s*)calloc(1, sizeof(ma_assistant_info_s));
	if (NULL == temp) {
		MAS_LOGE("[ERROR] Fail to allocate memory");
		xmlFreeDoc(doc);
		return -1;
	}

	temp->app_id = NULL;
	temp->name = NULL;
	temp->icon_path = NULL;
	memset(temp->wakeup_list, 0x00, sizeof(temp->wakeup_list));
	memset(temp->wakeup_language, 0x00, sizeof(temp->wakeup_language));
	temp->cnt_wakeup = 0;
	memset(temp->supported_lang, 0x00, sizeof(temp->supported_lang));
	temp->cnt_lang = 0;
	temp->wakeup_engine = NULL;
	temp->custom_ui_option = false;

	while (cur != NULL) {
		if (cur->name && 0 == xmlStrcmp(cur->name, (const xmlChar *)MA_TAG_ASSISTANT_LANGUAGE_SET)) {
			xmlNodePtr child_node = cur->xmlChildrenNode;
			while (child_node != NULL) {
				if (child_node->name && 0 == xmlStrcmp(child_node->name, (const xmlChar*)MA_TAG_ASSISTANT_LANGUAGE)) {
					key = xmlNodeGetContent(child_node);
					if (key) {
						temp->supported_lang[temp->cnt_lang++] = strdup((const char*)key);
						MAS_LOGD("Language : %s", key);
						xmlFree(key);
					}
				}

				child_node = child_node->next;
			}
		} else if (cur->name && 0 == xmlStrcmp(cur->name, (const xmlChar *)MA_TAG_ASSISTANT_WAKEUP_WORD_SET)) {
			xmlNodePtr child_node = cur->xmlChildrenNode;
			while (child_node != NULL) {
				if (child_node->name && 0 == xmlStrcmp(child_node->name, (const xmlChar*)MA_TAG_ASSISTANT_WAKEUP_WORD)) {
					key = xmlNodeGetContent(child_node);
					if (key) {
						temp->wakeup_list[temp->cnt_wakeup] = strdup((const char*)key);
						MAS_LOGD("Wakeup Word : %s", key);
						xmlFree(key);
					}
					xmlChar* prop = xmlNodeGetLang(child_node);
					if (prop) {
						temp->wakeup_language[temp->cnt_wakeup] = strdup((const char*)prop);
						MAS_LOGD("Wakeup Language for %s : %s", temp->wakeup_list[temp->cnt_wakeup], prop);
						xmlFree(prop);
					}
					temp->cnt_wakeup++;
				}

				child_node = child_node->next;
			}
		} else if (cur->name && 0 == xmlStrcmp(cur->name, (const xmlChar*)MA_TAG_ASSISTANT_NAME)) {
			key = xmlNodeGetContent(cur);
			if (key) {
				temp->name = strdup((const char*)key);
				MAS_LOGD("Name : %s", key);
				xmlFree(key);
			}
		} else if (cur->name && 0 == xmlStrcmp(cur->name, (const xmlChar*)MA_TAG_ASSISTANT_APPID)) {
			key = xmlNodeGetContent(cur);
			if (key) {
				temp->app_id = strdup((const char*)key);
				MAS_LOGD("ID : %s", key);
				xmlFree(key);
			}
		} else if (cur->name && 0 == xmlStrcmp(cur->name, (const xmlChar*)MA_TAG_ASSISTANT_ICON_PATH)) {
			key = xmlNodeGetContent(cur);
			if (key) {
				temp->icon_path = strdup((const char*)key);
				MAS_LOGD("Icon Path : %s", key);
				xmlFree(key);
			}
		} else if (cur->name && 0 == xmlStrcmp(cur->name, (const xmlChar*)MA_TAG_ASSISTANT_WAKEUP_ENGINE_APPID)) {
			key = xmlNodeGetContent(cur);
			if (key) {
				temp->wakeup_engine = strdup((const char*)key);
				MAS_LOGD("Wakeup Engine : %s", key);
				xmlFree(key);
			}
		} else if (cur->name && 0 == xmlStrcmp(cur->name, (const xmlChar*)MA_TAG_ASSISTANT_CUSTOM_UI)) {
			key = xmlNodeGetContent(cur);
			if (key) {
				if (0 == xmlStrcasecmp(key, "true")) {
					temp->custom_ui_option = true;
				}
				MAS_LOGD("Use custom UI : %d", temp->custom_ui_option);
				xmlFree(key);
			}
		}

		cur = cur->next;
	}

	if (callback) {
		callback(temp->app_id, temp->name, temp->icon_path,
			temp->wakeup_list, temp->wakeup_language, temp->cnt_wakeup,
			temp->supported_lang, temp->cnt_lang,
			temp->wakeup_engine, temp->custom_ui_option, user_data);
	}

	if (temp->app_id) {
		free((void*)temp->app_id);
	}
	if (temp->name) {
		free((void*)temp->name);
	}
	if (temp->icon_path) {
		free((void*)temp->icon_path);
	}
	for (loop = 0; loop < temp->cnt_wakeup; loop++) {
		if (temp->wakeup_list[loop]) {
			free((void*)(temp->wakeup_list[loop]));
		}
	}
	for (loop = 0; loop < temp->cnt_lang; loop++) {
		if (temp->supported_lang[loop]) {
			free((void*)(temp->supported_lang[loop]));
		}
	}
	if (temp->wakeup_engine) {
		free((void*)temp->wakeup_engine);
	}

	free(temp);
	xmlFreeDoc(doc);

	return 0;
}

int mas_config_get_assistant_info(mas_config_assistant_info_cb callback, void* user_data)
{
	const char *suffix = ".xml";

	DIR *d;
	struct dirent *dir;
	d = opendir(MA_ASSISTANT_INFO);

	if (d) {
		while (NULL != (dir = readdir(d))) {
			if (suffix && strlen(suffix) <= strlen(dir->d_name)) {
				if (0 == strcmp(dir->d_name + strlen(dir->d_name) - strlen(suffix), suffix)) {
					char fullpath[_POSIX_PATH_MAX];
					snprintf(fullpath, _POSIX_PATH_MAX - 1, "%s/%s", MA_ASSISTANT_INFO, dir->d_name);
					MAS_LOGD("Parsing file : %s\n", fullpath);
					mas_config_parse_assistant_info(callback, fullpath, user_data);
				}
			}
		}
		closedir(d);
	}

	return 0;
}
