/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <dbus/dbus.h>
#include <dirent.h>
#include <dlfcn.h>
#include <Ecore.h>

#include "multi_assistant_main.h"
#include "multi_assistant_dbus_server.h"
#include "multi_assistant_dbus.h"

static DBusConnection* g_conn_sender = NULL;
static DBusConnection* g_conn_listener = NULL;

static Ecore_Fd_Handler* g_dbus_fd_handler = NULL;

int mas_dbus_reconnect()
{
	if (!g_conn_sender || !g_conn_listener) {
		mas_dbus_close_connection();

		if (0 != mas_dbus_open_connection()) {
			MAS_LOGE("[ERROR] Fail to reconnect");
			return -1;
		}

		MAS_LOGD("[DBUS] Reconnect");
		return 0;
	}

	bool sender_connected = dbus_connection_get_is_connected(g_conn_sender);
	bool listener_connected = dbus_connection_get_is_connected(g_conn_listener);
	MAS_LOGW("[DBUS] Sender(%s) Listener(%s)",
		 sender_connected ? "Connected" : "Not connected", listener_connected ? "Connected" : "Not connected");

	if (false == sender_connected || false == listener_connected) {
		mas_dbus_close_connection();

		if (0 != mas_dbus_open_connection()) {
			MAS_LOGE("[ERROR] Fail to reconnect");
			return -1;
		}

		MAS_LOGD("[DBUS] Reconnect");
	}

	return 0;
}

static int __dbus_check()
{
	if (NULL == g_conn_sender || NULL == g_conn_listener) {
		MAS_LOGE("[ERROR] NULL connection");
		return mas_dbus_reconnect();
	}
	return 0;
}

int mas_check_dbus_connection()
{
	if (NULL == g_conn_sender || NULL == g_conn_listener) {
		MAS_LOGE("[ERROR] NULL connection sender(%p), listener(%p)", g_conn_sender, g_conn_listener);
		return -1;
	}
	return 0;
}

int masc_dbus_send_hello(int pid)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;
	DBusError err;
	dbus_error_init(&err);

	msg = dbus_message_new_method_call(
			MA_CLIENT_SERVICE_NAME,
			MA_CLIENT_SERVICE_OBJECT_PATH,
			MA_CLIENT_SERVICE_INTERFACE,
			MAS_METHOD_HELLO);

	if (NULL == msg) {
		MAS_LOGE("[DBus ERROR] Request masc hello : Fail to make message");
		return -1;
	}

	int result = -1;
	DBusMessage* result_msg = NULL;
	result_msg = dbus_connection_send_with_reply_and_block(g_conn_sender, msg, 500, &err);

	if (dbus_error_is_set(&err)) {
		MAS_LOGE("[Dbus ERROR] Dbus Error (%s)", err.message);
		dbus_error_free(&err);
	}

	dbus_message_unref(msg);

	if (NULL != result_msg) {
		dbus_message_unref(result_msg);
		result = 0;
	} else {
		result = -1; //MAS_ERROR_TIMED_OUT;
	}

	return result;
}

int masc_dbus_send_error_message(int reason, const char* err_msg)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	if (NULL == g_conn_sender) {
		MAS_LOGE("[Dbus ERROR] Dbus connection is not available");
		return -1;
	}

	DBusMessage* msg = NULL;

	/* create a message */
	msg = dbus_message_new_signal(
		MA_CLIENT_SERVICE_OBJECT_PATH,	/* object name of the signal */
		MA_CLIENT_SERVICE_INTERFACE,	/* interface name of the signal */
		MAS_METHOD_ERROR);		/* name of the signal */

	if (NULL == msg) {
		MAS_LOGE("[Dbus ERROR] Fail to create error message");
		return -1;
	}

	char* temp_err_msg = NULL;

	if (err_msg)
		temp_err_msg = strdup(err_msg);
	else
		temp_err_msg = strdup("#NULL");

	dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &reason,
		DBUS_TYPE_STRING, &temp_err_msg,
		DBUS_TYPE_INVALID);

	dbus_message_set_no_reply(msg, TRUE);

	if (!dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] <<<< error message : Out Of Memory !");
	} else {
		MAS_LOGD("<<<< Send error message : reason(%d), err_msg(%s)", reason, temp_err_msg);
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);

	if (temp_err_msg)
		free(temp_err_msg);
	return 0;
}

int masc_dbus_send_streaming_audio_data(int pid, int event, unsigned char* data, unsigned int data_size)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;
	DBusError err;
	dbus_error_init(&err);

	char service_name[64];
	memset(service_name, '\0', 64);
	snprintf(service_name, 64, "%s_%d", MA_CLIENT_SERVICE_NAME, pid);

	msg = dbus_message_new_method_call(
			service_name,
			MA_CLIENT_SERVICE_OBJECT_PATH,
			MA_CLIENT_SERVICE_INTERFACE,
			MAS_METHOD_STREAMING_AUDIO_DATA);

	static int count = 0;
	if (NULL == msg) {
		MAS_LOGE(">>>> Request mas send utterance stream : Fail to make message");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD(">>>> Request mas send utterance stream : %s event(%d) %d", service_name, event, count++);
	}

	if (true != dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &event,
		DBUS_TYPE_ARRAY, DBUS_TYPE_BYTE,
		&data, data_size,
		DBUS_TYPE_INVALID)) {
		dbus_message_unref(msg);
		MAS_LOGE("[ERROR] Fail to append args");
		return -1;
	}

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send utterance stream");
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);
	return 0;
}

int masc_dbus_active_state_change(int pid, int state)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;

	DBusError err;
	dbus_error_init(&err);

	char service_name[64];
	memset(service_name, '\0', 64);
	snprintf(service_name, 64, "%s_%d", MA_CLIENT_SERVICE_NAME, pid);

	msg = dbus_message_new_method_call(
			service_name,
			MA_CLIENT_SERVICE_OBJECT_PATH,
			MA_CLIENT_SERVICE_INTERFACE,
			MAS_METHOD_ACTIVE_STATE_CHANGE);

	static int count = 0;
	if (NULL == msg) {
		MAS_LOGE(">>>> Request mas send activate message : Fail to make message");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD(">>>> Request mas send activate message : %s", service_name);
	}

	if (true != dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &state,
		DBUS_TYPE_INVALID)) {
		dbus_message_unref(msg);
		MAS_LOGE("[ERROR] Fail to append args");
		return -1;
	}

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send activate message : %d", state);
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);
	return 0;
}

int masc_dbus_send_preprocessing_information(int pid, const char* app_id)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;

	DBusError err;
	dbus_error_init(&err);

	char service_name[64];
	memset(service_name, '\0', 64);
	snprintf(service_name, 64, "%s_%d", MA_CLIENT_SERVICE_NAME, pid);

	msg = dbus_message_new_method_call(
			service_name,
			MA_CLIENT_SERVICE_OBJECT_PATH,
			MA_CLIENT_SERVICE_INTERFACE,
			MAS_METHOD_SEND_PREPROCESSING_INFORMATION);

	static int count = 0;
	if (NULL == msg) {
		MAS_LOGE(">>>> Request mas send preprocessing assistant information : Fail to make message");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD(">>>> Request mas send preprocessing assistant information : %s", service_name);
	}

	char* temp_app_id = NULL;
	if (!app_id)
		temp_app_id = strdup("#NULL");
	else
		temp_app_id = strdup(app_id);

	if (true != dbus_message_append_args(msg,
		DBUS_TYPE_STRING, &temp_app_id,
		DBUS_TYPE_INVALID)) {
		dbus_message_unref(msg);
		MAS_LOGE("[ERROR] Fail to append args");
		if (temp_app_id)
			free(temp_app_id);
		return -1;
	}

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		if (temp_app_id)
			free(temp_app_id);
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send preprocessing assistant information : %s", app_id);
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);

	if (temp_app_id)
		free(temp_app_id);

	return 0;
}

int masc_dbus_send_streaming_section_changed(int pid, int section)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;

	DBusError err;
	dbus_error_init(&err);

	char service_name[64];
	memset(service_name, '\0', 64);
	snprintf(service_name, 64, "%s_%d", MA_CLIENT_SERVICE_NAME, pid);

	msg = dbus_message_new_method_call(
			service_name,
			MA_CLIENT_SERVICE_OBJECT_PATH,
			MA_CLIENT_SERVICE_INTERFACE,
			MAS_METHOD_AUDIO_STREAMING_DATA_SECTION);

	static int count = 0;
	if (NULL == msg) {
		MAS_LOGE(">>>> Request mas send streaming section changed information : Fail to make message");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD(">>>> Request mas send streaming section changed information : %s", service_name);
	}

	if (true != dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &section,
		DBUS_TYPE_INVALID)) {
		dbus_message_unref(msg);
		MAS_LOGE("[ERROR] Fail to append args");
		return -1;
	}

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send streaming section changed information : %d", section);
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);

	return 0;
}

int masc_dbus_send_preprocessing_result(int pid, bool result)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;

	DBusError err;
	dbus_error_init(&err);

	char service_name[64];
	memset(service_name, '\0', 64);
	snprintf(service_name, 64, "%s_%d", MA_CLIENT_SERVICE_NAME, pid);

	msg = dbus_message_new_method_call(
			service_name,
			MA_CLIENT_SERVICE_OBJECT_PATH,
			MA_CLIENT_SERVICE_INTERFACE,
			MAS_METHOD_SEND_PREPROCESSING_RESULT);

	static int count = 0;
	if (NULL == msg) {
		MAS_LOGE(">>>> Request mas send preprocessing result : Fail to make message");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD(">>>> Request mas send preprocessing result : %s", service_name);
	}

	int temp_result = result;

	if (true != dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &temp_result,
		DBUS_TYPE_INVALID)) {
		dbus_message_unref(msg);
		MAS_LOGE("[ERROR] Fail to append args");
		return -1;
	}

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send preprocessing result : %d", temp_result);
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);

	return 0;
}

int masc_ui_dbus_send_hello(void)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;

	DBusError err;
	dbus_error_init(&err);

	msg = dbus_message_new_method_call(
			MA_UI_CLIENT_SERVICE_NAME,
			MA_UI_CLIENT_SERVICE_OBJECT_PATH,
			MA_UI_CLIENT_SERVICE_INTERFACE,
			MAS_METHOD_HELLO);

	if (NULL == msg) {
		MAS_LOGE("[DBus ERROR] Request masc hello : Fail to make message");
		return -1;
	}

	DBusMessage* result_msg = NULL;
	int result = 0;

	result_msg = dbus_connection_send_with_reply_and_block(g_conn_sender, msg, 500, &err);

	if (dbus_error_is_set(&err)) {
		MAS_LOGE("[Dbus ERROR] Dbus Error (%s)", err.message);
		dbus_error_free(&err);
	}

	dbus_message_unref(msg);

	if (NULL != result_msg) {
		dbus_message_unref(result_msg);
		result = 0;
	} else {
		result = -1; //ERROR_TIMED_OUT;
	}

	return result;
}

int masc_ui_dbus_send_asr_result(int pid, int event, char* asr_result)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;

	msg = dbus_message_new_method_call(
			  MA_UI_CLIENT_SERVICE_NAME,
			  MA_UI_CLIENT_SERVICE_OBJECT_PATH,
			  MA_UI_CLIENT_SERVICE_INTERFACE,
			  MAS_UI_METHOD_SEND_ASR_RESULT);

	if (NULL == msg) {
		MAS_LOGE("@@ Request multi-assistant send ASR result : Fail to make message");
		return -1; //MA_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[DEBUG] multi-assistant send ASR result, asr_result(%p)", asr_result);
	}

	char* temp_asr_result = NULL;
	if (!asr_result)
		temp_asr_result = strdup("#NULL");
	else
		temp_asr_result = strdup(asr_result);

	dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &pid,
		DBUS_TYPE_INT32, &event,
		DBUS_TYPE_STRING, &temp_asr_result,
		DBUS_TYPE_INVALID);

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		if (NULL != temp_asr_result) {
			free(temp_asr_result);
			temp_asr_result = NULL;
		}
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send ASR result");
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);

	if (temp_asr_result)
		free(temp_asr_result);
	return 0;
}

int masc_ui_dbus_send_result(int pid, const char* display_text, const char* utterance_text, const char* result_json)
{
	if (0 != __dbus_check()) {
		return -1; //MA_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;

	msg = dbus_message_new_method_call(
			  MA_UI_CLIENT_SERVICE_NAME,
			  MA_UI_CLIENT_SERVICE_OBJECT_PATH,
			  MA_UI_CLIENT_SERVICE_INTERFACE,
			  MAS_UI_METHOD_SEND_RESULT);

	if (NULL == msg) {
		MAS_LOGE("@@ Request multi-assistant send result : Fail to make message");
		return -1; //MA_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[DEBUG] multi-assistant send result");
	}
	char* temp_display_text = NULL;
	char* temp_utterance_text = NULL;
	char* temp_result_json = NULL;

#if 0
	dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &pid,
		DBUS_TYPE_STRING, &display_text,
		DBUS_TYPE_STRING, &utterance_text,
		DBUS_TYPE_STRING, &result_json,
		DBUS_TYPE_INVALID);

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		return -1; //MA_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send result");
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);
#else
	if (!display_text)
		temp_display_text = strdup("#NULL");
	else
		temp_display_text = strdup(display_text);
	if (!utterance_text)
		temp_utterance_text = strdup("#NULL");
	else
		temp_utterance_text = strdup(utterance_text);
	if (!result_json)
		temp_result_json = strdup("#NULL");
	else
		temp_result_json = strdup(result_json);

	dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &pid,
		DBUS_TYPE_STRING, &temp_display_text,
		DBUS_TYPE_STRING, &temp_utterance_text,
		DBUS_TYPE_STRING, &temp_result_json,
		DBUS_TYPE_INVALID);

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		if (temp_display_text)
			free(temp_display_text);
		if (temp_utterance_text)
			free(temp_utterance_text);
		if (temp_result_json)
			free(temp_result_json);
		return -1; //MA_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send result");
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);

	if (temp_display_text)
		free(temp_display_text);
	if (temp_utterance_text)
		free(temp_utterance_text);
	if (temp_result_json)
		free(temp_result_json);
#endif
	return 0;
}

int masc_ui_dbus_change_assistant(char* app_id)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	if (NULL == app_id) {
		MAS_LOGE("@@ Request multi-assistant send change assistant request : Fail to make message");
		return -1; //MA_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[DEBUG] multi-assistant send change assistant request app_id(%s)", app_id);
	}

	DBusMessage* msg;

	msg = dbus_message_new_method_call(
			  MA_UI_CLIENT_SERVICE_NAME,
			  MA_UI_CLIENT_SERVICE_OBJECT_PATH,
			  MA_UI_CLIENT_SERVICE_INTERFACE,
			  MAS_UI_METHOD_CHANGE_ASSISTANT);

	dbus_message_append_args(msg,
		DBUS_TYPE_STRING, &app_id,
		DBUS_TYPE_INVALID);

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send change assistant request");
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);

	return 0;
}

int masc_ui_dbus_send_error_message(int reason, const char* err_msg)
{
	if (NULL == g_conn_sender) {
		MAS_LOGE("[Dbus ERROR] Dbus connection is not available");
		return -1;
	}

	DBusMessage* msg = NULL;

	/* create a message */
	msg = dbus_message_new_signal(
		MA_UI_CLIENT_SERVICE_OBJECT_PATH,	/* object name of the signal */
		MA_UI_CLIENT_SERVICE_INTERFACE,		/* interface name of the signal */
		MAS_UI_METHOD_ERROR);				/* name of the signal */

	if (NULL == msg) {
		MAS_LOGE("[Dbus ERROR] Fail to create error message");
		return -1;
	}

	char* temp_err_msg = NULL;

	if (err_msg)
		temp_err_msg = strdup(err_msg);
	else
		temp_err_msg = strdup("#NULL");

	dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &reason,
		DBUS_TYPE_STRING, &temp_err_msg,
		DBUS_TYPE_INVALID);

	dbus_message_set_no_reply(msg, TRUE);

	if (!dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] <<<< error message : Out Of Memory !");
	} else {
		MAS_LOGD("<<<< Send error message : reason(%d), err_msg(%s)", reason, temp_err_msg);
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);

	if (temp_err_msg)
		free(temp_err_msg);
	return 0;
}

int masc_ui_dbus_send_recognition_result(int pid, int result)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;

	msg = dbus_message_new_method_call(
			  MA_UI_CLIENT_SERVICE_NAME,
			  MA_UI_CLIENT_SERVICE_OBJECT_PATH,
			  MA_UI_CLIENT_SERVICE_INTERFACE,
			  MAS_UI_METHOD_SEND_RECOGNITION_RESULT);

	if (NULL == msg) {
		MAS_LOGE("@@ Request multi-assistant send recognition result : Fail to make message");
		return -1; //MA_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[DEBUG] multi-assistant send recognition result(%d)", result);
	}

	dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &pid,
		DBUS_TYPE_INT32, &result,
		DBUS_TYPE_INVALID);

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send ASR result");
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);

	return 0;
}

int masc_ui_dbus_enable_common_ui(int enable)
{
	if (0 != __dbus_check()) {
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	DBusMessage* msg;

	msg = dbus_message_new_method_call(
			  MA_UI_CLIENT_SERVICE_NAME,
			  MA_UI_CLIENT_SERVICE_OBJECT_PATH,
			  MA_UI_CLIENT_SERVICE_INTERFACE,
			  MAS_UI_METHOD_ENABLE_COMMON_UI);

	if (NULL == msg) {
		MAS_LOGE("@@ Request multi-assistant enable common ui : Fail to make message");
		return -1; //MA_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[DEBUG] multi-assistant enable common ui (%d)", enable);
	}

	dbus_message_append_args(msg,
		DBUS_TYPE_INT32, &enable,
		DBUS_TYPE_INVALID);

	dbus_message_set_no_reply(msg, TRUE);

	if (1 != dbus_connection_send(g_conn_sender, msg, NULL)) {
		MAS_LOGE("[Dbus ERROR] Fail to Send");
		return -1; // MAS_ERROR_OPERATION_FAILED;
	} else {
		MAS_LOGD("[Dbus DEBUG] Success to Send ASR result");
		dbus_connection_flush(g_conn_sender);
	}

	dbus_message_unref(msg);

	return 0;
}

static Eina_Bool listener_event_callback(void* data, Ecore_Fd_Handler *fd_handler)
{
	if (NULL == g_conn_listener)	return ECORE_CALLBACK_RENEW;

	dbus_connection_read_write_dispatch(g_conn_listener, 50);

	while (1) {
		DBusMessage* msg = NULL;
		msg = dbus_connection_pop_message(g_conn_listener);

		if (true != dbus_connection_get_is_connected(g_conn_listener)) {
			MAS_LOGE("[ERROR] Connection is disconnected");
			return ECORE_CALLBACK_RENEW;
		}

		/* loop again if we haven't read a message */
		if (NULL == msg) {
			return ECORE_CALLBACK_RENEW;
		}

		/* client event */
		if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_HELLO)) {
			ma_service_dbus_hello(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_INITIALIZE)) {
			ma_service_dbus_initialize(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_DEINITIALIZE)) {
			ma_service_dbus_deinitialize(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_GET_RECORDING_AUDIO_FORMAT)) {
			ma_service_dbus_get_audio_format(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_GET_RECORDING_AUDIO_SOURCE_TYPE)) {
			ma_service_dbus_get_audio_source_type(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_SEND_ASR_RESULT)) {
			ma_service_dbus_send_asr_result(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_SEND_RESULT)) {
			ma_service_dbus_send_result(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_SEND_RECOGNITION_RESULT)) {
			ma_service_dbus_send_recognition_result(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_START_STREAMING_AUDIO_DATA)) {
			ma_service_dbus_start_streaming_audio_data(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_STOP_STREAMING_AUDIO_DATA)) {
			ma_service_dbus_stop_streaming_audio_data(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_UPDATE_VOICE_FEEDBACK_STATE)) {
			ma_service_dbus_update_voice_feedback_state(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_SEND_ASSISTANT_SPECIFIC_COMMAND)) {
			ma_service_dbus_send_assistant_specific_command(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_SET_BACKGROUND_VOLUME)) {
			ma_service_dbus_set_background_volume(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_SET_PREPROCESSING_ALLOW_MODE)) {
			ma_service_dbus_set_preprocessing_allow_mode(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_SEND_PREPROCESSING_RESULT)) {
			ma_service_dbus_send_preprocessing_result(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_SET_WAKE_WORD_AUDIO_REQUIRE_FLAG)) {
			ma_service_dbus_set_wake_word_audio_require_flag(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_METHOD_SET_ASSISTANT_LANGUAGE)) {
			ma_service_dbus_set_assistant_language(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_UI_METHOD_INITIALIZE)) {
			ma_service_ui_dbus_initialize(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_UI_METHOD_DEINITIALIZE)) {
			ma_service_ui_dbus_deinitialize(g_conn_listener, msg);

		} else if (dbus_message_is_method_call(msg, MA_SERVER_SERVICE_INTERFACE, MA_UI_METHOD_CHANGE_ASSISTANT)) {
			ma_service_ui_dbus_change_assistant(g_conn_listener, msg);

		} else {
			MAS_LOGD("Message is NOT valid");
			/* Invalid method */
		}
		/* free the message */
		dbus_message_unref(msg);
	}

	return ECORE_CALLBACK_RENEW;
}

static void __mas_dbus_connection_free()
{
	if (NULL != g_conn_listener) {
		dbus_connection_close(g_conn_listener);
		dbus_connection_unref(g_conn_listener);
		g_conn_listener = NULL;
	}
	if (NULL != g_conn_sender) {
		dbus_connection_close(g_conn_sender);
		dbus_connection_unref(g_conn_sender);
		g_conn_sender = NULL;
	}
}

int mas_dbus_open_connection()
{
	DBusError err;
	dbus_error_init(&err);

	int ret;

	/* Create connection for sender */
	g_conn_sender = dbus_bus_get_private(DBUS_BUS_SESSION, &err);

	if (dbus_error_is_set(&err)) {
		MAS_LOGE("[Dbus ERROR] Fail dbus_bus_get : %s", err.message);
		dbus_error_free(&err);
	}

	if (NULL == g_conn_sender) {
		MAS_LOGE("[Dbus ERROR] Fail to get dbus connection");
		return -1;
	}

	dbus_connection_set_exit_on_disconnect(g_conn_sender, false);

	/* connect to the bus and check for errors */
	g_conn_listener = dbus_bus_get_private(DBUS_BUS_SESSION, &err);

	if (dbus_error_is_set(&err)) {
		MAS_LOGE("[Dbus ERROR] Fail dbus_bus_get : %s", err.message);
		dbus_error_free(&err);
	}

	if (NULL == g_conn_listener) {
		MAS_LOGE("[Dbus ERROR] Fail to get dbus connection");
		__mas_dbus_connection_free();
		return -1;
	}

	dbus_connection_set_exit_on_disconnect(g_conn_listener, false);

	/* request our name on the bus and check for errors */
	ret = dbus_bus_request_name(g_conn_listener, MA_SERVER_SERVICE_NAME, DBUS_NAME_FLAG_REPLACE_EXISTING, &err);

	if (DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER != ret) {
		printf("Fail to be primary owner in dbus request.");
		MAS_LOGE("[Dbus ERROR] Fail to be primary owner");
		__mas_dbus_connection_free();
		return -1;
	}

	if (dbus_error_is_set(&err)) {
		MAS_LOGE("[Dbus ERROR] dbus_bus_request_name() : %s", err.message);
		dbus_error_free(&err);
		__mas_dbus_connection_free();
		return -1;
	}

	/* Flush messages which are received before fd event handler registration */
	while (DBUS_DISPATCH_DATA_REMAINS == dbus_connection_get_dispatch_status(g_conn_listener)) {
		listener_event_callback(NULL, NULL);
	}

	/* add a rule for getting signal */
	char rule[128];
	snprintf(rule, 128, "type='signal',interface='%s'", MA_SERVER_SERVICE_INTERFACE);

	/* add a rule for which messages we want to see */
	dbus_bus_add_match(g_conn_listener, rule, &err);/* see signals from the given interface */

	if (dbus_error_is_set(&err)) {
		MAS_LOGE("[Dbus ERROR] dbus_bus_add_match() : %s", err.message);
		dbus_error_free(&err);
		__mas_dbus_connection_free();
		return -1;
	}

	int fd = 0;
	if (1 != dbus_connection_get_unix_fd(g_conn_listener, &fd)) {
		MAS_LOGE("fail to get fd from dbus ");
		__mas_dbus_connection_free();
		return -1;
	} else {
		MAS_LOGD("Get fd from dbus : %d", fd);
	}


	g_dbus_fd_handler = ecore_main_fd_handler_add(fd, ECORE_FD_READ, (Ecore_Fd_Cb)listener_event_callback, g_conn_listener, NULL, NULL);

	if (NULL == g_dbus_fd_handler) {
		MAS_LOGE("[Dbus ERROR] Fail to get fd handler");
		__mas_dbus_connection_free();
		return -1;
	}

	return 0;
}

int mas_dbus_close_connection()
{
	DBusError err;
	dbus_error_init(&err);

	if (NULL != g_dbus_fd_handler) {
		ecore_main_fd_handler_del(g_dbus_fd_handler);
		g_dbus_fd_handler = NULL;
	}

	dbus_bus_release_name(g_conn_listener, MA_SERVER_SERVICE_NAME, &err);

	if (dbus_error_is_set(&err)) {
		MAS_LOGE("[Dbus ERROR] dbus_bus_release_name() : %s", err.message);
		dbus_error_free(&err);
	}

	__mas_dbus_connection_free();

	return 0;
}
