/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <tizen.h>
#include <service_app.h>
#include <app_manager.h>
#include <app.h>
#include <aul.h>
#include <malloc.h>
#include <Ecore.h>
#include <vconf.h>
#include <package_manager.h>
#include <pkgmgr-info.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "multi_assistant_main.h"
#include "multi_assistant_service.h"
#include "multi_assistant_service_plugin.h"
#include "multi_assistant_dbus.h"
#include "multi_assistant_config.h"

static const char *g_current_lang = "en_US";

#define ENABLE_MULTI_ASSISTANT_BY_DEFAULT

#define MULTI_ASSISTANT_SETTINGS_ACTIVATED "db/multi-assistant/activated"
#define WAKEUP_SETTINGS_KEY_PREPROCESSING_ASSISTANT_APPID "db/multi-assistant/preprocessing_assistant_appid"
#define WAKEUP_SETTINGS_KEY_PRELAUNCH_MODE "db/multi-assistant/prelaunch_mode"

#define MAX_MACLIENT_INFO_NUM 16
#define MAX_WAKEUP_WORDS_NUM 255
#define MAX_WAKEUP_WORD_LEN 32
#define MAX_SUPPORTED_LANGUAGES_NUM 255
#define MAX_SUPPORTED_LANGUAGE_LEN 16
typedef struct {
	bool used;
	char appid[MAX_APPID_LEN];
	char wakeup_word[MAX_WAKEUP_WORDS_NUM][MAX_WAKEUP_WORD_LEN];
	char wakeup_language[MAX_WAKEUP_WORDS_NUM][MAX_SUPPORTED_LANGUAGE_LEN];
	char wakeup_engine[MAX_APPID_LEN];
	char supported_language[MAX_SUPPORTED_LANGUAGES_NUM][MAX_SUPPORTED_LANGUAGE_LEN];
	bool custom_ui_option;

	ma_preprocessing_allow_mode_e preprocessing_allow_mode;
	char preprocessing_allow_appid[MAX_APPID_LEN];
} ma_client_info;

static ma_client_info g_maclient_info[MAX_MACLIENT_INFO_NUM];

typedef struct {
	int pid;
	char appid[MAX_APPID_LEN];
} ma_client_s;

static int g_current_maclient_info = 0;
static int g_current_preprocessing_maclient_info = -1;
static const char *g_wakeup_maclient_appid = NULL;
static package_manager_h g_pkgmgr = NULL;

static PREPROCESSING_STATE g_current_preprocessing_state = PREPROCESSING_STATE_NONE;

/* client list */
static GSList* g_client_list = NULL;

int ma_client_create(ma_client_s *info)
{
	if (NULL == info) {
		MAS_LOGE("Input parameter is NULL"); //LCOV_EXCL_LINE
		return -1;
	}

	ma_client_s* data = NULL;

	data = (ma_client_s*)calloc(1, sizeof(ma_client_s));
	if (NULL == data) {
		MAS_LOGE("[ERROR] Fail to allocate memory"); //LCOV_EXCL_LINE
		return -1;// MA_ERROR_OUT_OF_MEMORY;
	}

	*data = *info;

	g_client_list = g_slist_append(g_client_list, data);

	return 0;
}

int ma_client_destroy(ma_client_s *client)
{
	if (NULL == client) {
		MAS_LOGE("Input parameter is NULL"); //LCOV_EXCL_LINE
		return -1;// MA_ERROR_OPERATION_FAILED;
	}

	g_client_list =  g_slist_remove(g_client_list, client);

	free(client);
	client = NULL;

	return 0;
}

ma_client_s* ma_client_find_by_appid(const char *appid)
{
	if (NULL == appid) {
		MAS_LOGE("Input parameter is NULL"); //LCOV_EXCL_LINE
		return NULL;
	}

	ma_client_s *data = NULL;

	int count = g_slist_length(g_client_list);
	int i;

	for (i = 0; i < count; i++) {
		data = g_slist_nth_data(g_client_list, i);

		if (NULL != data) {
			if (0 == strncmp(data->appid, appid, MAX_APPID_LEN)) {
				return data;
			}
		}
	}

	MAS_LOGE("[ERROR] client Not found");

	return NULL;
}

ma_client_s* ma_client_find_by_pid(int pid)
{
	ma_client_s *data = NULL;

	int count = g_slist_length(g_client_list);
	int i;

	for (i = 0; i < count; i++) {
		data = g_slist_nth_data(g_client_list, i);

		if (NULL != data) {
			if (data->pid == pid) {
				return data;
			}
		}
	}

	MAS_LOGE("[ERROR] client Not found");

	return NULL;
}

bool check_preprocessing_assistant_exists()
{
	bool ret = false;

	char* vconf_str = vconf_get_str(WAKEUP_SETTINGS_KEY_PREPROCESSING_ASSISTANT_APPID);
	if (vconf_str) {
		for (int loop = 0; loop < MAX_MACLIENT_INFO_NUM; loop++) {
			if (g_maclient_info[loop].used) {
				if (strncmp(vconf_str, g_maclient_info[loop].appid, MAX_APPID_LEN) == 0) {
					ma_client_s* client = ma_client_find_by_appid(vconf_str);
					if (client && client->pid > 0) {
						ret = true;
					}
				}
			}
		}
		free(vconf_str);
		vconf_str = NULL;
	}

	MAS_LOGD("result : %d", ret);

	return ret;
}

bool is_current_preprocessing_assistant(const char* appid)
{
	if (NULL == appid) return false;

	bool ret = false;

	char* vconf_str = vconf_get_str(WAKEUP_SETTINGS_KEY_PREPROCESSING_ASSISTANT_APPID);
	if (vconf_str) {
		if (strncmp(vconf_str, appid, MAX_APPID_LEN) == 0) {
			ret = true;
		}
		free(vconf_str);
		vconf_str = NULL;
	}

	return ret;
}

int mas_client_initialize(int pid)
{
	MAS_LOGD("[Enter] pid(%d)", pid);

	char appid[MAX_APPID_LEN] = {'\0',};
	if (AUL_R_OK == aul_app_get_appid_bypid(pid, appid, sizeof(appid))) {
		appid[MAX_APPID_LEN - 1] = '\0';

		MAS_LOGD("appid for pid %d is : %s", pid, (appid ? appid : "Unknown"));

		/* Remove existing client that has same appid, if there's any */
		ma_client_s *old_client = NULL;
		old_client = ma_client_find_by_appid(appid);
		if (old_client) {
			ma_client_destroy(old_client);
			old_client = NULL;
		}

		/* And remove a client that has same pid also */
		old_client = ma_client_find_by_pid(pid);
		if (old_client) {
			ma_client_destroy(old_client);
			old_client = NULL;
		}

		ma_client_s new_client;
		new_client.pid = pid;
		strncpy(new_client.appid, appid, MAX_APPID_LEN);
		new_client.appid[MAX_APPID_LEN - 1] = '\0';
		ma_client_create(&new_client);

		const char *current_maclient_appid = NULL;
		if (g_current_maclient_info >= 0 && g_current_maclient_info < MAX_MACLIENT_INFO_NUM) {
			current_maclient_appid = g_maclient_info[g_current_maclient_info].appid;
		}

		mas_client_send_preprocessing_information(pid);
		if (current_maclient_appid && 0 == strncmp(current_maclient_appid, appid, MAX_APPID_LEN)) {
			MAS_LOGD("MA client with current maclient appid connected!");

			if (g_wakeup_maclient_appid && strncmp(g_wakeup_maclient_appid, appid, MAX_APPID_LEN) == 0) {
				g_wakeup_maclient_appid = NULL;
				mas_client_activate(pid);
				mas_process_preprocessing_state_event(PREPROCESSING_STATE_EVENT_ASSISTANT_ACTIVATED);
			} else {
				MAS_LOGE("[ERROR] g_wakeup_maclient_appid and appid differ : %s %s",
					(g_wakeup_maclient_appid ? g_wakeup_maclient_appid : "NULL"), appid);
			}
		} else {
			MAS_LOGD("MA client connected, but its appid does not match with current maclient");
		}
	} else {
		MAS_LOGE("[ERROR] Fail to retrieve appid");
	}

	return 0;
}

int mas_client_deinitialize(int pid)
{
	MAS_LOGD("[Enter] pid(%d)", pid);
	ma_client_s *client = ma_client_find_by_pid(pid);
	if (client) {
		ma_client_destroy(client);
		client = NULL;
	}
	return 0;
}

int mas_client_get_audio_format(int pid, int* rate, int* channel, int* audio_type)
{
	MAS_LOGD("[Enter] pid(%d)", pid);

	int ret = multi_assistant_service_plugin_get_recording_audio_format(rate, channel, audio_type);
	if (0 != ret){
		MAS_LOGE("[ERROR] Fail to get recording audio format, ret(%d)", ret);
	}

	return ret;
}

int mas_client_get_audio_source_type(int pid, char** type)
{
	MAS_LOGD("[Enter] pid(%d)", pid);

	int ret = multi_assistant_service_plugin_get_recording_audio_source_type(type);
	if (0 != ret){
		MAS_LOGE("[ERROR] Fail to get recording audio source type, ret(%d)", ret);
	}

	return ret;
}

int mas_client_send_preprocessing_information(int pid)
{
	int ret = -1;
	MAS_LOGD("[Enter] pid(%d)", pid);

	char* vconf_str = vconf_get_str(WAKEUP_SETTINGS_KEY_PREPROCESSING_ASSISTANT_APPID);
	MAS_LOGD("preprocessing_assistant_appid : %s", vconf_str);
	if (vconf_str) {
		ret = masc_dbus_send_preprocessing_information(pid, vconf_str);
		free(vconf_str);
		vconf_str = NULL;
	}

	return ret;
}

int mas_client_activate(int pid)
{
	int ret = -1;
	MAS_LOGD("[Enter] pid(%d)", pid);

	ret = masc_dbus_active_state_change(pid, MA_ACTIVE_STATE_ACTIVE);

	return ret;
}

int mas_client_deactivate(int pid)
{
	int ret = -1;
	MAS_LOGD("[Enter] pid(%d)", pid);

	ret = masc_dbus_active_state_change(pid, MA_ACTIVE_STATE_INACTIVE);

	return ret;
}

int mas_client_send_asr_result(int pid, int event, char* asr_result)
{
	MAS_LOGD("[Enter] pid(%d), event(%d), asr_result(%s)", pid, event, asr_result);
	int ret = masc_ui_dbus_send_asr_result(pid, event, asr_result);
	if (0 != ret){
		MAS_LOGE("[ERROR] Fail to send asr result, ret(%d)", ret);
	}

	// if final event is , launch assistant app which is invoked with wakeup word.
	/* TO_DO */
	return ret;
}

int mas_client_send_result(int pid, char* display_text, char* utterance_text, char* result_json)
{
	MAS_LOGD("[Enter] pid(%d), display_text(%s), utterance_text(%s), result_json(%s)", pid, display_text, utterance_text, result_json);
	int ret = masc_ui_dbus_send_result(pid, display_text, utterance_text, result_json);
	if (0 != ret){
		MAS_LOGE("[ERROR] Fail to send result, ret(%d)", ret);
	}

	const char* pid_appid = NULL;
	char buf[MAX_APPID_LEN] = {'\0',};
	if (AUL_R_OK == aul_app_get_appid_bypid(pid, buf, sizeof(buf))) {
		buf[MAX_APPID_LEN - 1] = '\0';
		pid_appid = buf;
	}
	multi_assistant_service_plugin_update_recognition_result(pid_appid, MA_RECOGNITION_RESULT_EVENT_SUCCESS);

	return ret;
}

int mas_client_send_recognition_result(int pid, int result)
{
	MAS_LOGD("[Enter] pid(%d), result(%d)", pid, result);
	int ret = masc_ui_dbus_send_recognition_result(pid, result);
	if (0 != ret){
		MAS_LOGE("[ERROR] Fail to send recognition result, ret(%d)", ret);
	}

	const char* pid_appid = NULL;
	char buf[MAX_APPID_LEN] = {'\0',};
	if (AUL_R_OK == aul_app_get_appid_bypid(pid, buf, sizeof(buf))) {
		buf[MAX_APPID_LEN - 1] = '\0';
		pid_appid = buf;
	}
	multi_assistant_service_plugin_update_recognition_result(pid_appid, result);

	return ret;
}

int mas_client_start_streaming_audio_data(int pid, int type)
{
	int ret = -1;
	switch(type) {
		case MA_AUDIO_STREAMING_DATA_TYPE_CURRENT_UTTERANCE:
			ret = multi_assistant_service_plugin_start_streaming_utterance_data();
			mas_process_preprocessing_state_event(PREPROCESSING_STATE_EVENT_UTTERANCE_STREAMING_STARTED);
			break;
		case MA_AUDIO_STREAMING_DATA_TYPE_PREVIOUS_UTTERANCE:
			ret = multi_assistant_service_plugin_start_streaming_previous_utterance_data();
			/* Preprocessing is not required for previous utterance streaming */
			break;
		case MA_AUDIO_STREAMING_DATA_TYPE_FOLLOW_UP_SPEECH:
			ret = multi_assistant_service_plugin_start_streaming_follow_up_data();
			mas_process_preprocessing_state_event(PREPROCESSING_STATE_EVENT_FOLLOW_UP_STREAMING_STARTED);
			break;
	}
	return ret;
}

int mas_client_stop_streaming_audio_data(int pid, int type)
{
	int ret = -1;
	switch(type) {
		case MA_AUDIO_STREAMING_DATA_TYPE_CURRENT_UTTERANCE:
			ret = multi_assistant_service_plugin_stop_streaming_utterance_data();
			break;
		case MA_AUDIO_STREAMING_DATA_TYPE_PREVIOUS_UTTERANCE:
			ret = multi_assistant_service_plugin_stop_streaming_previous_utterance_data();
			break;
		case MA_AUDIO_STREAMING_DATA_TYPE_FOLLOW_UP_SPEECH:
			ret = multi_assistant_service_plugin_stop_streaming_follow_up_data();
			break;
	}
	return ret;
}

int mas_client_update_voice_feedback_state(int pid, int state)
{
	const char* pid_appid = NULL;
	char buf[MAX_APPID_LEN] = {'\0',};
	if (AUL_R_OK == aul_app_get_appid_bypid(pid, buf, sizeof(buf))) {
		buf[MAX_APPID_LEN - 1] = '\0';
		pid_appid = buf;
	}
	multi_assistant_service_plugin_update_voice_feedback_state(pid_appid, state);
	return 0;
}

int mas_client_send_assistant_specific_command(int pid, const char *command)
{
	const char* pid_appid = NULL;
	char buf[MAX_APPID_LEN] = {'\0',};
	if (AUL_R_OK == aul_app_get_appid_bypid(pid, buf, sizeof(buf))) {
		buf[MAX_APPID_LEN - 1] = '\0';
		pid_appid = buf;
	}
	multi_assistant_service_plugin_send_assistant_specific_command(pid_appid, command);
	return 0;
}

int mas_client_set_background_volume(int pid, double ratio)
{
	const char* pid_appid = NULL;
	char buf[MAX_APPID_LEN] = {'\0',};
	if (AUL_R_OK == aul_app_get_appid_bypid(pid, buf, sizeof(buf))) {
		buf[MAX_APPID_LEN - 1] = '\0';
		pid_appid = buf;
	}
	multi_assistant_service_plugin_set_background_volume(pid_appid, ratio);
	return 0;
}

int mas_client_set_preprocessing_allow_mode(int pid, int mode, const char* appid)
{
	const char* pid_appid = NULL;
	char buf[MAX_APPID_LEN] = {'\0',};
	if (AUL_R_OK == aul_app_get_appid_bypid(pid, buf, sizeof(buf))) {
		buf[MAX_APPID_LEN - 1] = '\0';
		pid_appid = buf;
	}

	for (int loop = 0; loop < MAX_MACLIENT_INFO_NUM; loop++) {
		if (g_maclient_info[loop].used) {
			if (pid_appid && strncmp(pid_appid, g_maclient_info[loop].appid, MAX_APPID_LEN) == 0) {
				g_maclient_info[loop].preprocessing_allow_mode = mode;
				if (appid) {
					strncpy(g_maclient_info[loop].preprocessing_allow_appid, appid, MAX_APPID_LEN);
					g_maclient_info[loop].preprocessing_allow_appid[MAX_APPID_LEN - 1] = '\0';
				} else {
					g_maclient_info[loop].preprocessing_allow_appid[0] = '\0';
				}
			}
		}
	}

	mas_process_preprocessing_state_event(PREPROCESSING_STATE_EVENT_PREPROCESSING_ALLOW_MODE_CHANGED);

	return 0;
}

int mas_client_send_preprocessing_result(int pid, bool result)
{
	const char* pid_appid = NULL;
	char buf[MAX_APPID_LEN] = {'\0',};
	if (AUL_R_OK == aul_app_get_appid_bypid(pid, buf, sizeof(buf))) {
		buf[MAX_APPID_LEN - 1] = '\0';
		pid_appid = buf;
	}
	if (!is_current_preprocessing_assistant(pid_appid)) return -1;

	const char *current_maclient_appid = NULL;
	if (g_current_maclient_info >= 0 && g_current_maclient_info < MAX_MACLIENT_INFO_NUM) {
		current_maclient_appid = g_maclient_info[g_current_maclient_info].appid;
	}

	if (result) {
		MAS_LOGD("Preprocessing succeeded, bring (%s) to foreground", pid_appid);
		mas_process_preprocessing_state_event(PREPROCESSING_STATE_EVENT_PREPROCESSING_SUCCEEDED);
	} else {
		MAS_LOGD("Preprocessing failed, bring (%s) to foreground", current_maclient_appid);
		mas_process_preprocessing_state_event(PREPROCESSING_STATE_EVENT_PREPROCESSING_FAILED);
	}

	ma_client_s* client = ma_client_find_by_appid(current_maclient_appid);
	if (client) {
		masc_dbus_send_preprocessing_result(client->pid, result);
	}

	return 0;
}

int mas_client_set_wake_word_audio_require_flag(int pid, bool require)
{
	const char* pid_appid = NULL;
	char buf[MAX_APPID_LEN] = {'\0',};
	if (AUL_R_OK == aul_app_get_appid_bypid(pid, buf, sizeof(buf))) {
		buf[MAX_APPID_LEN - 1] = '\0';
		pid_appid = buf;
	}

	multi_assistant_service_plugin_set_wake_word_audio_require_flag(pid_appid, require);
	return 0;
}

int mas_client_set_assistant_language(int pid, const char* language)
{
	const char* pid_appid = NULL;
	char buf[MAX_APPID_LEN] = {'\0',};
	int ret = aul_app_get_appid_bypid(pid, buf, sizeof(buf));
	if (AUL_R_OK == ret) {
		buf[MAX_APPID_LEN - 1] = '\0';
		pid_appid = buf;
	}

	multi_assistant_service_plugin_assistant_language(pid_appid, language);
	return 0;
}

int mas_ui_client_initialize(int pid)
{
	MAS_LOGD("[Enter] pid(%d)", pid);

	return 0;
}

int mas_ui_client_deinitialize(int pid)
{
	MAS_LOGD("[Enter] pid(%d)", pid);

	return 0;
}

int mas_ui_client_change_assistant(const char* appid)
{
	MAS_LOGD("[Enter]");

	if (NULL == appid) {
		MAS_LOGE("NULL parameter");
		return -1;
	}

	bool use_custom_ui = mas_get_client_custom_ui_option_by_appid(appid);
	masc_ui_dbus_enable_common_ui(!use_custom_ui);

	mas_set_current_client_by_appid(appid);
	int pid = mas_get_client_pid_by_appid(appid);
	if (pid != -1) {
		mas_bring_client_to_foreground(appid);
		mas_client_send_preprocessing_information(pid);
		mas_client_activate(pid);
		MAS_LOGD("MA Client with appid %s exists, requesting speech data", (appid ? appid : "NULL"));
		/*
		int ret = multi_assistant_service_plugin_start_streaming_utterance_data();
		if (0 != ret) {
			MAS_LOGE("[ERROR] Fail to start streaming utterance data(%d)", ret);
		}
		*/
	} else {
		// Appropriate MA Client not available, trying to launch new one
		MAS_LOGD("MA Client with appid %s does not exist, launching client", (appid ? appid : "NULL"));

		/* The appid parameter might not exist after this function call, so we use appid string in our g_maclient_info */
		for (int loop = 0; loop < MAX_MACLIENT_INFO_NUM; loop++) {
			if (g_maclient_info[loop].used &&
				0 < strlen(g_maclient_info[loop].appid) &&
				0 < strlen(g_maclient_info[loop].wakeup_word[0])) {
				if (strncmp(appid, g_maclient_info[loop].appid, MAX_APPID_LEN) == 0) {
					mas_launch_client_by_appid(g_maclient_info[loop].appid, CLIENT_LAUNCH_MODE_ACTIVATION);
				}
			}
		}
	}

	return 0;
}

int __mas_assistant_info_cb(const char* appid, const char* name, const char* icon_path,
		const char* wakeup_list[], const char* wakeup_language[], int cnt_wakeup,
		const char* supported_lang[], int cnt_lang, const char* wakeup_engine,
		bool custom_ui_option, void* user_data) {
	MAS_LOGD("__mas_assistant_info_cb called");

	if (NULL == appid) {
		MAS_LOGD("app_id NULL, returning");
		return -1;
	}

	int index = -1;
	int loop = 0;
	while(-1 == index && loop < MAX_MACLIENT_INFO_NUM) {
		if (false == g_maclient_info[loop].used) {
			index = loop;
		}
		loop++;
	}
	if (-1 != index) {
		g_maclient_info[index].used = true;
		g_maclient_info[index].preprocessing_allow_mode = MA_PREPROCESSING_ALLOW_NONE;
		g_maclient_info[index].preprocessing_allow_appid[0] = '\0';
		MAS_LOGD("app_id(%s)", appid);
		strncpy(g_maclient_info[index].appid, appid, MAX_APPID_LEN);
		g_maclient_info[index].appid[MAX_APPID_LEN - 1] = '\0';

		if (is_current_preprocessing_assistant(g_maclient_info[index].appid)) {
			g_current_preprocessing_maclient_info = index;
		}

		for (loop = 0;loop < MAX_WAKEUP_WORDS_NUM;loop++) {
			if (loop < cnt_wakeup && wakeup_list[loop]) {
				MAS_LOGD("wakeup_list(%d)(%s)(%s)", loop, wakeup_list[loop], wakeup_language[loop]);
				strncpy(g_maclient_info[index].wakeup_word[loop], wakeup_list[loop], MAX_WAKEUP_WORD_LEN);
				g_maclient_info[index].wakeup_word[loop][MAX_WAKEUP_WORD_LEN - 1] = '\0';
				if (wakeup_language[loop]) {
					strncpy(g_maclient_info[index].wakeup_language[loop], wakeup_language[loop], MAX_SUPPORTED_LANGUAGE_LEN);
					g_maclient_info[index].wakeup_language[loop][MAX_SUPPORTED_LANGUAGE_LEN - 1] = '\0';
				} else {
					strncpy(g_maclient_info[index].wakeup_language[loop], "", MAX_SUPPORTED_LANGUAGE_LEN);
				}
			} else {
				strncpy(g_maclient_info[index].wakeup_word[loop], "", MAX_WAKEUP_WORD_LEN);
			}
		}

		for (loop = 0;loop < MAX_SUPPORTED_LANGUAGES_NUM;loop++) {
			if (loop < cnt_lang && supported_lang[loop]) {
				MAS_LOGD("supported_lang(%d)(%s)", loop, supported_lang[loop]);
				strncpy(g_maclient_info[index].supported_language[loop], supported_lang[loop], MAX_SUPPORTED_LANGUAGE_LEN);
				g_maclient_info[index].supported_language[loop][MAX_SUPPORTED_LANGUAGE_LEN - 1] = '\0';
			} else {
				strncpy(g_maclient_info[index].supported_language[loop], "", MAX_SUPPORTED_LANGUAGE_LEN);
			}
		}

		MAS_LOGD("wakeup_engine(%s)", wakeup_engine);
		if (wakeup_engine) {
			strncpy(g_maclient_info[index].wakeup_engine, wakeup_engine, MAX_APPID_LEN);
			g_maclient_info[index].wakeup_engine[MAX_APPID_LEN - 1] = '\0';
		} else {
			g_maclient_info[index].wakeup_engine[0] = '\0';
			MAS_LOGW("Wakeup engine information not provided for : %s", appid);
		}
		g_maclient_info[index].custom_ui_option = custom_ui_option;
	} else {
		MAS_LOGD("Couldn't find an empty slot for storing assistant info");
	}

	MAS_LOGD("__mas_assistant_info_cb end");

	return 0;
}

static void mas_active_state_changed_cb(keynode_t* key, void* data)
{
	int vconf_value = 0;
	if (vconf_get_bool(MULTI_ASSISTANT_SETTINGS_ACTIVATED, &vconf_value) == 0) {
		MAS_LOGD("multi-assistant active state : %d\n", vconf_value);

		if (vconf_value) {
			multi_assistant_service_plugin_activate();
		} else {
			multi_assistant_service_plugin_deactivate();
		}
	}
}

static int init_plugin(void)
{
	if (0 != multi_assistant_service_plugin_initialize()) {
		MAS_LOGE("Fail to ws intialize");
		return -1;
	}

	if (0 != multi_assistant_service_plugin_set_language(g_current_lang)) {
		MAS_LOGE("Fail to ws set language");
		return -1;
	}

	if (0 == mas_config_get_assistant_info(__mas_assistant_info_cb, NULL)) {
		for (int loop = 0; loop < MAX_MACLIENT_INFO_NUM; loop++) {
			int inner_loop;
			if (g_maclient_info[loop].used &&
				0 < strlen(g_maclient_info[loop].appid)) {
				if (0 < strlen(g_maclient_info[loop].wakeup_engine)) {
					multi_assistant_service_plugin_set_assistant_wakeup_engine(
						g_maclient_info[loop].appid,
						g_maclient_info[loop].wakeup_engine);
				}
				for (inner_loop = 0; inner_loop < MAX_WAKEUP_WORDS_NUM; inner_loop++) {
					if (0 < strlen(g_maclient_info[loop].wakeup_word[inner_loop])) {
						MAS_LOGD("Registering wakeup word %s for app %s",
							g_maclient_info[loop].wakeup_word[inner_loop], g_maclient_info[loop].appid);
						if (0 != multi_assistant_service_plugin_add_assistant_wakeup_word(
							g_maclient_info[loop].appid,
							g_maclient_info[loop].wakeup_word[inner_loop],
							g_maclient_info[loop].wakeup_language[inner_loop])) {
							MAS_LOGE("Fail to add assistant's wakeup word");
						}
					}
				}
				for (inner_loop = 0; inner_loop < MAX_SUPPORTED_LANGUAGE_NUM; inner_loop++) {
					if (0 < strlen(g_maclient_info[loop].supported_language[inner_loop])) {
						MAS_LOGD("Adding language %s for app %s",
							g_maclient_info[loop].supported_language[inner_loop], g_maclient_info[loop].appid);
						if (0 != multi_assistant_service_plugin_add_assistant_language(
							g_maclient_info[loop].appid,
							g_maclient_info[loop].supported_language[inner_loop])) {
							MAS_LOGE("Fail to add assistant's language");
						}
					}
				}
			}
		}
	} else {
		MAS_LOGE("Fail to load assistant info");
	}

	if (0 != multi_assistant_service_plugin_set_callbacks()) {
		MAS_LOGE("Fail to set callbacks");
		return -1;
	}

	return 0;
}

static int deinit_plugin(void)
{
	MAS_LOGI("[ENTER]");
	if (0 != multi_assistant_service_plugin_deactivate()) {
		MAS_LOGE("Fail to deactivate");
	}
	if (0 != multi_assistant_service_plugin_deinitialize()) {
		MAS_LOGE("Fail to deinitialize");
	}
	MAS_LOGI("[END]");
	return 0;
}

static int process_activated_setting()
{
	if (0 == vconf_notify_key_changed(MULTI_ASSISTANT_SETTINGS_ACTIVATED, mas_active_state_changed_cb, NULL)) {
		/* Activate / deactivate according to the vconf key setting */
		mas_active_state_changed_cb(NULL, NULL);
	} else {
#ifdef ENABLE_MULTI_ASSISTANT_BY_DEFAULT
		/* Multi-assistant needs to be enabled by default, unless disabled explicitly */
		multi_assistant_service_plugin_activate();
		const char *default_assistant = NULL;
		if (0 == multi_assistant_service_plugin_get_default_assistant(&default_assistant)) {
			if (NULL == default_assistant) {
				if (g_maclient_info[0].used) {
					default_assistant = g_maclient_info[0].appid;
					MAS_LOGW("No default assistant, setting %s as default", default_assistant);
					multi_assistant_service_plugin_set_default_assistant(default_assistant);
				} else {
					MAS_LOGE("No default assistant, and no assistant installed");
				}
			}
		}
#endif
	}
	return 0;
}

static int init_wakeup(void)
{
	MAS_LOGD("[Enter] init_wakeup");

	int ret = mas_dbus_open_connection();
	if (0 != ret) {
		MAS_LOGE("[ERROR] Fail to open connection");
	}

	init_plugin();

	process_activated_setting();

	mas_prelaunch_default_assistant();

	/* For the case of preprocessing assistant, it always have to be launched beforehand */
	char *vconf_str;
	vconf_str = vconf_get_str(WAKEUP_SETTINGS_KEY_PREPROCESSING_ASSISTANT_APPID);
	if (vconf_str) {
		MAS_LOGD("prelaunching preprocessing_assistant_appid : %s", vconf_str);
		mas_launch_client_by_appid(vconf_str, CLIENT_LAUNCH_MODE_PRELAUNCH);
		free(vconf_str);
		vconf_str = NULL;
	}

	return 0;
}

static void deinit_wakeup(void)
{
	MAS_LOGI("[Enter] deinit_wakeup");

/*	if (NULL != g_current_lang) {
		free(g_current_lang);
		g_current_lang = NULL;
	}
*/
	deinit_plugin();

	vconf_ignore_key_changed(MULTI_ASSISTANT_SETTINGS_ACTIVATED, mas_active_state_changed_cb);

	int ret = mas_dbus_close_connection();
	if (0 != ret) {
		MAS_LOGE("[ERROR] Fail to close connection");
	}
	MAS_LOGI("[END]");
}

int mas_get_current_client_pid()
{
	int ret = -1;
	if (g_current_maclient_info >= 0 && g_current_maclient_info < MAX_MACLIENT_INFO_NUM) {
		const char *appid = g_maclient_info[g_current_maclient_info].appid;
		ma_client_s* client = ma_client_find_by_appid(appid);
		if (client) {
			ret = client->pid;
		}
	}
	return ret;
}

int mas_get_current_preprocessing_client_pid()
{
	int ret = -1;
	if (g_current_preprocessing_maclient_info >= 0 && g_current_preprocessing_maclient_info < MAX_MACLIENT_INFO_NUM) {
		const char *appid = g_maclient_info[g_current_preprocessing_maclient_info].appid;
		ma_client_s* client = ma_client_find_by_appid(appid);
		if (client) {
			ret = client->pid;
		}
	}
	return ret;
}

int mas_get_client_pid_by_appid(const char *appid)
{
	int ret = -1;

	if (appid) {
		ma_client_s *client = NULL;
		client = ma_client_find_by_appid(appid);
		if (client) {
			ret = client->pid;
		}
	}

	int status = aul_app_get_status_bypid(ret);
	if (-1 != ret && 0 > status) {
		MAS_LOGE("The PID for %s was %d, but it seems to be terminated : %d",
			(appid ? appid : "NULL"), ret, status);
		mas_client_deinitialize(ret);
		ret = -1;
	}

	return ret;
}

bool mas_get_client_custom_ui_option_by_appid(const char *appid)
{
	bool ret = false;
	for (int loop = 0; loop < MAX_MACLIENT_INFO_NUM; loop++) {
		if (g_maclient_info[loop].used &&
			0 < strlen(g_maclient_info[loop].appid) &&
			0 < strlen(g_maclient_info[loop].wakeup_word[0])) {
			if (strncmp(appid, g_maclient_info[loop].appid, MAX_APPID_LEN) == 0) {
				ret = g_maclient_info[loop].custom_ui_option;
			}
		}
	}
	return ret;
}

int mas_get_client_pid_by_wakeup_word(const char *wakeup_word)
{
	const char *appid = mas_get_client_appid_by_wakeup_word(wakeup_word);
	return mas_get_client_pid_by_appid(appid);
}

const char* mas_get_client_appid_by_wakeup_word(const char *wakeup_word)
{
	int loop;
	const char *appid = NULL;

	if (NULL == wakeup_word) return NULL;

	for (loop = 0; loop < MAX_MACLIENT_INFO_NUM && NULL == appid; loop++) {
		if (g_maclient_info[loop].used &&
			0 < strlen(g_maclient_info[loop].appid)) {
			for (int inner_loop = 0; inner_loop < MAX_WAKEUP_WORDS_NUM; inner_loop++) {
				if (0 < strlen(g_maclient_info[loop].wakeup_word[inner_loop])) {
					if (0 == strncmp(wakeup_word, g_maclient_info[loop].wakeup_word[inner_loop], MAX_WAKEUP_WORD_LEN)) {
						appid = g_maclient_info[loop].appid;
					}
				}
			}
		}
	}

	/* Perform extended search, by eliminating blank characters */
	if (NULL == appid) {
		for (loop = 0; loop < MAX_MACLIENT_INFO_NUM && NULL == appid; loop++) {
			if (g_maclient_info[loop].used &&
				0 < strlen(g_maclient_info[loop].appid)) {
				for (int inner_loop = 0; inner_loop < MAX_WAKEUP_WORDS_NUM; inner_loop++) {
					if (0 < strlen(g_maclient_info[loop].wakeup_word[inner_loop])) {
						char comparand[MAX_WAKEUP_WORD_LEN];
						int comparand_index = 0;
						for (int index = 0; index < MAX_WAKEUP_WORD_LEN; index++) {
							if (' ' != g_maclient_info[loop].wakeup_word[inner_loop][index]) {
								comparand[comparand_index++] = g_maclient_info[loop].wakeup_word[inner_loop][index];
							}
						}
						if (0 == strncmp(wakeup_word, comparand, MAX_WAKEUP_WORD_LEN)) {
							appid = g_maclient_info[loop].appid;
						}
					}
				}
			}
		}
	}

	return appid;
}

int mas_set_current_client_by_wakeup_word(const char *wakeup_word)
{
	int loop;
	int ret = -1;
	int prev_selection = g_current_maclient_info;

	for (loop = 0; loop < MAX_MACLIENT_INFO_NUM && -1 == ret; loop++) {
		if (g_maclient_info[loop].used &&
			0 < strlen(g_maclient_info[loop].appid)) {
			for (int inner_loop = 0; inner_loop < MAX_WAKEUP_WORDS_NUM; inner_loop++) {
				if (0 < strlen(g_maclient_info[loop].wakeup_word[inner_loop])) {
					if (0 == strncmp(wakeup_word, g_maclient_info[loop].wakeup_word[inner_loop], MAX_WAKEUP_WORD_LEN)) {
						g_current_maclient_info = loop;
						ret = 0;
					}
				}
			}
		}
	}
	/* Perform extended search, by eliminating blank characters */
	if (ret == -1) {
		for (loop = 0; loop < MAX_MACLIENT_INFO_NUM && -1 == ret; loop++) {
			if (g_maclient_info[loop].used &&
				0 < strlen(g_maclient_info[loop].appid)) {
				for (int inner_loop = 0; inner_loop < MAX_WAKEUP_WORDS_NUM; inner_loop++) {
					if (0 < strlen(g_maclient_info[loop].wakeup_word[inner_loop])) {
						char comparand[MAX_WAKEUP_WORD_LEN];
						int comparand_index = 0;
						for (int index = 0; index < MAX_WAKEUP_WORD_LEN; index++) {
							if (' ' != g_maclient_info[loop].wakeup_word[inner_loop][index]) {
								comparand[comparand_index++] = g_maclient_info[loop].wakeup_word[inner_loop][index];
							}
						}
						if (0 == strncmp(wakeup_word, comparand, MAX_WAKEUP_WORD_LEN)) {
							g_current_maclient_info = loop;
							ret = 0;
						}
					}
				}
			}
		}
	}

	if (g_current_maclient_info != prev_selection) {
		if (prev_selection >= 0 && prev_selection < MAX_MACLIENT_INFO_NUM) {
			mas_client_deactivate(mas_get_client_pid_by_appid(g_maclient_info[prev_selection].appid));
		}
	}

	return ret;
}

int mas_set_current_client_by_appid(const char *appid)
{
	int ret = -1;
	int prev_selection = g_current_maclient_info;

	for (int loop = 0; loop < MAX_MACLIENT_INFO_NUM; loop++) {
		if (g_maclient_info[loop].used &&
			0 < strlen(g_maclient_info[loop].appid) &&
			0 < strlen(g_maclient_info[loop].wakeup_word[0])) {
			if (strncmp(appid, g_maclient_info[loop].appid, MAX_APPID_LEN) == 0) {
				g_current_maclient_info = loop;
				ret = 0;
			}
		}
	}

	if (g_current_maclient_info != prev_selection) {
		if (prev_selection >= 0 && prev_selection < MAX_MACLIENT_INFO_NUM) {
			mas_client_deactivate(mas_get_client_pid_by_appid(g_maclient_info[prev_selection].appid));
		}
	}

	return ret;
}

int mas_launch_client_by_appid(const char *appid, CLIENT_LAUNCH_MODE launch_mode)
{
	if (NULL == appid || 0 == strlen(appid)) {
		MAS_LOGE("appid invalid, failed launching MA Client");
		return -1;
	}

	bundle *b = NULL;
	b = bundle_create();
	if (NULL == b) {
		MAS_LOGE("Failed creating bundle for aul operation");
		return -1;
	}

	int result = aul_svc_set_background_launch(b, TRUE);
	if (result < AUL_R_OK) {
		MAS_LOGE("ERROR : aul_svc_set_background_launch failed. app_id [%s] bundle[%p] result[%d : %s]",
			appid, b, result, get_error_message(result));
	}

	result = aul_launch_app(appid, b);
	if (result < AUL_R_OK) {
		MAS_LOGE("ERROR : aul_launch_app failed. app_id [%s] bundle[%p] result[%d : %s]",
			appid, b, result, get_error_message(result));
	}

	if (CLIENT_LAUNCH_MODE_ACTIVATION == launch_mode) {
		bool found = false;
		for (int loop = 0; loop < MAX_MACLIENT_INFO_NUM; loop++) {
			if (g_maclient_info[loop].used &&
				0 < strlen(g_maclient_info[loop].appid)) {
				if (strncmp(appid, g_maclient_info[loop].appid, MAX_APPID_LEN) == 0) {
					g_wakeup_maclient_appid = g_maclient_info[loop].appid;
					found = true;
				}
			}
		}
		MAS_LOGD("g_wakeup_maclient_appid : %s, %d", g_wakeup_maclient_appid, found);
	}

	if (b) bundle_free(b);
	b = NULL;

	return result;
}

int mas_bring_client_to_foreground(const char* appid)
{
	/* Bring MA client to foreground - is there a better way other than launching? */
	if (NULL == appid || 0 == strlen(appid)) {
		MAS_LOGE("appid invalid, failed launching MA Client");
		return -1;
	}

	bundle *b = NULL;
	b = bundle_create();
	if (NULL == b) {
		MAS_LOGE("Failed creating bundle for aul operation");
		return -1;
	}

	int result = aul_launch_app(appid, b);
	if (result < AUL_R_OK) {
		MAS_LOGE("ERROR : aul_launch_app failed. app_id [%s] bundle[%p] result[%d : %s]",
			appid, b, result, get_error_message(result));
	}

	if (b) bundle_free(b);
	b = NULL;

	return result;
}

int mas_launch_client_by_wakeup_word(const char *wakeup_word)
{
	const char *appid = mas_get_client_appid_by_wakeup_word(wakeup_word);
	return mas_launch_client_by_appid(appid, CLIENT_LAUNCH_MODE_ACTIVATION);
}

int mas_prelaunch_default_assistant()
{
	/* CHECK NEEDED : should the code segment below and activation logic above be moved to wakeup manger? */
	int prelaunch_mode;
	int res = vconf_get_bool(WAKEUP_SETTINGS_KEY_PRELAUNCH_MODE, &prelaunch_mode);
	if (0 == res && 0 != prelaunch_mode) {
		const char *default_assistant = NULL;
		if (0 == multi_assistant_service_plugin_get_default_assistant(&default_assistant)) {
			if (0 == aul_app_is_running(default_assistant)) {
				MAS_LOGD("prelaunching default_assistant_appid : %s", default_assistant);
				mas_launch_client_by_appid(default_assistant, CLIENT_LAUNCH_MODE_PRELAUNCH);
			}
		}
	}
	return 0;
}


int mas_process_voice_key_event(bool pressed)
{
	if (pressed) {
		multi_assistant_service_plugin_process_event(MA_PLUGIN_EVENT_VOICE_KEY_PRESSED, NULL, 0);
	} else {
		multi_assistant_service_plugin_process_event(MA_PLUGIN_EVENT_VOICE_KEY_RELEASED, NULL, 0);
	}
	return 0;
}

ma_preprocessing_allow_mode_e get_preprocessing_allow_mode(const char* appid)
{
	for (int loop = 0; loop < MAX_MACLIENT_INFO_NUM; loop++) {
		if (appid && g_maclient_info[loop].used) {
			if (strncmp(appid, g_maclient_info[loop].appid, MAX_APPID_LEN) == 0) {
				return g_maclient_info[loop].preprocessing_allow_mode;
			}
		}
	}
	return MA_PREPROCESSING_ALLOW_NONE;
}

/* This might need to be read from settings in the future, but using macro for now */
//#define BRING_PREPROCESSING_ASSISTANT_TO_FRONT

int mas_process_preprocessing_state_event(PREPROCESSING_STATE_EVENT event)
{
	const char* current_maclient_appid = NULL;
	const char* preprocessing_allow_appid = NULL;
	if (g_current_maclient_info >= 0 && g_current_maclient_info < MAX_MACLIENT_INFO_NUM) {
		current_maclient_appid = g_maclient_info[g_current_maclient_info].appid;
		preprocessing_allow_appid = g_maclient_info[g_current_maclient_info].preprocessing_allow_appid;
	}
	ma_preprocessing_allow_mode_e mode = get_preprocessing_allow_mode(current_maclient_appid);

	switch (event) {
		case PREPROCESSING_STATE_EVENT_ASSISTANT_ACTIVATED:
		{
#ifndef BRING_PREPROCESSING_ASSISTANT_TO_FRONT
			/* If there is no need to bring preprocessing assistant to front,
				current_maclient should always be brought to front */
			mas_bring_client_to_foreground(current_maclient_appid);
#endif
			g_current_preprocessing_state = PREPROCESSING_STATE_WAKEUP_PREPROCESS_DISABLED;
			if (check_preprocessing_assistant_exists()) {
				if (MA_PREPROCESSING_ALLOW_UTTERANCE == mode ||
					MA_PREPROCESSING_ALLOW_ALL == mode) {
					if (is_current_preprocessing_assistant(preprocessing_allow_appid)) {
						g_current_preprocessing_state = PREPROCESSING_STATE_WAKEUP_PREPROCESS_ENABLED;
					}
				}
			} else {
#ifdef BRING_PREPROCESSING_ASSISTANT_TO_FRONT
				/* If preprocessing assistant does not exist, there is no way to enable
					preprocessing assistant, so bring current maclient to front right away */
				mas_bring_client_to_foreground(current_maclient_appid);
#endif
			}
		}
		break;
		case PREPROCESSING_STATE_EVENT_PREPROCESSING_ALLOW_MODE_CHANGED:
		{
			g_current_preprocessing_state = PREPROCESSING_STATE_WAKEUP_PREPROCESS_DISABLED;
			/* Enable preprocessing mode only if the preprocessing assistant exists */
			if (check_preprocessing_assistant_exists()) {
				if (MA_PREPROCESSING_ALLOW_UTTERANCE == mode ||
					MA_PREPROCESSING_ALLOW_ALL == mode) {
					if (is_current_preprocessing_assistant(preprocessing_allow_appid)) {
						g_current_preprocessing_state = PREPROCESSING_STATE_WAKEUP_PREPROCESS_ENABLED;
					}
				}
			}
		}
		break;
		case PREPROCESSING_STATE_EVENT_UTTERANCE_STREAMING_STARTED:
		{
			if (PREPROCESSING_STATE_WAKEUP_PREPROCESS_ENABLED == g_current_preprocessing_state) {
				g_current_preprocessing_state = PREPROCESSING_STATE_PREPROCESSING_UTTERANCE;
			} else if (PREPROCESSING_STATE_WAKEUP_PREPROCESS_DISABLED == g_current_preprocessing_state) {
				/* If preprocessing assistant does not exist, the current_maclient
					would have been brought to front already on wakeup event */
#ifdef BRING_PREPROCESSING_ASSISTANT_TO_FRONT
				if (check_preprocessing_assistant_exists()) {
					mas_bring_client_to_foreground(current_maclient_appid);
				}
#endif
				g_current_preprocessing_state = PREPROCESSING_STATE_NONE;
			}
		}
		break;
		case PREPROCESSING_STATE_EVENT_FOLLOW_UP_STREAMING_STARTED:
		{
			g_current_preprocessing_state = PREPROCESSING_STATE_NONE;
			if (check_preprocessing_assistant_exists()) {
				if (MA_PREPROCESSING_ALLOW_FOLLOW_UP == mode ||
					MA_PREPROCESSING_ALLOW_ALL == mode) {
					g_current_preprocessing_state = PREPROCESSING_STATE_PREPROCESSING_FOLLOW_UP;
				}
			}
		}
		break;
		case PREPROCESSING_STATE_EVENT_PREPROCESSING_SUCCEEDED:
		{
#ifdef BRING_PREPROCESSING_ASSISTANT_TO_FRONT
			if (PREPROCESSING_STATE_EVENT_UTTERANCE_STREAMING_STARTED == g_current_preprocessing_state ||
				PREPROCESSING_STATE_EVENT_FOLLOW_UP_STREAMING_STARTED == g_current_preprocessing_state) {
				char* vconf_str = vconf_get_str(WAKEUP_SETTINGS_KEY_PREPROCESSING_ASSISTANT_APPID);
				MAS_LOGD("preprocessing_assistant_appid : %s", vconf_str);
				if (vconf_str) {
					mas_bring_client_to_foreground(vconf_str);
					free(vconf_str);
					vconf_str = NULL;
				}
			}
#endif
			g_current_preprocessing_state = PREPROCESSING_STATE_NONE;
		}
		break;
		case PREPROCESSING_STATE_EVENT_PREPROCESSING_FAILED:
		{
#ifdef BRING_PREPROCESSING_ASSISTANT_TO_FRONT
			if (PREPROCESSING_STATE_EVENT_UTTERANCE_STREAMING_STARTED == g_current_preprocessing_state ||
				PREPROCESSING_STATE_EVENT_FOLLOW_UP_STREAMING_STARTED == g_current_preprocessing_state) {
				mas_bring_client_to_foreground(current_maclient_appid);
			}
#endif
			g_current_preprocessing_state = PREPROCESSING_STATE_NONE;
		}
		break;
	}
	return 0;
}

static int pkg_app_list_cb(const pkgmgrinfo_appinfo_h handle, void *user_data)
{
	char *appid = NULL;

	int ret = pkgmgrinfo_appinfo_get_appid (handle, &appid);
	if (PMINFO_R_OK == ret && NULL != appid) {
		int *result = (int*)user_data;
		if (result) {
			for (int loop = 0; loop < MAX_MACLIENT_INFO_NUM;loop++) {
				if (g_maclient_info[loop].used) {
					LOGD("comparing appid : %s %s", g_maclient_info[loop].wakeup_engine, appid);
					if (0 == strncmp(g_maclient_info[loop].wakeup_engine, appid, MAX_APPID_LEN)) {
						*result = 1;
					}
				}
			}
		}
	} else {
		LOGE("pkgmgrinfo_appinfo_get_appid failed! error code=%d", ret);
		return 0;
	}

	return 0;
}

/*
INFO: Package install/update/uninstall scenario
Install and Uninstall are obviously simple.
   Install: just INSTALL
   Uninstall: just UNINSTALL
Update package (change the source codes and Run As again), there are four scenarios:
1. UPDATE
   Source code change
2. UNINSTALL -> INSTALL
   This happens when Tizen IDE Property > Tizen SDK > Rapid Development Support > Check "Enable Project specific settings"
   and change Application ID in tizen-manifest.xml file and Run As.
3. UPDATE -> INSTALL
   This happens when Tizen IDE Property > Tizen SDK > Rapid Development Support > Uncheck "Enable Project specific settings"
   and change Application ID in tizen-manifest.xml file and Run As.
   At UPDATE event, pkgid (package parameter) is invalid...
4. UPDATE
   Exceptionally, only UPDATE can be called when Application ID in tizen-manifest.xml file is changed.
   At UPDATE event, pkgid (package parameter) is valid, and only appid is changed; the previous appid is invalid.
*/
static void _package_manager_event_cb(const char *type, const char *package, package_manager_event_type_e event_type, package_manager_event_state_e event_state, int progress, package_manager_error_e error, void *user_data)
{
	int ret = 0;
	uid_t uid = getuid ();
	pkgmgrinfo_pkginfo_h handle = NULL;
	static bool in_progress = false;

	if (!package || !type)
		return;

	if (PACKAGE_MANAGER_EVENT_TYPE_UPDATE != event_type &&
		PACKAGE_MANAGER_EVENT_TYPE_INSTALL != event_type &&
		PACKAGE_MANAGER_EVENT_TYPE_UNINSTALL != event_type)
		return;

	if (PACKAGE_MANAGER_EVENT_STATE_STARTED != event_state &&
		PACKAGE_MANAGER_EVENT_STATE_COMPLETED != event_state)
		return;

	bool user = false;
	LOGD("type=%s package=%s event_type=%d event_state=%d progress=%d error=%d",
		type, package, event_type, event_state, progress, error);
	ret = pkgmgrinfo_pkginfo_get_pkginfo(package, &handle);
	if (ret != PMINFO_R_OK || NULL == handle) {
		LOGW("Failed to call pkgmgrinfo_pkginfo_get_pkginfo & get_usr_pkginfo(\"%s\",~) returned %d, uid : %d", package, ret, getuid ());
		/* Try to get in user packages */
		user = true;
		ret = pkgmgrinfo_pkginfo_get_usr_pkginfo (package, uid, &handle);
		if (ret != PMINFO_R_OK || NULL == handle) {
			LOGW("Failed to call pkgmgrinfo_pkginfo_get_pkginfo & get_usr_pkginfo(\"%s\",~) returned %d, uid : %d", package, ret, getuid ());
			return;
		}
	}

	if (user) {
		/* Try to get in user packages */
		pkgmgrinfo_appinfo_get_usr_list(handle, PMINFO_ALL_APP, pkg_app_list_cb, (void *)&ret, uid);
	}  else {
		/* Try to get in global packages */
		pkgmgrinfo_appinfo_get_list(handle, PMINFO_ALL_APP, pkg_app_list_cb, (void *)&ret);
	}
	if (1 == ret) {
		if (PACKAGE_MANAGER_EVENT_STATE_STARTED == event_state) {
			LOGD("processing PACKAGE_MANAGER_EVENT_STATE_STARTED event");
			if (false == in_progress) {
				in_progress = true;
				deinit_plugin();
			}
		} else if (PACKAGE_MANAGER_EVENT_STATE_COMPLETED == event_state) {
			LOGD("processing PACKAGE_MANAGER_EVENT_STATE_COMPLETED event");
			if (true == in_progress) {
				init_plugin();
				process_activated_setting();
				in_progress = false;
			}
		}
	}

	pkgmgrinfo_pkginfo_destroy_pkginfo(handle);

	return;
}

bool service_app_create(void *data)
{
	// Todo: add your code here.

	MAS_LOGD("[Enter] Service app create");

	if (0 != init_wakeup()) {
		MAS_LOGE("Fail to init wakeup service");
		return false;
	}

	if (!g_pkgmgr) {
		int ret = package_manager_create(&g_pkgmgr);
		if (ret == PACKAGE_MANAGER_ERROR_NONE) {
			ret = package_manager_set_event_cb(g_pkgmgr, _package_manager_event_cb, NULL);
			if (ret == PACKAGE_MANAGER_ERROR_NONE) {
				LOGD("package_manager_set_event_cb succeeded.");
			}
			else {
				LOGE("package_manager_set_event_cb failed(%d)", ret);
			}
		}
		else {
			LOGE("package_manager_create failed(%d)", ret);
		}
	}

	return true;
}

void service_app_terminate(void *data)
{
	MAS_LOGI("[ENTER]");
	if (g_pkgmgr) {
		package_manager_unset_event_cb(g_pkgmgr);
		package_manager_destroy(g_pkgmgr);
		g_pkgmgr = NULL;
	}

	// Todo: add your code here.
	deinit_wakeup();
	MAS_LOGI("[END]");
	return;
}

void service_app_control(app_control_h app_control, void *data)
{
	// Todo: add your code here.
	return;
}

static void
service_app_lang_changed(app_event_info_h event_info, void *user_data)
{
	/*APP_EVENT_LANGUAGE_CHANGED*/
	return;
}

static void
service_app_region_changed(app_event_info_h event_info, void *user_data)
{
	/*APP_EVENT_REGION_FORMAT_CHANGED*/
}

static void
service_app_low_battery(app_event_info_h event_info, void *user_data)
{
	/*APP_EVENT_LOW_BATTERY*/
}

static void
service_app_low_memory(app_event_info_h event_info, void *user_data)
{
	/*APP_EVENT_LOW_MEMORY*/
}

int main(int argc, char* argv[])
{
	char ad[50] = {0,};
	service_app_lifecycle_callback_s event_callback;
	app_event_handler_h handlers[5] = {NULL, };

	event_callback.create = service_app_create;
	event_callback.terminate = service_app_terminate;
	event_callback.app_control = service_app_control;

	service_app_add_event_handler(&handlers[APP_EVENT_LOW_BATTERY], APP_EVENT_LOW_BATTERY, service_app_low_battery, &ad);
	service_app_add_event_handler(&handlers[APP_EVENT_LOW_MEMORY], APP_EVENT_LOW_MEMORY, service_app_low_memory, &ad);
	service_app_add_event_handler(&handlers[APP_EVENT_LANGUAGE_CHANGED], APP_EVENT_LANGUAGE_CHANGED, service_app_lang_changed, &ad);
	service_app_add_event_handler(&handlers[APP_EVENT_REGION_FORMAT_CHANGED], APP_EVENT_REGION_FORMAT_CHANGED, service_app_region_changed, &ad);

	int ret = service_app_main(argc, argv, &event_callback, ad);
	LOGI("Main function exits with : %d", ret);
	return ret;
}
