/*
 * Copyright 2018  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <tizen.h>
#include <service_app.h>
#include <app_manager.h>
#include <app.h>
#include <malloc.h>
#include <Ecore.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <dlfcn.h>

#include "multi_wakeup_recognizer.h"
#include "multi_assistant_main.h"
#include "multi_assistant_service.h"
#include "multi_assistant_service_plugin.h"
#include "multi_assistant_dbus.h"

/* Sound buf save for test */
#if 0
#define BUF_SAVE_MODE
#endif

#ifdef BUF_SAVE_MODE
static char g_temp_file_name[128] = {'\0',};

static FILE* g_pFile = NULL;

static int g_count = 1;
#endif

static void *g_handle = NULL;

static wakeup_manager_interface _wakeup_manager_interface = { NULL, };
static ma_plugin_settings* g_plugin_settings = NULL;

static bool is_ui_panel_enabled()
{
	/* By default we assume the ui panel is always enabled unless explicitly turned off */
	bool ret = true;
	if (g_plugin_settings) {
		ret = g_plugin_settings->ui_panel_enabled;
	}
	MAS_LOGD("UI Panel Enabled : %d", ret);
	return ret;
}

#if 0 /* + TEST_CODE */
Eina_Bool __send_asr_result(void *data)
{
	MAS_LOGD("[ENTER]");

	if (!strcmp((char*)data, "Today's")) {
		masc_ui_dbus_send_asr_result(-1, 1, "Today's");
	}
	if (!strcmp((char*)data, "weather.")) {
		masc_ui_dbus_send_asr_result(-1, 0, "Today's weather.");
	}

	MAS_LOGD("END");
	return EINA_FALSE;
}

Eina_Bool __send_result(void *data)
{
	MAS_LOGD("[ENTER]");

	int ret = masc_ui_dbus_send_result(-1, (const char*)data, (const char*)data, "test");
	if (0 != ret) {
		MAS_LOGE("[ERROR] Fail to stop recording(%d)", ret);
		return EINA_TRUE;
	}

	MAS_LOGD("END");
	return EINA_FALSE;
}
#endif /* -TEST_CODE */

Eina_Bool process_wakeup_event_by_appid_timer(char* appid)
{
	MAS_LOGD("[ENTER] appid(%s)", appid);

	int pid = -1;
	if (!appid) return ECORE_CALLBACK_CANCEL;

	bool use_custom_ui = mas_get_client_custom_ui_option_by_appid(appid);
	bool ui_panel_enabled = is_ui_panel_enabled();
	if (ui_panel_enabled) masc_ui_dbus_enable_common_ui(!use_custom_ui);

	mas_set_current_client_by_appid(appid);
	if (ui_panel_enabled) masc_ui_dbus_change_assistant(appid);
	if ((pid = mas_get_client_pid_by_appid(appid)) != -1) {
		mas_client_send_preprocessing_information(pid);
		mas_client_activate(pid);
		mas_process_preprocessing_state_event(PREPROCESSING_STATE_EVENT_ASSISTANT_ACTIVATED);
	} else {
		// Appropriate MA Client not available, trying to launch new one
		MAS_LOGD("MA Client with appid %s does not exist, launching client", appid);
		mas_launch_client_by_appid(appid, CLIENT_LAUNCH_MODE_ACTIVATION);
	}

	if (appid) free(appid);

	MAS_LOGD("END");
	return ECORE_CALLBACK_CANCEL;
}

Eina_Bool process_wakeup_event_by_word_timer(char* wakeup_word)
{
	MAS_LOGD("[ENTER]");

	if (!wakeup_word) return EINA_FALSE;

	const char* appid = mas_get_client_appid_by_wakeup_word(wakeup_word);
	process_wakeup_event_by_appid_timer(strdup(appid));

	if (wakeup_word) free(wakeup_word);

	MAS_LOGD("END");
	return ECORE_CALLBACK_CANCEL;
}

static void __wakeup_event_cb(wakeup_event_info wakeup_info, void* user_data)
{
	MAS_LOGD("[SUCCESS] __wakeup_event_cb is called, wakeup_word(%s)", wakeup_info.wakeup_word);
	int ret = -1;

	if (is_ui_panel_enabled()) {
		int retry_cnt = 0;
		while (0 != ret) {
			ret = masc_ui_dbus_send_hello();
			retry_cnt++;
			if (5 < retry_cnt) {
				MAS_LOGE("[ERROR] Fail to receive reply for hello, ret(%d)", ret);
				break;
			}
		}
	}

#ifdef BUF_SAVE_MODE
	if (g_pFile) {
		fclose(g_pFile);
		g_pFile = NULL;
	} else {
		MAS_LOGD("[Recorder Info] File not found!");
	}

	while (1) {
		snprintf(g_temp_file_name, sizeof(g_temp_file_name), "/tmp/ma_service_%d_%d", getpid(), g_count);
		int ret = access(g_temp_file_name, 0);

		if (0 == ret) {
			MAS_LOGD("[Recorder ERROR] File is already exist");
			if (0 == remove(g_temp_file_name)) {
				MAS_LOGD("[Recorder] Remove file");
				break;
			} else {
				g_count++;
			}
		} else {
			break;
		}
	}

	MAS_LOGD("[Recorder] Temp file name=[%s]", g_temp_file_name);

	/* open test file */
	g_pFile = fopen(g_temp_file_name, "wb+x");
	if (!g_pFile) {
		MAS_LOGD("[Recorder ERROR] File not found!");
		return;
	}
	g_count++;
#endif

#if 0 /* + TEST_CODE */
	if (WAKEUP_EVENT_SUCCESS == event) {
		ecore_thread_main_loop_begin();
		ecore_timer_add(1.0, __send_asr_result, "Today's");
		ecore_thread_main_loop_end();

		ecore_thread_main_loop_begin();
		ecore_timer_add(2.0, __send_asr_result, "weather.");
		ecore_thread_main_loop_end();

		ecore_thread_main_loop_begin();
		ecore_timer_add(3.0, __send_result, "Partly cloudy with temperatures from 75 to 88");
		ecore_thread_main_loop_end();
	}
#endif /* - TEST_CODE */
	if (wakeup_info.wakeup_appid) {
		ecore_thread_main_loop_begin();
		ecore_timer_add(0.0f, process_wakeup_event_by_appid_timer, (void*)strdup(wakeup_info.wakeup_appid));
		ecore_thread_main_loop_end();
	} else if (wakeup_info.wakeup_word) {
		ecore_thread_main_loop_begin();
		ecore_timer_add(0.0f, process_wakeup_event_by_word_timer, (void*)strdup(wakeup_info.wakeup_word));
		ecore_thread_main_loop_end();
	}

}

static bool __validate_streaming_event_order(int pid, wakeup_speech_streaming_event_e *event)
{
	bool ret = false;

	static int previous_pid = -1;
	static wakeup_speech_streaming_event_e previous_event = WAKEUP_SPEECH_STREAMING_EVENT_FINISH;

	if (NULL == event) return false;

	wakeup_speech_streaming_event_e expected_sequence [][2] = {
		{WAKEUP_SPEECH_STREAMING_EVENT_START, WAKEUP_SPEECH_STREAMING_EVENT_CONTINUE},
		{WAKEUP_SPEECH_STREAMING_EVENT_START, WAKEUP_SPEECH_STREAMING_EVENT_FINISH},
		{WAKEUP_SPEECH_STREAMING_EVENT_CONTINUE, WAKEUP_SPEECH_STREAMING_EVENT_CONTINUE},
		{WAKEUP_SPEECH_STREAMING_EVENT_CONTINUE, WAKEUP_SPEECH_STREAMING_EVENT_FINISH},
		{WAKEUP_SPEECH_STREAMING_EVENT_FINISH, WAKEUP_SPEECH_STREAMING_EVENT_START},
	};

	if (pid != previous_pid) {
		/* When sending streaming event to a new client, it always sends START message first */
		previous_event = WAKEUP_SPEECH_STREAMING_EVENT_FINISH;
	}

	for (int loop = 0;loop < sizeof(expected_sequence) / sizeof(expected_sequence[0]);loop++) {
		if (previous_event == expected_sequence[loop][0] &&
			*event == expected_sequence[loop][1]) {
			ret = true;
		}
	}
	if (!ret) {
		/* In case of FINISH -> CONTINUE without START, simply modify current event value */
		if (WAKEUP_SPEECH_STREAMING_EVENT_FINISH == previous_event &&
			WAKEUP_SPEECH_STREAMING_EVENT_CONTINUE == *event) {
			*event = WAKEUP_SPEECH_STREAMING_EVENT_START;
			ret = true;

			MAS_LOGD("[WARNING] forcibly changed CONTINUE to START : %d -> %d (PID %d -> %d)",
				previous_event, *event, previous_pid, pid);
		}
	}

	if (ret) {
		previous_pid = pid;
		previous_event = *event;
	} else {
		MAS_LOGE("[ERROR] State sequence validation failed : %d -> %d (PID %d -> %d)",
			previous_event, *event, previous_pid, pid);
	}
	return ret;
}

static void handle_speech_streaming_event_failure(void *data)
{
	mas_client_send_recognition_result(0, MA_RECOGNITION_RESULT_EVENT_ERROR);
}

static void __audio_streaming_cb(wakeup_speech_streaming_event_e event, unsigned char* buffer, int len, void *user_data)
{
	if (event == WAKEUP_SPEECH_STREAMING_EVENT_FAIL) {
		ecore_main_loop_thread_safe_call_async(handle_speech_streaming_event_failure, NULL);
		return;
	}
	static int count = 0;
	if (event != WAKEUP_SPEECH_STREAMING_EVENT_CONTINUE || count % 100 == 0) {
		MAS_LOGD( "[SUCCESS] __audio_streaming_cb is called, event(%d), buffer(%p), len(%d)", event, buffer, len);
	}
	++count;

	int pid = mas_get_current_client_pid();
	int preprocessing_pid = mas_get_current_preprocessing_client_pid();
	if (pid == -1) {
		MAS_LOGE("[ERROR] Fail to retrieve pid of current MA client");
	} else {
		if (__validate_streaming_event_order(pid, &event)) {
			int ret = masc_dbus_send_streaming_audio_data(pid, event, buffer, len);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to send speech data, ret(%d)", ret);
			}
			if (pid != preprocessing_pid && -1 != preprocessing_pid) {
				int ret = masc_dbus_send_streaming_audio_data(preprocessing_pid, event, buffer, len);
				if (0 != ret) {
					MAS_LOGE("[ERROR] Fail to send speech data to preprocessing client, ret(%d)", ret);
				}
			}
		}
	}

#ifdef BUF_SAVE_MODE
	/* write pcm buffer */
	if (g_pFile)
		fwrite(buffer, 1, len, g_pFile);

	if (WAKEUP_SPEECH_STREAMING_EVENT_FINISH == event) {
		if (g_pFile) {
			MAS_LOGE("[Recorder SUCCESS] File Close");
			fclose(g_pFile);
			g_pFile = NULL;
		} else {
			MAS_LOGE("[Recorder ERROR] File not found!");
		}
	}
#endif
}

static void __speech_status_cb(wakeup_speech_status_e status, void *user_data)
{
	MAS_LOGD( "[SUCCESS] __speech_status_cb is called, status(%d)", status);
}

static void __error_cb(int error, const char* err_msg, void* user_data)
{
	MAS_LOGD( "[SUCCESS] __error_cb is called, error(%d), err_msg(%s)", error, err_msg);

	if (is_ui_panel_enabled()) {
		int ret = masc_ui_dbus_send_error_message(error, err_msg);
		if (0 != ret) {
			MAS_LOGE("[ERROR] Fail to send error message, ret(%d)", ret);
		}
	}
}

static void __setting_changed_cb(void *user_data)
{
	mas_prelaunch_default_assistant();
	MAS_LOGD( "[SUCCESS] __setting_changed_cb is called");
}

static void __streaming_section_changed_cb(ma_audio_streaming_data_section_e section, void* user_data)
{
	MAS_LOGD( "[SUCCESS] __streaming_section_changed_cb is called, section(%d)", section);

	int pid = mas_get_current_client_pid();
	int ret = masc_dbus_send_streaming_section_changed(pid, (int)section);
	if (0 != ret) {
		MAS_LOGE("[ERROR] Fail to send streaming section changed information, ret(%d)", ret);
	}
}

int multi_assistant_service_plugin_initialize(void)
{
	MAS_LOGD( "[Enter]");

	char filepath[512] = {'\0',};
	const char *default_engine_path = MA_WAKEUP_MANAGER_PATH;
	snprintf(filepath, 512, "%s/%s", default_engine_path, MA_DEFAULT_WAKEUP_MANAGER_FILENAME);

	char *error;
	g_handle = NULL;
	g_handle = dlopen(filepath, RTLD_LAZY);
	if (NULL != (error = dlerror())) {
		MAS_LOGE("[ERROR] Fail to dlopen(%s), error(%s)", filepath, error);
		return -1; //MAS_ERROR_OPERATION_FAILED;
	}

	_wakeup_manager_interface.initialize =
		(wakeup_manager_initialize)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_INITIALIZE);
	_wakeup_manager_interface.deinitialize =
		(wakeup_manager_deinitialize)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_DEINITIALIZE);
	_wakeup_manager_interface.get_settings =
		(wakeup_manager_get_settings)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_GET_SETTINGS);
	_wakeup_manager_interface.add_assistant_wakeup_word =
		(wakeup_manager_add_assistant_wakeup_word)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_ADD_ASSISTANT_WAKEUP_WORD);
	_wakeup_manager_interface.add_assistant_language =
		(wakeup_manager_add_assistant_language)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_ADD_ASSISTANT_LANGUAGE);
	_wakeup_manager_interface.set_assistant_wakeup_engine =
		(wakeup_manager_set_assistant_wakeup_engine)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_ASSISTANT_WAKEUP_ENGINE);
	_wakeup_manager_interface.set_default_assistant =
		(wakeup_manager_set_default_assistant)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_DEFAULT_ASSISTANT);
	_wakeup_manager_interface.get_default_assistant =
		(wakeup_manager_get_default_assistant)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_GET_DEFAULT_ASSISTANT);
	_wakeup_manager_interface.set_language =
		(wakeup_manager_set_language)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_LANGUAGE);
	_wakeup_manager_interface.activate =
		(wakeup_manager_activate)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_ACTIVATE);
	_wakeup_manager_interface.deactivate =
		(wakeup_manager_deactivate)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_DEACTIVATE);
	_wakeup_manager_interface.update_voice_feedback_state =
		(wakeup_manager_update_voice_feedback_state)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_UPDATE_VOICE_FEEDBACK_STATE);
	_wakeup_manager_interface.send_assistant_specific_command =
		(wakeup_manager_send_assistant_specific_command)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SEND_ASSISTANT_SPECIFIC_COMMAND);
	_wakeup_manager_interface.set_background_volume =
		(wakeup_manager_set_background_volume)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_BACKGROUND_VOLUME);
	_wakeup_manager_interface.update_recognition_result =
		(wakeup_manager_update_recognition_result)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_UPDATE_RECOGNITION_RESULT);
	_wakeup_manager_interface.process_event =
		(wakeup_manager_process_event)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_PROCESS_EVENT);
	_wakeup_manager_interface.start_streaming_utterance_data =
		(wakeup_manager_start_streaming_utterance_data)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_START_STREAMING_UTTERANCE_DATA);
	_wakeup_manager_interface.stop_streaming_utterance_data =
		(wakeup_manager_stop_streaming_utterance_data)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_STOP_STREAMING_UTTERANCE_DATA);
	_wakeup_manager_interface.start_streaming_previous_utterance_data =
		(wakeup_manager_start_streaming_previous_utterance_data)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_START_STREAMING_PREVIOUS_UTTERANCE_DATA);
	_wakeup_manager_interface.stop_streaming_previous_utterance_data =
		(wakeup_manager_stop_streaming_previous_utterance_data)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_STOP_STREAMING_PREVIOUS_UTTERANCE_DATA);
	_wakeup_manager_interface.start_streaming_follow_up_data =
		(wakeup_manager_start_streaming_follow_up_data)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_START_STREAMING_FOLLOW_UP_DATA);
	_wakeup_manager_interface.stop_streaming_follow_up_data =
		(wakeup_manager_stop_streaming_follow_up_data)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_STOP_STREAMING_FOLLOW_UP_DATA);
	_wakeup_manager_interface.get_audio_format =
		(wakeup_manager_get_audio_format)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_GET_AUDIO_FORMAT);
	_wakeup_manager_interface.get_audio_source_type =
		(wakeup_manager_get_audio_source_type)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_GET_AUDIO_SOURCE_TYPE);
	_wakeup_manager_interface.set_wake_word_audio_require_flag =
		(wakeup_manager_set_wake_word_audio_require_flag)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_WAKE_WORD_AUDIO_REQUIRE_FLAG);
	_wakeup_manager_interface.set_assistant_language =
		(wakeup_manager_set_assistant_language)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_ASSISTANT_LANGUAGE);
	_wakeup_manager_interface.set_wakeup_event_callback =
		(wakeup_manager_set_wakeup_event_callback)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_WAKEUP_EVENT_CALLBACK);
	_wakeup_manager_interface.set_utterance_streaming_callback =
		(wakeup_manager_set_utterance_streaming_callback)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_UTTERANCE_STREAMING_CALLBACK);
	_wakeup_manager_interface.set_previous_utterance_streaming_callback =
		(wakeup_manager_set_previous_utterance_streaming_callback)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_PREVIOUS_UTTERANCE_STREAMING_CALLBACK);
	_wakeup_manager_interface.set_follow_up_streaming_callback =
		(wakeup_manager_set_follow_up_streaming_callback)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_FOLLOW_UP_STREAMING_CALLBACK);
	_wakeup_manager_interface.set_speech_status_callback =
		(wakeup_manager_set_speech_status_callback)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_SPEECH_STATUS_CALLBACK);
	_wakeup_manager_interface.set_setting_changed_callback =
		(wakeup_manager_set_setting_changed_callback)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_SETTING_CHANGED_CALLBACK);
	_wakeup_manager_interface.set_error_callback =
		(wakeup_manager_set_error_callback)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_ERROR_CALLBACK);
	_wakeup_manager_interface.set_streaming_section_changed_callback =
		(wakeup_manager_set_streaming_section_changed_callback)dlsym(g_handle,
		MA_WAKEUP_MANAGER_FUNC_SET_STREAMING_SECTION_CHANGED_CALLBACK);

	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_initialize func = _wakeup_manager_interface.initialize;

		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_INITIALIZE);
		} else {
			ret = func();
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to initialize, ret(%d)", ret);
			}
		}

		wakeup_manager_get_settings get_settings_func = _wakeup_manager_interface.get_settings;

		if (NULL == get_settings_func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_GET_SETTINGS);
		} else {
			size_t struct_size;
			ret = get_settings_func(&g_plugin_settings, &struct_size);
			if (0 != ret || struct_size != sizeof(ma_plugin_settings)) {
				MAS_LOGE("[ERROR] Fail to get settings, ret(%d), size %zu", ret, struct_size);
				g_plugin_settings = NULL;
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_deinitialize(void)
{
#ifdef BUF_SAVE_MODE
	if (g_pFile) {
		fclose(g_pFile);
		g_pFile = NULL;
	} else {
		MAS_LOGD("[Recorder ERROR] File not found!");
	}
#endif

	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_deinitialize func = _wakeup_manager_interface.deinitialize;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_DEINITIALIZE);
		} else {
			ret = func();
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to deinitialize, ret(%d)", ret);
			}
		}

		dlclose(g_handle);
		g_handle = NULL;
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}

	return ret;
}

int multi_assistant_service_plugin_get_settings(ma_plugin_settings **settings, size_t *struct_size)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_get_settings func = _wakeup_manager_interface.get_settings;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_GET_SETTINGS);
		} else {
			ret = func(settings, struct_size);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to get settings, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_set_language(const char* language)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_language func = _wakeup_manager_interface.set_language;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_LANGUAGE);
		} else {
			ret = func(language);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set langauge(%s), ret(%d)", language, ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_add_assistant_wakeup_word(const char* appid, const char* wakeup_word, const char* language)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_add_assistant_wakeup_word func = _wakeup_manager_interface.add_assistant_wakeup_word;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_ADD_ASSISTANT_WAKEUP_WORD);
		} else {
			ret = func(appid, wakeup_word, language);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set wakeup word(%s)(%s)(%s), ret(%d)", appid, wakeup_word, language, ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_add_assistant_language(const char* appid, const char* language)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_add_assistant_language func = _wakeup_manager_interface.add_assistant_language;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_ADD_ASSISTANT_LANGUAGE);
		} else {
			ret = func(appid, language);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set wakeup word(%s)(%s), ret(%d)", appid, language, ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_set_assistant_wakeup_engine(const char* appid, const char* engine)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_assistant_wakeup_engine func = _wakeup_manager_interface.set_assistant_wakeup_engine;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_ASSISTANT_WAKEUP_ENGINE);
		} else {
			ret = func(appid, engine);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set wakeup engine(%s)(%s), ret(%d)", appid, engine, ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_set_default_assistant(const char* appid)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_default_assistant func = _wakeup_manager_interface.set_default_assistant;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_DEFAULT_ASSISTANT);
		} else {
			ret = func(appid);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set default assistant(%s), ret(%d)", appid, ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_get_default_assistant(const char** appid)
{
	int ret = -1;
	if (NULL == appid) {
		MAS_LOGE("[ERROR] appid is not valid");
		return ret;
	}
	if (NULL != g_handle) {
		wakeup_manager_get_default_assistant func = _wakeup_manager_interface.get_default_assistant;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_GET_DEFAULT_ASSISTANT);
		} else {
			ret = func(appid);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to get default assistant, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_activate(void)
{
	int ret = -1;
	if (NULL != g_handle) {
 		wakeup_manager_activate func = _wakeup_manager_interface.activate;
		 if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_ACTIVATE);
		} else {
			ret = func();
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to start recording, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_deactivate(void)
{
	int ret = -1;
	if (NULL != g_handle) {
 		wakeup_manager_deactivate func = _wakeup_manager_interface.deactivate;
		 if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_DEACTIVATE);
		} else {
			ret = func();
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to stop recording, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_update_voice_feedback_state(const char* appid, int state)
{
	int ret = -1;
	if (NULL != g_handle) {
 		wakeup_manager_update_voice_feedback_state func = _wakeup_manager_interface.update_voice_feedback_state;
		 if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_UPDATE_VOICE_FEEDBACK_STATE);
		} else {
			ret = func(appid, state);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to stop recording, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_send_assistant_specific_command(const char* appid, const char* command)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_send_assistant_specific_command func = _wakeup_manager_interface.send_assistant_specific_command;
		 if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SEND_ASSISTANT_SPECIFIC_COMMAND);
		} else {
			ret = func(appid, command);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to stop recording, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_set_background_volume(const char* appid, double ratio)
{
	int ret = -1;
	if (NULL != g_handle) {
 		wakeup_manager_set_background_volume func = _wakeup_manager_interface.set_background_volume;
		 if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_BACKGROUND_VOLUME);
		} else {
			ret = func(appid, ratio);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set background volume, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_update_recognition_result(const char* appid, int state)
{
	int ret = -1;
	if (NULL != g_handle) {
 		wakeup_manager_update_recognition_result func = _wakeup_manager_interface.update_recognition_result;
		 if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_UPDATE_RECOGNITION_RESULT);
		} else {
			ret = func(appid, state);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to update result state, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_process_event(int event, void *data, int len)
{
	int ret = -1;
	if (NULL != g_handle) {
 		wakeup_manager_process_event func = _wakeup_manager_interface.process_event;
		 if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_PROCESS_EVENT);
		} else {
			ret = func(event, data, len);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to stop recording, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_start_streaming_utterance_data(void)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_start_streaming_utterance_data func = _wakeup_manager_interface.start_streaming_utterance_data;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_START_STREAMING_UTTERANCE_DATA);
		} else {
			ret = func();
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to request speech data, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_stop_streaming_utterance_data(void)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_stop_streaming_utterance_data func = _wakeup_manager_interface.stop_streaming_utterance_data;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_STOP_STREAMING_UTTERANCE_DATA);
		} else {
			ret = func();
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to request speech data, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_start_streaming_previous_utterance_data(void)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_start_streaming_previous_utterance_data func = _wakeup_manager_interface.start_streaming_previous_utterance_data;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_START_STREAMING_PREVIOUS_UTTERANCE_DATA);
		} else {
			ret = func();
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to request previous speech data, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_stop_streaming_previous_utterance_data(void)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_stop_streaming_previous_utterance_data func = _wakeup_manager_interface.stop_streaming_previous_utterance_data;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_STOP_STREAMING_PREVIOUS_UTTERANCE_DATA);
		} else {
			ret = func();
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to request previous speech data, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_start_streaming_follow_up_data(void)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_start_streaming_follow_up_data func = _wakeup_manager_interface.start_streaming_follow_up_data;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_START_STREAMING_FOLLOW_UP_DATA);
		} else {
			ret = func();
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to request speech data, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_stop_streaming_follow_up_data(void)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_stop_streaming_follow_up_data func = _wakeup_manager_interface.stop_streaming_follow_up_data;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_STOP_STREAMING_FOLLOW_UP_DATA);
		} else {
			ret = func();
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to request speech data, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_get_recording_audio_format(int *rate, int *channel, int *audio_type)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_get_audio_format func = _wakeup_manager_interface.get_audio_format;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_GET_AUDIO_FORMAT);
		} else {
			ret = func(rate, channel, audio_type);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to get recording audio format, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_get_recording_audio_source_type(char** type)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_get_audio_source_type func = _wakeup_manager_interface.get_audio_source_type;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_GET_AUDIO_SOURCE_TYPE);
		} else {
			ret = func(type);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to get recording audio source type, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_set_wake_word_audio_require_flag(const char* appid, bool require)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_wake_word_audio_require_flag func = _wakeup_manager_interface.set_wake_word_audio_require_flag;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_WAKE_WORD_AUDIO_REQUIRE_FLAG);
		} else {
			ret = func(require);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set wake word audio require flag, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_assistant_language(const char* appid, const char* language)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_assistant_language func = _wakeup_manager_interface.set_assistant_language;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_ASSISTANT_LANGUAGE);
		} else {
			ret = func(appid, language);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set assistant language, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_set_callbacks(void)
{
	int ret = multi_assistant_service_plugin_set_wakeup_event_callback(__wakeup_event_cb, NULL);
	if (0 != ret) {
		MAS_LOGE("Fail to set wakeup event cb");
		return ret;
	}

	ret = multi_assistant_service_plugin_set_utterance_streaming_callback(__audio_streaming_cb, NULL);
	if (0 != ret) {
		MAS_LOGE("Fail to set utterance streaming cb");
		return ret;
	}

	ret = multi_assistant_service_plugin_set_previous_utterance_streaming_callback(__audio_streaming_cb, NULL);
	if (0 != ret) {
		MAS_LOGE("Fail to set previous utterance streaming cb");
		return ret;
	}

	ret = multi_assistant_service_plugin_set_follow_up_streaming_callback(__audio_streaming_cb, NULL);
	if (0 != ret) {
		MAS_LOGE("Fail to set follow-up streaming cb");
		return ret;
	}

	ret = multi_assistant_service_plugin_set_speech_status_callback(__speech_status_cb, NULL);
	if (0 != ret) {
		MAS_LOGE("Fail to set speech status changed cb");
		return ret;
	}

	ret = multi_assistant_service_plugin_set_setting_changed_callback(__setting_changed_cb, NULL);
	if (0 != ret) {
		MAS_LOGE("Fail to set setting changed cb");
		return ret;
	}

	ret = multi_assistant_service_plugin_set_error_callback(__error_cb, NULL);
	if (0 != ret) {
		MAS_LOGE("Fail to set error cb");
		return ret;
	}

	ret = multi_assistant_service_plugin_set_streaming_section_changed_callback(__streaming_section_changed_cb, NULL);
	if (0 != ret) {
		MAS_LOGE("Fail to set streaming section changed cb");
		return ret;
	}

	return 0;
}

int multi_assistant_service_plugin_set_wakeup_event_callback(wakeup_service_wakeup_event_cb callback, void* user_data)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_wakeup_event_callback func = _wakeup_manager_interface.set_wakeup_event_callback;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_WAKEUP_EVENT_CALLBACK);
		} else {
			ret = func(callback, user_data);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set wakeup event callback, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_set_utterance_streaming_callback(wakeup_service_speech_streaming_cb callback, void* user_data)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_utterance_streaming_callback func = _wakeup_manager_interface.set_utterance_streaming_callback;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_UTTERANCE_STREAMING_CALLBACK);
		} else {
			ret = func(callback, user_data);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set utterance streaming callback, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_set_previous_utterance_streaming_callback(wakeup_service_speech_streaming_cb callback, void* user_data)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_previous_utterance_streaming_callback func = _wakeup_manager_interface.set_previous_utterance_streaming_callback;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_PREVIOUS_UTTERANCE_STREAMING_CALLBACK);
		} else {
			ret = func(callback, user_data);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set utterance streaming callback, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_set_follow_up_streaming_callback(wakeup_service_speech_streaming_cb callback, void* user_data)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_follow_up_streaming_callback func = _wakeup_manager_interface.set_follow_up_streaming_callback;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_FOLLOW_UP_STREAMING_CALLBACK);
		} else {
			ret = func(callback, user_data);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set follow-up streaming callback, ret(%d)", ret);
			}
		}
	} else {
		MAS_LOGE("[ERROR] g_handle is not valid");
	}
	return ret;
}

int multi_assistant_service_plugin_set_speech_status_callback(wakeup_service_speech_status_cb callback, void* user_data)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_speech_status_callback func = _wakeup_manager_interface.set_speech_status_callback;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_SPEECH_STATUS_CALLBACK);
		} else {
			ret = func(callback, user_data);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set speech status callback, ret(%d)", ret);
			}
		}
	}
	return ret;
}

int multi_assistant_service_plugin_set_setting_changed_callback(wakeup_service_setting_changed_cb callback, void* user_data)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_setting_changed_callback func = _wakeup_manager_interface.set_setting_changed_callback;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_SETTING_CHANGED_CALLBACK);
		} else {
			ret = func(callback, user_data);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set setting_changed callback, ret(%d)", ret);
			}
		}
	}
	return ret;
}

int multi_assistant_service_plugin_set_error_callback(wakeup_service_error_cb callback, void* user_data)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_error_callback func = _wakeup_manager_interface.set_error_callback;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_ERROR_CALLBACK);
		} else {
			ret = func(callback, user_data);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set error callback, ret(%d)", ret);
			}
		}
	}
	return ret;
}

int multi_assistant_service_plugin_set_streaming_section_changed_callback(wakeup_service_streaming_section_changed_cb callback, void* user_data)
{
	int ret = -1;
	if (NULL != g_handle) {
		wakeup_manager_set_streaming_section_changed_callback func = _wakeup_manager_interface.set_streaming_section_changed_callback;
		if (NULL == func) {
			MAS_LOGE("[ERROR] symbol lookup failed : %s", MA_WAKEUP_MANAGER_FUNC_SET_STREAMING_SECTION_CHANGED_CALLBACK);
		} else {
			ret = func(callback, user_data);
			if (0 != ret) {
				MAS_LOGE("[ERROR] Fail to set streaming section changed callback, ret(%d)", ret);
			}
		}
	}
	return ret;
}
